
//*************************************************************************************
// GM_BezierCurve.h: interface for the BezierCurve classes.
//*************************************************************************************
#ifndef GM_BEZIERCURVE_H_
#define GM_BEZIERCURVE_H_

#include "GM_PCurve.h"

namespace GM_lib{

template <class T, int n>
class BezierCurve: public PCurve<T,n>
{
	bool			_cl;			// Closed?
	bool			_ctrl_poly;		// Shall the control-polygon be displayed? 
	bool			_ctrl_pt_moved;	// Short time status if a ctrl-point is moved
	bool			_pre_eval;		// Pre evaluation are used!

	void				init();
	int					findIndex(T t)		{return (__sam-1)*(t-startPar())/(deltaPar())+0.1; }
	void				calcCtrlPts(const DVector<Vector<T,n> >& c, T t T dt);
	const DMatrix<T>&	bernstainPoly(int k, T y, T dx = T(1));

	void localSimulate(double dt){}

protected:
	DVector<Point<T,n> >		_c;				// The control points
	DVector<Selector<T,n>*>		_edit_pts;	// The selectors, one at each control point 
	SelectorGrid<T,n>*			_grid;			// A pointer to the selector grid
	DMatrix<DMatrix<T> >		_b;				// Preevaluated basis functions in u-direction

	void _eval(T t, int d);

	T _st_p() { return T(0); }
	T _end_p(){ return T(1); }

 public:
	BezierCurve(const DVector<Point<T,n> >& c):PCurve<T,n>() {_c.setSize(c.size()); _c[0]=c; init();}
	BezierCurve(const Array<Point<T,n> >& c):PCurve<T,n>() {_c.setSize(c.size()); _c[0]=c; init();}
	BezierCurve(const BezierCurve<T,n>& c):PCurve<T,n>(c) { _c=c._c; }

	virtual
	~BezierCurve() {}

	PCurve<T,n>* split(T t) { return NULL;}

	const char* identity() const { return "BEZIER-CURVE";}
};


} // end namespace GM_lib




#include <GM_DMatrix.h>

using namespace GM_lib;

template <class T, int n>
inline
BezierCurve<T,n>::BezierCurve( const DMatrix<Vector<T,n> >& c, DisplayObject* sp, bool interpolate, T t, T dt):PCurve<T,n>() 
{
	__parent		= sp;
	_ctrl_poly		= false;
	_pre_eval		= false;
	_ctrl_pt_moved	= false;
	_cl				= false;
	if(interpolate)	{ calcCtrlPts(c,t,dt); init(); }
	else			{ _c = c;			   init(); }
}


template <class T, int n>
inline	
BezierCurve<T,n>::BezierCurve(const BezierCurve<T,n>& c):PCurve<T,n>(c) 
{ 
	_c				= c._c; 
	_clu			= c._clu;
	_clv			= c._clv;
	_ctrl_poly		= false;
	_ctrl_pt_moved	= false;
}


template <class T, int n>
void BezierCurve<T,n>::init()
{
	_edit_pts.setDim(_c.dim());
	_grid  = new SelectorGrid<T,n>(_c[0],this,GL_Color(0.0,0.9,0.0));	
	for (int i=0; i<_c.dim(); i++)
	{
		_edit_pts[r]=new Selector<T,n>( _c[r], i, this, 0.125, GL_Color(0.8,0.2,0.2), NULL);
		_edit_pts[r]->onOff();				// set default off
		insert(_edit_pts[r]);
		if(r!=0) _grid->add(_c[r-1], _c[r]);	// Vertical ilines in grid
	}
	_grid->onOff(); // set default off
	insert(_grid);

	resetPlot(__sam);
	setVisible(true);
}


template <class T, int n>
void BezierCurve<T,n>::calcCtrlPts(const DVector<Vector<T,n> >& c, T t, T dt)
{
	DMatrix<T> b = bernstainPoly(c.dim(), t, 1.0/dt);
	b.invert();
	_c = b*c; 
	for(int i=0; i<_c.dim1(); i++)
	{ 
		_c[i] -= c(0);
		_editPoints[i]->translate(-c(0));
	}
	LA_DisplayObject::translate(c(0));
}


template <class T, int n>
const DMatrix<T>& BezierCurve<T,n>::bernstainPolyNew(int k, T y, T dx )
{
	static DMatrix<T> bv;		// Matrix storing result
	int i, j, d = k-1;
	T	x = T(1) - y;
	bv.setDim(k,k);

	// Ordinary Cox/deBoor like algorithm
	bv[d-1][0] = x; bv[d-1][1] = y;
	for(i=d-2; i>=0; i--)	
	{
		bv[i][0] = x*bv[i+1][0];
		for( j=1; j<d-i; j++) bv[i][j] = y*bv[i+1][j-1]+x*bv[i+1][j];
		bv[i][j] = y*bv[i+1][j-1];
	}

	// Computing all derivatives for all basisfunctions
	bv[d][0] = -dx; bv[d][1] = dx;
	for(int d1=2; d1<k; d1++)
	{
		T dy = dx * d1;
		for( i=d; i>d-d1; i--)
		{
			bv[i][d1] = dy*bv[i][d1-1];
			for( j=d1-1; j>0; j--) bv[i][j] =  dy*(bv[i][j-1] - bv[i][j]);
			bv[i][0] = -dy*bv[i][0];
		}
	}
    return bv;
}


template <class T, int n>
void BezierCurve<T,n>::_eval(T t, int d)
{
	__p.setDim(d+1);
	int dim = min(__p.dim(),_c.dim());
	if(_pre_eval) 
	{
		int i=0;
		findIndex(t, i); 
		DMatrix<T>& b = _b[i];
		__p = b*_c;
	}
	 else
	{
		DMatrix<T> b = bernstainPoly(_c.dim(), t);		
		__p = b*_c;
	}

	for( int i = dim; i < __p.dim(); i++)
			__p[i] = Vector<T,3>(T(0));
}

#endif   /* GM_BEZIERCURVE_H_ */