//********************************************************
// Camera.h: interface for the Camera class.
//********************************************************
#ifndef _CAMERA_H__
#define _CAMERA_H__

#include <iostream>
using namespace std;
#include <GM_Stream.h>

#include <GM_LA_DisplayObject.h>
#include <GM_Scene.h>

namespace GM_lib {

//*********************************************************************
//	Inneholder:
//		privat;
//				Et navn for identifikasjon i select funksjonaliteten.
//				To matriser for lokal transformasjon av objektet.
//				En Omsluttende kule til objektet som m lages.
//				En array av pekere til alle sub-objektene til dette objektet
//		public:
//				Konstuktr med mulige default verdier og sletter
//				En display funksjon for displaying av objektet.
//				En select funksjon for plukking(select) av objektet med mus.
//				En simulate funksjon for at objektet kan bevege subobjekter.
//				En find funksjon for  finne en funksjon med et gitt navn.
//				En rotate, scale og translate funksjon for  
//				   posisjonere objektet relativt i forhold til morsobjektet
//		protected:
//				Det er to virtuelle funksjoner localDisplay og localSelect
//					som alle avledede klasser m ha sin versjon av.
//				Arbeidsdelingen er slik at display og select foretar 
//					transformasjonen lagret i matrix, og displayer alle 
//					sub-objektene til objektet. Men etter transformasjonen
//					og fr sub-objektene displayes kalles localDisplay
//					(eventuelt localSelect). I localDisplay kan en s 
//					displaye det som ikke displayes i sub-objektene og/eller
//					foreta transformasjoner av sub-objekter for  
//					 simulere bevegelser.
//				Det er ogs en insert-funksjon som avledede klasser
//				    kan bruke for  lage sub-objekt. Hvis en avledet klasse 
//					trenger tilgang til et objekt for bevegelsessimulering
//					br den lage en egen peker for det.
//*********************************************************************

class Camera : public LA_DisplayObject  
{
	static Scene	def;
	static GLuint	display_list;

	Scene*			_scene;
	int				_x,_y,_w,_h;					// Viewport position and size.

	bool			_coord_sys;
	bool			_frustrum_on;

	double			_focal_length;
	double			_eye_dist;
	double			_ed_fd;
	GLfloat			_angle_tan;

protected:
	GLfloat			_near_plane;
	GLfloat			_far_plane;
	GLfloat			_ratio;
	Frustum			_frustum;
	GL_Matrix		_frustum_mat;					// Frustrum matrix



	void	basisChange(const Vector<GLfloat,3>& x, const Vector<GLfloat,3>& y,
						const Vector<GLfloat,3>& z, const Vector<GLfloat,3>& p)
											{ _matrix.basisChangeCam(x,y,z,p);}
	virtual
	void			_setPerspective();
	void			_reset(GLfloat z = 1);	// NB!!! Only for the constructors.

	void			_select(int type_id);
	void			_display();

	DisplayObject*	_find(GLuint name);

	virtual
	void			_drawActivCam();
	virtual
	void			makeGraphic();

	// Virtual functions from DisplayObject
	GL_Matrix&		matrix();

	void			localDisplay();
	void			localSelect();


public:
	Camera(Scene& s=def);
	Camera(Scene* s);

	Camera( const Point<GLfloat,3>&  pos,
		const Point<GLfloat,3>&  look_at_pos);

	Camera( const Point<GLfloat,3>&  pos,
		const Vector<GLfloat,3>& dir,
		const Vector<GLfloat,3>& up,
		GLfloat zoom=1);

	virtual ~Camera();

	void		changeScene(Scene& s) {_scene = &s;}
	void		changeScene(Scene *s) {_scene = s;}

	void		viewPort(int& w1, int& w2, int& h1, int& h2);
	int			viewPortW() { return _w; }
	int			viewPortH() { return _h; }

	GLfloat		getNearPlane()	{ return _near_plane; }
	GLfloat		getFarPlane()	{ return _far_plane; }
	GLfloat		getRatio()		{ return _ratio; }
	GLfloat		getFocalLength(){ return _focal_length; }

	void		setEyeDist(double eye_dist=0.08);
	void		increaseEyeDist(double delta=0.01);
	void		decreaseEyeDist(double delta=0.01);

	void		setFocalDist(double focal=50.0);
	void		increaseFocalDist(double delta=1);
	void		decreaseFocalDist(double delta=1);

	void		setCutPlane(GLfloat near_plane, GLfloat far_plane);

	virtual
	DisplayObject*	lockAtTargetAtPixel(int,int);
	virtual
	double			deltaTranslate(LA_DisplayObject *);

	DisplayObject*	findSelectObject(int, int, int type_id=0);
	double			distanceToObject(int, int);
	double			distanceToObject(DisplayObject*);

	void reset();						// To be used when changing Camera.
	void reshape(int,int,int,int);		// To be used when changing size of window
	virtual void go(bool stereo=false);	// Running the Camera.
	virtual void zoom(GLfloat z);

	void coordSys(bool on=true)		{ _coord_sys=on; }
	bool isCoordSysOn()				{ return _coord_sys; }
	void frustrumOn(bool on=true)	{ _frustrum_on=on; }
	bool isFrustrumOn()				{ return _frustrum_on;}

	const char* identity() const { return "Camera";}
};




inline
DisplayObject* Camera::_find(GLuint name)
{
	return	_scene->find(name);
}

inline
void Camera::setCutPlane(GLfloat near_plane, GLfloat far_plane)
{
	_near_plane = near_plane;
	_far_plane	= far_plane;
	_angle_tan	= 13.0f*_near_plane/_focal_length;
}


inline
void  Camera::setEyeDist(double d)
{
	_eye_dist=d;
	_ed_fd=_eye_dist/_focal_length;
}
inline
void  Camera::increaseEyeDist(double d)
{
	_eye_dist+=d;
	_ed_fd=_eye_dist/_focal_length;
}
inline
void  Camera::decreaseEyeDist(double d)
{
	_eye_dist-=d;
	_ed_fd=_eye_dist/_focal_length;
}

inline
void  Camera::setFocalDist(double focal)
{
	_focal_length   = focal;
	_ed_fd			= _eye_dist/_focal_length;
	_angle_tan		= 13.0f*_near_plane/_focal_length;
}

inline
void  Camera::increaseFocalDist(double d)
{
	setFocalDist(_focal_length+d);
}

inline
void  Camera::decreaseFocalDist(double d)
{
	increaseFocalDist(-d);
}

inline
void Camera::reset()
{
	int	wp[4];
	glGetIntegerv(GL_VIEWPORT,wp);
	reshape(wp[0],wp[1],wp[2],wp[3]);
}


inline
void Camera::viewPort(int& w1, int& w2, int& h1, int& h2)
{
	w1 = _x;
	h1 = _y;
	w2 = _x+_w;
	h2 = _y+_h;
}

inline
void Camera::_display()
{
	_setPerspective();
	glViewport(_x,_y,_w,_h);
	glPushMatrix();
		_matrix.mult();
		_mat_inv.mult();
		if(_coord_sys) _drawActivCam();
		_scene->culling(_frustum);
		_scene->lighting();
		_scene->display();
	glPopMatrix();
}

} // end namespace GM_lib

#endif // _CAMERA_H__

