#include <GM_DMatrix.h>

using namespace GM_lib; 

//*********************************************************
//****************    	Constructors       ****************
//*********************************************************

template<class T>
inline
DMatrix<T>::DMatrix(int i, int j)		// Default constructor
{
	pt = (i>4 ? new DVector<T>[i]:a); n=i;
	for (int k=0; k<n; k++) pt[k].setDim(j);
}



template<class T>
inline
DMatrix<T>::DMatrix(int i, int j, T val)// Standard initializing (all elements equal)
{
	pt = (i>4 ? new DVector<T>[i]:a); n=i;
	for(int k=0; k<n; k++) pt[k].increaseDim(j, val);
}

template<class T>
inline
DMatrix<T>::DMatrix(int i, int j, const T p[])// Standard initializing from c-array
{
	pt = (i>4 ? new DVector<T>[i]:a); n=i;
	for (int k=0; k<n; k++) pt[k].setDim(j);
	copy(p);
}

template<class T>
inline
DMatrix<T>::DMatrix(const DMatrix<T>& v)// Copy constructor
{
	n=0; pt=a;
	copy(v);
}


//*********************************************************
//	Private help functions implementations
//*********************************************************

template <class T>
inline
void  DMatrix<T>::copy(const DMatrix<T>& v)
{
	if(v.n>4 && v.n>n)
	{
		if (pt != a) delete [] pt;
		pt = new DVector<T>[v.n];
	}
	n = v.n;
	for(int i=0; i<n; i++) pt[i] = v.pt[i]; 
}


template <class T>
inline
void  DMatrix<T>::copy(const T p[])		
{  
	for(int k=0,i=0; i<n; i++)
		for(int j=0;j<pt[i].dim();j++)	pt[i][j] = p[k++];
}



//*********************************************************
//	Member operators and functions implementations
//*********************************************************

template <class T>
inline
void  DMatrix<T>::setDim(int i, int j)
{ 
	if(i>4 && i>n)
	{
		if(pt != a) delete [] pt;
		pt = new DVector<T>[i];
	}
	n = i;
	for (int k=0; k<n; k++) pt[k].setDim(j); 
}

template <class T>
inline
void  DMatrix<T>::resetDim(int i, int j)
{
	int m, k = min(i,n);
	DVector<T>* tmp = (i>4 ? new DVector<T>[i]:a);
	for(m=0; m<k; m++) 
	{
		tmp[m] = pt[m];
		tmp[m].resetDim(j);
	}
	for(; m<i; m++) 
	{
		tmp[m].resetDim(j);
		tmp[m].clear();
	}
	if(pt != a) delete [] pt;
	n = i;
	pt = tmp;
}

template <class T>
inline
void  DMatrix<T>::increaseDim(int i, int j, T val, bool h_end, bool v_end)
{ 
	if(i>0)
	{
		int m = n+i;
		int k = (v_end ? n:0);
		int r = (v_end ? 0:i);
		if(m>4)
		{
			DVector<T>* tmp = new DVector<T>[m];
			for(m=r; m<r+n; m++) 
			{
				tmp[m] = pt[m-r];
				tmp[m].increaseDim(j,val,h_end);
			}
			if(pt != a) delete [] pt;
			pt = tmp;
		}
		else if(!v_end)
			for(m=n-1; m>=0; m--) 
			{
				pt[m+i] = pt[m];
				tmp[m+i].increaseDim(j,val,h_end);
			}
		for(m=k; m<i+k; m++) 
		{
			pt[m].resetdim(j);
			pt[m].clear(val);
		}
		n += i;
	}
}

template <class T>
inline
void  DMatrix<T>::setIdentity()
{
	for(int i=0; i<dim1(); i++)
		for(int j=0; j<dim2(); j++) 
			if(i==j)	(*this)[i][j]=T(1.0); 
			else		(*this)[i][j]=T(0.0);
}

//**********************************************
//	[] and () and < operators  implementations
//  and determinant() function  implementations.
//**********************************************

template <class T>
inline
DVector<T>&  DMatrix<T>::operator[](int i)
{
#ifdef DEBUG
	if (i<0 || i>=n) cerr << "Error index m " << i << " is outside(0," << n << ")\n";
#endif
	return pt[i];
}

template <class T>
inline
const DVector<T>&  DMatrix<T>::operator()(int i) const
{
#ifdef DEBUG
	if (i<0 || i>=n) cerr << "Error index m " << i << " is outside(0," << n << ")\n";
#endif
	return pt[i];
}

template <class T>
inline
T DMatrix<T>::determinant() const // Dummy not implemented !!
{
	T r = T(0);
	return r;
}

template <class T>
inline
bool DMatrix<T>::operator<(const DMatrix<T>& v) const
{
	return determinant() < v.determinant();
}

//*******************************************
//	+= -= + -  operators  implementations
//*******************************************

template<class T>
inline
DMatrix<T>& DMatrix<T>::operator+=(const DMatrix<T>& v)
{
#ifdef DEBUG
	if (n != v.dim()) 
		cerr << "Matrix dimension error, dim=" << n << " ,dim=" << v.dim() << endl;
#endif
	for (int i=0; i <n; i++) pt[i] += v.pt[i];
	return *this;
}

template<class T>
inline
DMatrix<T>& DMatrix<T>::operator-=(const DMatrix<T>& v)
{
#ifdef DEBUG
	if (n != v.dim()) 
		cerr << "Matrix dimension error, dim=" << n << " ,dim=" << v.dim() << endl;
#endif
	for (int i=0; i <n; i++) pt[i] -= v.pt[i];
	return *this;
}

template<class T>
inline
DMatrix<T>	DMatrix<T>::operator+(const DMatrix<T>& a) const
{ 
	DMatrix<T> na=(*this);
	return na += a;
}

template<class T>
inline
DMatrix<T>	DMatrix<T>::operator-(const DMatrix<T>& a) const
{
	DMatrix<T> na=(*this);
	return na -= a;
}


//*******************************************
//	Scaling  operators  implementations
//*******************************************

template <class T>
inline
DMatrix<T>& DMatrix<T>::operator*=(double d)
{  
	for(int i=0; i<n; i++) pt[i] *= d;
	return *this;
}

template <class T>
inline
DMatrix<T>& DMatrix<T>::operator/=(double d)
{  
	for(int i=0; i<n; i++) pt[i] /= d;
	return *this;
}

template <class T>
inline
DMatrix<T> DMatrix<T>::operator*(double d) const
{
	DMatrix<T> ret = *this;
	return ret *= d;
}

template <class T>
inline
DMatrix<T> DMatrix<T>::operator/(double d) const
{
	DMatrix<T> ret = *this;
	return ret /= d;
}


//*******************************************
//	Transpose and flat implementations
//*******************************************

template <class T>
inline
DMatrix<T>&	DMatrix<T>::transpose()
{
	int i,j;
	if(dim1()!=dim2())
	{
		DVector<T>*    ptn= new DVector<T>[dim2()];
		for (int s=0;s<dim2();s++) ptn[s].setDim(dim1());
		for(i=0; i<dim1(); i++)
			for(j=0;j<dim2();j++)
				ptn[j][i]=pt[i][j];
		delete [] pt;
		pt = ptn;
	}else
	{
		for(i=0; i<dim1(); i++)
			for(j=0;j<i;j++)
				swap(pt[j][i],pt[i][j]);
	}
	return (*this);
}

template <class T>
inline
DVector<T> DMatrix<T>::flat()
{
	DVector<T> r(dim1()*dim2());
	for(int k=0,i=0; i<dim1(); i++)
		for(int j=0; j<dim2(); j++)
			r[k++]=(*this)[i][j];
	return r;
}

//*********************************************************
//	Implementation of inverting using either lapack 
//  or local non optimal implementation.
//*********************************************************

#ifdef GM_INTEL
#include <mkl_lapack32.h>	// NBNBNB!!!   Remember also to link with mkl_c.lib
#endif

template <class T>
DMatrix<T>& DMatrix<T>::invert()
{
#ifdef GM_INTEL				// This requires MKL from Intel (mkl_c.lib), 
							// or BLAS and Lapack binaries in a compatible format
	if(dim1()==dim2() && dim1()>1)
	{
		int nk=dim1();
		Array<float> aa;
		aa.setSize(nk*nk);
		for(int i=0; i<nk; i++)	// daft cast
			for(int j=0; j<nk; j++) aa[i+nk*j] = (float) (*this)[i][j];	

		Array<float> work; work.setSize(nk*32);	// temporary work array
		int wsize=nk*32;						// work array size - should be optimized
		Array<int> ipiv; ipiv.setSize(nk);		// pivot table (result), size max(1,mmm,nnn)
		int info=0;								// error message, i=info>0 means that a[i][i]=0 ): singular, 
		int mmm=nk;								// dimentsion 
		int nnn=nk;								// dimentsion  i.e. a is mxn matrix
		int lda=nk;								// leading dimentsion
		// float *a								// input array

		// LAPACK When calling LAPACK routines from C-language programs, make sure that you follow Fortran rules: 
		// Pass variables by 'address' as opposed to pass by 'value'. Be sure to store your data Fortran-style, 
		// i.e. data stored column-major rather than row-major order 
		// Project neeeds mkl_c.lib and mkl_lapack32.h

		sgetrf(&mmm, &nnn, aa.ptr(), &lda, ipiv.ptr(), &info);				 // using Lapack LU-fact a is overwritten by LU
		sgetri(&mmm, aa.ptr(), &lda, ipiv.ptr(), work.ptr(), &wsize, &info); // a should now contain the inverse

		// if(info==0)
		for(int i=0; i<nk; i++)									
			for(int j=0; j<nk; j++) (*this)[i][j]=(T) aa[i+j*nk];
	}
#else							// gauss-jordan implementation from Numerical recipes
								// Ordinary LU-decomp used.
	DMatrix<T> a=(*this);
	Array<int> indx(a.dim2());

	const double TINY=1.0e-20;
	int i,imax,j,k;
	double big,dum,sum,temp;

	int n=a.dim1();  //nrows
	Array<double> vv(n);
	double d=1.0;
	for (i=0;i<n;i++) 
	{
		big=0.0;
		for (j=0;j<n;j++) if ((temp=fabs(a[i][j])) > big) big=temp;
		if (big == 0.0) return (*this);   //nrerror("Singular matrix in routine ludcmp");
		vv[i]=1.0/big;
	}
	for (j=0;j<n;j++) 
	{
		for (i=0;i<j;i++) 
		{
			sum=a[i][j];
			for (k=0;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=0.0;
		for (i=j;i<n;i++) 
		{
			sum=a[i][j];
			for (k=0;k<j;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ((dum=vv[i]*fabs(sum)) >= big) { big=dum; imax=i; }
		}
		if (j != imax) 
		{
			for (k=0;k<n;k++) 
			{
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			d = -d;
			vv[imax]=vv[j];
		}
		indx[j]=imax;
		if (a[j][j] == 0.0) a[j][j]=TINY;
		if (j != n-1) 
		{
			dum=1.0/(a[j][j]);
			for (i=j+1;i<n;i++) a[i][j] *= dum;
		}
	}									// LU-decomp. finished, stored in a[][]

	DVector<T> b(a.dim1());				// LU-back subst. begins 
	for(int cols=0; cols<a.dim2(); cols++)
	{
		int i,ii=0,ip,j;
		for(i=0; i<dim1(); i++) if(i==cols) b[i]=T(1.0); else b[i]=T(0.0);
		T sum;
		int n=a.dim1();
		for (i=0;i<n;i++) 
		{
			ip=indx[i];
			sum=b[ip];
			b[ip]=b[i];
			if (ii != 0)
				for (j=ii-1;j<i;j++) sum -= a[i][j]*b[j];
			else if (sum != 0.0) ii=i+1;
			b[i]=sum;
		}
		for (i=n-1;i>=0;i--) 
		{
			sum=b[i];
			for (j=i+1;j<n;j++) sum -= a[i][j]*b[j];
			b[i]=sum/a[i][i];
		}
		// LU-back subst. finished,
		for(i=0; i<dim1(); i++)		// inverse stored in this, a and b is disappearing?
			(*this)[i][cols]=b[i];
	}
#endif	
	return (*this);
}