//Test
//********************************************************
// GM_DMatrix.h: interface for the DMatrix class.
//********************************************************
#ifndef GM_DMATRIX_H_
#define GM_DMATRIX_H_

#include <GM_DVector.h>
#include <GM_Stream.h>
//*********************************************************
//****  GM_DMatrix.h: interface for the DMatrix class. ****
//	  -------------------------------------------------
//	DMatrix (Dynamic Matrix) is a container class which can
//	dynamic change dimention. together with DVector a lot  
//	of standard vector/matrix operations are available.
//  setDim:		Keep the previous contents intact if possible.
//				Set dim to i, but if the previous dim is greater than i then
//				the memory is not realocated and therefor not accessable later!!
//				This function is the speedy one.
//  resetDim:	Keep the previous contents intact if possible.
//				Realocating and cleaning up the memory
//				This function is safe but slow.
//  increaseDim:Increasing the dimention of the vector.
//				The previous contents is intact, but if "at_end=false"
//				then the previous contents is moved to the end of the vector.
//				The new element is initialized by "val".
//  clear:		All elements set to "v", (dimension not changed).

//	----------------------------------------------------
//				FOR FUTHER DEVELOPMENT !!
//	The invariants (stats) for DMatrix are:
//		if no allocation: pt  = a,
//		if allocation:	  pt != a.
//		NB! There might be more memory allocated than used,
//		but there is no knowledge about the real size!!
//		To clean up there is the function: resetDim(int d)
//*********************************************************

namespace GM_lib {

template <class T>
class DMatrix
{
	int			n;
	DVector<T>*	pt;
	DVector<T>	a[4];
   
	// Private help functions
	void copy(const DMatrix<T>& v);
	void copy(const T p[]);
   
 public:

	DMatrix(int i=0, int j=0);				// Default constructor
	DMatrix(int i, int j, T val);			// Standard initializing (all elements equal)
	DMatrix(int i, int j, const T p[]);		// Standard initializing from c-array
	DMatrix(const DMatrix<T>& v);			// Copy constructor

	~DMatrix() { if(pt != a) delete []pt; }	// NB! According to invariant

	int         dim1() const		{ return n;} 
	int         dim2() const		{ if (pt) return pt[0].dim(); return 0; }

	void		setDim(int i, int j);
	void		resetDim(int i, int j);
	void		increaseDim(int i, int j, T val=T(0), bool h_end=true, bool v_end=true);
	void		setIdentity();

	DMatrix<T>& operator=(const DMatrix<T>& v)	{ copy(v); return(*this); }
	DMatrix<T>& operator=(T p[])				{ copy(p); return(*this); }

	// The += -=  + - operators NB!!!! assume the dimensions to be equal
	// NB!! this operators are not doubling the dimensions!!
	//-----------------------------------------------------------------
	DMatrix<T>& operator+=(const DMatrix<T>&);			// (+=) On all elements.
	DMatrix<T>& operator-=(const DMatrix<T>&);			// (-=) On all elements.
	DMatrix<T>	operator+( const DMatrix<T>&) const;	// (+)  On all elements.
	DMatrix<T>	operator-( const DMatrix<T>&) const;	// (-)  On all elements.

	// The scalingoperator, there is also a non-member scaling operator with oposit order!!
	//-------------------------------------------------------------------------------------
	DMatrix<T>&	operator*=(double d);			// Scaling all elements on this matrix.
	DMatrix<T>	operator*(double d) const;		// Scaling all elements on the returned matrix.
	DMatrix<T>&	operator/=(double d);			// Scaling all elements on this matrix.
	DMatrix<T>	operator/(double d) const;		// Scaling all elements on the returned matrix.

	DVector<T>&		  operator[](int i);			// return given element.
	const DVector<T>& operator()(int i) const;		// return copy of given element.

	bool		operator<(const DMatrix<T>& m) const;

	DMatrix<T>&	transpose();
	DVector<T>	flat();

	T			determinant() const;
	DMatrix<T>& invert();
};


//***************************************************
//******   Scaling operators (opposit order)   ******
//***************************************************

template <class T>
inline
DMatrix<T> operator*(const double d, DMatrix<T> p) { return p*=d; }


//********************************************************
//******  Template iostream operators for DMatrix  ******
//********************************************************


template <class T_Stream, class T>
T_Stream& operator<<(T_Stream& out, const DMatrix<T>& v)
{ 
	out << v.dim1() << element_separator << v.dim2() << group_separator;
	for (int i=0; i<v.dim1(); i++)
	{
		for(int j=0;j<v.dim2(); j++) out << v(i)(j) << element_separator;
		out << group_separator;
	}
	return out;
}

template< class T_Stream, class T>
T_Stream& operator>>(T_Stream& in, DMatrix<T>& v)
{ 
	static Separator gs(group_separator);
	static Separator es(element_separator);
	int a,b;
	in >> a >> es >> b >> gs;
	v.setDim(a,b);
	for (int i=0; i<v.dim1(); i++)
	{
		for(int j=0;j<v.dim2(); j++) in >> v[i][j] >> es;
		in >> gs;
	}
	return in;
}


//***********************************************************
// The multiplication operators between DMatrix and DVectors
// and multiplication operators between DMatrix and DMatrix
//***********************************************************

template <class T, class G>
inline
const DVector<T>&  operator*(const DMatrix<G>& m, const DVector<T>& b)             
{
	static DVector<T> r; 

	if(m.dim2() != b.dim()) return r;
	r.setDim(m.dim1());
	int i,k;
	for(i=0;i<m.dim1();i++)
	{
		r[i] = m(i)(0)*b(0);
		for(k=1;k<b.dim();k++) 
			r[i] += m(i)(k)*b(k);
	}
	return r;
}

   
template <class T, class G>
inline
const DMatrix<T>&  operator*(const DMatrix<G>& m, const DMatrix<T>& b)             
{
	static DMatrix<T> r; 

	if(m.dim2() != b.dim1()) return r;
	r.setDim(m.dim1(),b.dim2());
	int i,j,k;
	for(i=0;i<m.dim1();i++)
		for(j=0;j<b.dim2();j++)
		{
			r[i][j] = m(i)(0)*b(0)(j);
			for(k=1;k<m.dim2();k++) 
				r[i][j] += m(i)(k)*b(k)(j);
		}
	return r;
}

template <class T, class G>
inline
const DVector<G>&  operator^(const DMatrix<G>& m, const DVector<T>& b)             
{
	static DVector<G> r; 

	if(m.dim2() != b.dim()) return r;
	r.setDim(m.dim1());
	int i,k;
	for(i=0;i<m.dim1();i++)
	{
		r[i] = m(i)(0)*b(0);
		for(k=1;k<b.dim();k++) 
			r[i] += m(i)(k)*b(k);
	}
	return r;
}

   
template <class T, class G>
inline
const DMatrix<G>&  operator^(const DMatrix<G>& m, const DMatrix<T>& b)             
{
	static DMatrix<G> r; 

	if(m.dim2() != b.dim1()) return r;
	r.setDim(m.dim1(),b.dim2());
	int i,j,k;
	for(i=0;i<m.dim1();i++)
		for(j=0;j<b.dim2();j++)
		{
			r[i][j] = m(i)(0)*b(0)(j);
			for(k=1;k<m.dim2();k++) 
				r[i][j] += m(i)(k)*b(k)(j);
		}
	return r;
}

	// NB!!	All non member template operator/function definition 
	//		must be inside the namespace
}	// end namespace GM_lib,  



#include "GM_DMatrix.c"



#endif /*GM_DMATRIX_H_*/
