#ifndef GM_DVECTOR_H_
#define GM_DVECTOR_H_

#include <GM_Stream.h>
#include <GM_Array.h>

//*********************************************************
//****  GM_DVector.h: interface for the DVector class. ****
//	  -------------------------------------------------
//	DVector (Dynamic Vector) is a container class which can
//	dynamic change dimention. together with DMatrix a lot  
//	of standard vector/matrix operations are available.
//  setDim:		Keep the previous contents intact if possible.
//				Set dim to i, but if the previous dim is greater than i then
//				the memory is not realocated and therefor not accessable later!!
//				This function is the speedy one.
//  resetDim:	Keep the previous contents intact if possible.
//				Realocating and cleaning up the memory
//				This function is safe but slow.
//  increaseDim:Increasing the dimention of the vector.
//				The previous contents is intact, but if "at_end=false"
//				then the previous contents is moved to the end of the vector.
//				The new element is initialized by "val".
//  clear:		All elements set to "v", (dimension not changed).

//	----------------------------------------------------
//				FOR FUTHER DEVELOPMENT !!
//	The invariants (stats) for DVector are:
//		if no allocation: pt  = a,
//		if allocation:	  pt != a.
//		NB! There might be more memory allocated than used,
//		but there is no knowledge about the real size!!
//		To clean up there is the function: resetDim(int d)
//*********************************************************

namespace GM_lib {

template <class T>
class DVector
{
	int   n;
	T*    pt;
	T	  a[4];

	// Private help functions
	void copy(const DVector<T>& v);
	void copy(const T p[]);

public:

	DVector(int i=0);				// Default constructor
	DVector(int i, T val);			// Standard initializing (all elements equal)
	DVector(int i, const T p[]);	// Standard initializing from c-array
	DVector(const DVector<T>& v);	// Copy constructor

	~DVector() { if(pt != a) delete []pt; }		// NB! According to invariant

	int         dim() const	{ return n;}		// Returning the vector dimention
		
	void		setDim(int i);
	void		resetDim(int i);
	void		increaseDim(int i, T val=T(0), bool at_end=true);
	void		clear(T v = T(0));

	DVector<T>& operator=(const DVector<T>& v)	{ copy(v); return *this; }
	DVector<T>& operator=(T p[])				{ copy(p); return *this; }

	// The += -= %= + - % operators NB!!!! assume the dimension to be equal
	// NB!! this operators are not doubling the dimension!!
	//-----------------------------------------------------------------
	DVector<T>& operator+=(const DVector<T>&);			// (+=) On all elements.
	DVector<T>& operator-=(const DVector<T>&);			// (-=) On all elements.
	DVector<T>	operator+( const DVector<T>&) const;	// (+)  On all elements.
	DVector<T>	operator-( const DVector<T>&) const;	// (-)  On all elements.


	// The scalingoperator, there is also a non-member scaling operator with oposit order!!
	//-------------------------------------------------------------------------------------
	DVector<T>&	operator*=(double d);			// Scaling all elements on this vector.
	DVector<T>	operator*(double d) const;		// Scaling all elements on the returned vector.
	DVector<T>&	operator/=(double d);			// Scaling all elements on this vector.
	DVector<T>	operator/(double d) const;		// Scaling all elements on the returned vector.

	T&          operator[](int i);				// return given element.
	const T&	operator()(int i) const;		// return copy of given element.

	void		putFirst(T val, int i=1);
	void		putFirst(const DVector<T>& v);
	void		append(T val, int i=1);
	void		append(const DVector<T>& v);

	T			length() const;
	bool		operator<(const DVector<T>& m) const;

	T&			sum() const;
	T&			sum(int start, int end) const;

	DVector<T>& getReversed() const;
	DVector<T>& getSubVector(int start, int end) const;
	Array<T>&	toArray() const;
};


//***************************************************
//******   Scaling operators (opposit order)   ******
//***************************************************

template <class T>
inline
DVector<T> operator*(const double d, DVector<T> p) { return p*=d; }


//********************************************************
//******  Template iostream operators for DVectors  ******
//********************************************************

template <class T_Stream, class T>
T_Stream& operator<<(T_Stream& out, const DVector<T>& v)
{ 
	out << v.dim() << group_separator;
	for (int i=0; i<v.dim(); i++)
		out << v(i) << element_separator;
	return out;
}

template<class T_Stream, class T>
T_Stream& operator>>(T_Stream& in, DVector<T>& v)
{ 
	static Separator gs(group_separator);
	static Separator es(element_separator);
	int a;
	in >> a >> gs;
	v.setDim(a);
	for (int i=0; i<v.dim(); i++)
		in >> v[i] >> es;
	return in;
}


//********************************************************
// The multiplication operators between two DVectors
//********************************************************

template<class T, class K>
inline
T operator*(const DVector<K>& m, const DVector<T>& v)
{
#ifdef DEBUG
	if (m.dim() != v.dim()) 
		cerr << "Dimension error, dim1=" << m.dim() << " ,dim2=" << v.dim() << endl;
#endif
	T r(0.0); 
	for(int i=0;i<m.dim();i++) r += m(i)*v(i);
	return r; 
}

template<class T, class K>
inline
K operator^(const DVector<K>& m, const DVector<T>& v)
{  
#ifdef DEBUG
	if (m.dim() != v.dim()) 
		cerr << "Dimension error, dim1=" << m.dim() << " ,dim2=" << v.dim() << endl;
#endif
	K r(0); 
	for(int i=0;i<m.dim();i++) r += m(i)*v(i);
	return r; 
}

	// NB!!	All non member template operator/function definition 
	//		must be inside the namespace
}	// end namespace GM_lib,  




//****************************************************************
// The rest of the file should be in:  GM_DVector.c
//******************************************************************
// #include <GM_DVector.h>

using namespace GM_lib; 


//*********************************************************
//****************    	Constructors       ****************
//*********************************************************


template<class T>
inline
DVector<T>::DVector(int i)				// Default constructor
{
	pt = (i>4 ? new T[i]:a); n=i;
}

template<class T>
inline
DVector<T>::DVector(int i, T val)		// Standard initializing (all elements equal)
{
	pt = (i>4 ? new T[i]:a); n=i;
	for(int j=0; j<n; j++) pt[j] = val;
}

template<class T>
inline
DVector<T>::DVector(int i, const T p[])	// Standard initializing from c-array
{
	pt = (i>4 ? new T[i]:a); n=i;
	copy(p);
}

template<class T>
inline
DVector<T>::DVector(const DVector<T>& v)// Copy constructor
{
	n=0; pt=a;
	copy(v);
}


//*********************************************************
//	Private help functions implementations
//*********************************************************

template <class T>
inline
void  DVector<T>::copy(const DVector<T>& v)
{
	if(v.n>4 && v.n>n)
	{
		if (pt != a) delete [] pt;
		pt = new T[v.n];
	}
	n = v.n;
	for(int i=0; i<n; i++) pt[i] = v.pt[i]; 
}

template <class T>
inline
void  DVector<T>::copy(const T p[])		
{
	for(int i=0; i<n; i++) pt[i] = p[i];
}

//*********************************************************
//	Member operators and functions implementations
//*********************************************************

template <class T>
inline
void  DVector<T>::setDim(int i)
{ 
	if(i>4 && i>n)
	{
		if(pt != a) delete [] pt;
		pt = new T[i];
	}
	n = i;
}

template <class T>
inline
void  DVector<T>::resetDim(int i)
{
	int j, k = min(i,n);
	T* tmp = (i>4 ? new T[i]:a);
	for(j=0; j<k; j++) tmp[j] = pt[j];
	for(; j<i; j++) tmp[j] = T(0);
	if(pt != a) delete [] pt;
	n = i;
	pt = tmp;
}

template <class T>
inline
void  DVector<T>::increaseDim(int i, T val, bool at_end)
{ 
	if(i>0)
	{
		int j = n+i;
		int k = (at_end ? n:0);
		int r = (at_end ? 0:i);
		if(j>4)
		{
			T* tmp = new T[j];
			for(j=r; j<r+n; j++) tmp[j] = pt[j-r];
			if(pt != a) delete [] pt;
			pt = tmp;
		}
		else if(!at_end)
			for(j=n-1; j>=0; j--) pt[j+i] = pt[j];
		for(j=k; j<i+k; j++) pt[j] = val;
		n += i;
	}
}

template <class T>
inline
void DVector<T>::clear(T v)
{
	for(int i=0; i<n; i++) pt[i] = v;
}

//**********************************************
//	[] and () and < operators  implementations
//  and length() function  implementations.
//**********************************************

template <class T>
inline
T&  DVector<T>::operator[](int i)
{
#ifdef DEBUG
	if (i<0 || i>=n) cerr << "Error index " << i << " is outside(0," << n << ")\n";
#endif
	return pt[i];
}

template <class T>
inline
const T&  DVector<T>::operator()(int i) const
{
#ifdef DEBUG
	if (i<0 || i>=n) cerr << "Error index " << i << " is outside(0," << n << ")\n";
#endif
	return pt[i];
}

template <class T>
inline
T DVector<T>::length() const
{
	T r=T(0);
	for(int i=0; i<n; i++) r += (*this)(i)*(*this)(i);
	return sqrt(r);
}

template <class T>
inline
bool DVector<T>::operator<(const DVector<T>& v) const
{
	return length() < v.length();
}

//*******************************************
//	+= -= + -  operators  implementations
//*******************************************

template<class T>
inline
DVector<T>& DVector<T>::operator+=(const DVector<T>& v)
{
#ifdef DEBUG
	if (n != v.dim()) 
		cerr << "Vector dimension error, dim=" << n << " ,dim=" << v.dim() << endl;
#endif
	for (int i=0; i <n; i++) pt[i] += v.pt[i];
	return *this;
}

template<class T>
inline
DVector<T>& DVector<T>::operator-=(const DVector<T>& v)
{
#ifdef DEBUG
	if (n != v.dim()) 
		cerr << "Vector dimension error, dim=" << n << " ,dim=" << v.dim() << endl;
#endif
	for (int i=0; i <n; i++) pt[i] -= v.pt[i];
	return *this;
}

template<class T>
inline
DVector<T>	DVector<T>::operator+(const DVector<T>& a) const
{ 
	DVector<T> na=(*this);
	return na += a;
}

template<class T>
inline
DVector<T>	DVector<T>::operator-(const DVector<T>& a) const
{
	DVector<T> na=(*this);
	return na -= a;
}


//*******************************************
//	Scaling  operators  implementations
//*******************************************

template <class T>
inline
DVector<T>& DVector<T>::operator*=(double d)
{  
	for(int i=0; i<n; i++) pt[i] *= d;
	return *this;
}

template <class T>
inline
DVector<T>& DVector<T>::operator/=(double d)
{  
	for(int i=0; i<n; i++) pt[i] /= d;
	return *this;
}

template <class T>
inline
DVector<T> DVector<T>::operator*(double d) const
{
	DVector<T> ret = *this;
	return ret *= d;
}

template <class T>
inline
DVector<T> DVector<T>::operator/(double d) const
{
	DVector<T> ret = *this;
	return ret /= d;
}

template <class T>
inline
DVector<T>& DVector<T>::getReversed() const
{
	static DVector<T> ret;
	ret.setDim(n);
	for(int i=0; i<n; i++)
		ret[i] = pt[n-1-i];
	return ret;
}

template <class T>
inline
DVector<T>& DVector<T>::getSubVector(int start, int end) const
{
	static DVector<T> ret;
	if(start < 0)	start = 0;
	if(end > n)		end = n;
	if(start < end)
	{
		ret.setDim(end-start);
		int u=0;
		for(int i=start; i<end; i++)
			ret[u++] = pt[i];
	}
	else
		ret.setDim(0);
	return ret;
}

template <class T>
inline
T& DVector<T>::sum() const
{
	static T ret;
	ret = T(0);
	for(int i=0; i<n; i++) ret += pt[i];
	return ret;
}

template <class T>
inline
T& DVector<T>::sum(int start, int end) const
{
	static T ret;
	if(start < 0)	start = 0;
	if(end   > n)	end = n;
	ret = T(0);
	for(int i=start; i<end; i++) ret += pt[i];
	return ret;
}

template <class T>
inline
void DVector<T>::putFirst(T val, int i)
{
	if(i < 1) return;
	increaseDim(i,val,false);
}

template <class T>
inline
void DVector<T>::append(T val, int i)
{
	if(i < 1) return;
	increaseDim(i, val);
}

template <class T>
inline
void DVector<T>::putFirst(const DVector<T>& v)
{
	if(v.n>0)
	{
		int j = n+v.n;
		if(j>4)
		{
			T* tmp = new T[j];
			for(j=0; j<n; j++) tmp[j+v.n] = pt[j];
			if(pt != a) delete [] pt;
			pt = tmp;
		}
		else
			for(j=n-1; j>=0; j--) pt[j+v.n] = pt[j];
		for(j=0; j<v.n; j++) pt[j] = v(j);
		n += v.n;
	}
}

template <class T>
inline
void DVector<T>::append(const DVector<T>& v)
{
	if(v.n > 0)
	{
		int j = n + v.n;
		if(j>4)
		{
			T* tmp = new T[j];
			for(j=0; j<n; j++) tmp[j] = pt[j];
			if(pt != a) delete [] pt;
			pt = tmp;
		}
		for(j=0; j<v.n; j++) pt[j+n] = v(j);
		n += v.n;
	}
}

template <class T>
inline
Array<T>&	DVector<T>::toArray() const
{
	static Array<T> a;
	a.setSize(dim());
	for(int i=0; i<n; i++) a[i] = (*this)(i);
	return a;
}
#endif /*GM_DVECTOR_H_*/
