
//////////////////////////////////////////////////////////////////////
// DisplayObject.cpp: implementation of the DisplayObject class.
//////////////////////////////////////////////////////////////////////


#include <GM_DisplayObject.h>
#include <GM_Light.h>
using namespace GM_lib;


GLuint DisplayObject::_free_name=1;


void DisplayObject::setSurroundingSphere(const Sphere<GLfloat,3>& b)
{
	_sphere = b;
}

void DisplayObject::updateSurroundingSphere(const Point<GLfloat,3>& p)
{
	_sphere += p;
}

void DisplayObject::insert(DisplayObject* obj)
{
	if(obj)  _object.insert(obj);
}


void DisplayObject::remove(DisplayObject* obj)
{
	if(obj)
		if(!_object.remove(obj))
			for(int i=0; i< _object.size(); i++)
				_object[i]->remove(obj);
}


DisplayObject::DisplayObject(const Vector<GLfloat,3>& trans_vector, 
							 const Point<GLfloat,3>& scale_factor, 
							 const Vector<GLfloat,3>& rot_axel, Angle a):_scale(scale_factor)
{
	_matrix.translate(trans_vector);
	_matrix.rotate(a,rot_axel);
	_name	= _free_name++;
	_aktiv	= false;
	_local_cs = true;
	_type_id  = 1;
	_visible = true;
	_collapsed = false;
	_default	= Blue; 
	_marked		= _default.inverse();
	_ir = 0.07;
}


DisplayObject::DisplayObject(const DisplayObject&  d)
{
	_matrix	= d._matrix;
	_object	= d._object;
	_sphere	= d._sphere;
	_scale  = d._scale;
	_name	= _free_name++;
	_aktiv	= false;
	_local_cs = d._local_cs;
	_type_id  = 1;
	_visible  = d._visible;
	_collapsed = d._collapsed;
	_default  = d._default;
	_marked  = d._marked;
	_ir = d._ir;
}


DisplayObject::~DisplayObject()
{
	for(int i=0; i < _object.size(); i++)
		delete _object[i];
}


int DisplayObject::prepare(Array<Light*>& obj, Array<GL_Matrix>& mat, Scene* s, DisplayObject* mother)
{
	int nr = 1;
	_scene  = s;
	_mother = mother;
	Light * pl  = dynamic_cast<Light *>(this);
	if(pl) obj += pl;

	mat.push();
	_LAPrepare(mat.last());
	mat.last() *= matrix();
	_present = mat.last();
	_global_total_sphere = _global_sphere = _present*_sphere;
	if(_scale.isActiv())
	{
		_global_sphere *= _scale.maxScale();
		_global_sphere %= _scale.getScale();
		_global_total_sphere *= _scale.maxScale();
		_global_total_sphere %= _scale.getScale();
	}
	for(int i=0; i< _object.size(); i++)
	{
		nr += _object[i]->prepare(obj,mat,s,this);
		_global_total_sphere += _object[i]->surroundingSphere();
	}
	mat.pop();
	return nr;
}

Sphere<GLfloat,3>	DisplayObject::cleanSurroundingSphere() const
{
	Sphere<GLfloat,3> sp;

	for(int i=0; i< _object.size(); i++)
		sp += _object(i)->cleanSurroundingSphere();
	if(_type_id!=2 && _type_id!=3) sp += _global_sphere;
	return sp;
}

void DisplayObject::simulate(double dt)
{
	glPushMatrix();
	localSimulate(dt);
	glPopMatrix();
	for(int i=0; i< _object.size(); i++)
		_object[i]->simulate(dt);
}
	
void DisplayObject::_fillObj(Array<DisplayObject*>& obj)
{
	if(_sphere.valid()) 
		obj += this;
	for(int i=0; i< _object.size(); i++)
		_object[i]->_fillObj(obj);
}


void DisplayObject::culling(Array<DisplayObject*>& obj, const Frustum& f)
{
	if(!_visible) return;
	int i, k = f.isInterfering(surroundingSphere());

	if(k < 0) return;	// Outside
	if(k > 0)			// Inside
	{
		if(_sphere.valid()) obj += this;
		for(i=0; i< _object.size(); i++)
			_object[i]->_fillObj(obj);
	}
	else // if(k == 0)     Intersecting
	{
		if(f.isInterfering(_global_sphere) >= 0) obj += this;
		for(i=0; i< _object.size(); i++)
			_object[i]->culling(obj, f);
	}
}

DisplayObject* DisplayObject::find(GLuint name)
{
	DisplayObject* d;

	if(name == _name)	return this;
	for(int i=0; i < _object.size(); i++)
		if(d = _object(i)->find(name)) return d;
	return 0;
}


void DisplayObject::rotate(Angle a, const Vector<GLfloat,3>& rot_axel)
{
	_matrix.rotate(a,rot_axel);
}

void DisplayObject::rotateGlobal(Angle a, const Vector<GLfloat,3>& rot_axel)
{
	_matrix.rotateGlobal(a,rot_axel);
}

void DisplayObject::rotate(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& rot_axel)
{
	_matrix.rotate(a,p,rot_axel);
}

void DisplayObject::rotateGlobal(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& rot_axel)
{
	_matrix.rotateGlobal(a,p,rot_axel);
}

void DisplayObject::translate(const Vector<GLfloat,3>& trans_vector)
{
	_matrix.translate(trans_vector);
}

void DisplayObject::translateGlobal(const Vector<GLfloat,3>& trans_vector)
{
	_matrix.translateGlobal(trans_vector);
}


void DisplayObject::scale(const Point<GLfloat,3>& scale_factor)
{	
	for(int i=0; i<_object.size(); i++) 
	{
		Point3D<GLfloat> tr = -(_object[i]->_matrix*Point<GLfloat,3>(0.0f));
		_object[i]->translate(scale_factor%tr);
		_object[i]->scale(scale_factor); // BB
	}
	_scale.scale(scale_factor);
	_sphere=_scale.sphere(_sphere);
}


void DisplayObject::imposter() 
{
		glBegin(GL_QUAD_STRIP);
		glVertex3f( _ir,-_ir,-_ir ); glVertex3f( _ir, _ir,-_ir );
		glVertex3f( _ir,-_ir, _ir ); glVertex3f( _ir, _ir, _ir );
		glVertex3f(-_ir,-_ir, _ir ); glVertex3f(-_ir, _ir, _ir );
		glVertex3f(-_ir,-_ir,-_ir ); glVertex3f(-_ir, _ir,-_ir );
		glVertex3f( _ir,-_ir,-_ir ); glVertex3f( _ir, _ir,-_ir );
		glEnd();
		glBegin(GL_QUADS);
		glVertex3f(-_ir,-_ir,-_ir ); glVertex3f( _ir,-_ir,-_ir );
		glVertex3f( _ir,-_ir, _ir ); glVertex3f(-_ir,-_ir, _ir );
		glVertex3f(-_ir, _ir,-_ir ); glVertex3f(-_ir, _ir, _ir );
		glVertex3f( _ir, _ir, _ir ); glVertex3f( _ir, _ir,-_ir );
		glEnd();
}

