
//********************************************************
// DisplayObject.h: interface for the DisplayObject class.
//********************************************************
#ifndef _DISPLAYOBJECT_H__
#define _DISPLAYOBJECT_H__

#include <GM_OpenGL.h>
#include <GM_Material.h>


namespace GM_lib {

class Camera;
class Scene;
class Light;

//*********************************************************************
//	Inneholder:
//		privat;
//				Et navn for identifikasjon i select funksjonaliteten.
//				En array av pekere til alle sub-objektene til dette objektet
//		protected:
//				To matriser for lokal transformasjon av objektet.
//				En Omsluttende kule til objektet som m lages.
//				Det er to virtuelle funksjoner localDisplay og localSelect
//					som alle avledede klasser m ha sin versjon av.
//				Arbeidsdelingen er slik at display og select foretar 
//					transformasjonen lagret i matrix, og displayer alle 
//					sub-objektene til objektet. Men etter transformasjonen
//					og fr sub-objektene displayes kalles localDisplay
//					(eventuelt localSelect). I localDisplay kan en s 
//					displaye det som ikke displayes i sub-objektene og/eller
//					foreta transformasjoner av sub-objekter for  
//					 simulere bevegelser.
//				Det er ogs en insertSubObject-funksjon som avledede klasser
//				    kan bruke for  lage sub-objekt. Hvis en avledet klasse 
//					trenger tilgang til et objekt for bevegelsessimulering
//					br den lage en egen peker for det.
//		public:
//				Konstuktr med mulige default verdier og sletter
//				En display funksjon for displaying av objektet.
//				En select funksjon for plukking(select) av objektet med mus.
//				En simulate funksjon for at objektet kan bevege subobjekter.
//				En find funksjon for  finne en funksjon med et gitt navn.
//				En rotate, scale og translate funksjon for  
//				   posisjonere objektet relativt i forhold til morsobjektet
//*********************************************************************


class DisplayObject
{
	static GLuint			_free_name;	// For automatisk name-generations.
	DisplayObject*			_mother;	// the mother in the hierarchy (tree).
	GLuint					_name;		// Unic name for this object, used for selecting
	Sphere<GLfloat,3>		_sphere;	// Surrounding sphere for this object

	void _fillObj(Array<DisplayObject*>&);

protected:	
	int						_type_id;

	friend class Scene;					// Funksjoner som bare brukes av en scene
	friend class Camera;				// Funksjoner som bare brukes av et kamera

	Array<DisplayObject*>	_object;

	Scene*					_scene;		// Den aktuelle scenen til displayhierarkiet

	GL_Matrix				_matrix;	// The difference matrix from mother to this.	
	GL_Matrix				_present;	// The difference matrix from global to this.
	ScaleObj				_scale;		// The scaling for this and the children.

	bool					_local_cs;	// Using local coordinate system, default is true
	bool					_aktiv;		// This variable is only for camera.

	Sphere<GLfloat,3>		_global_sphere;			// for this object
	Sphere<GLfloat,3>		_global_total_sphere;	// included all children
	float					_ir;					// Size of cube (collapsed object)

	GL_Color				_default;
	GL_Color				_marked;
	bool					_selected;	// BB
	bool					_visible;	// culling on invisible items
	bool					_collapsed;	// represented by a small cube
	
	void	setSurroundingSphere(const Sphere<GLfloat,3>& b);
	void    updateSurroundingSphere(const Point<GLfloat,3>& p);
	const	GL_Matrix& motherGlobalMatrix() const { return _mother->globalMatrix();}
	void	reset()								  { _matrix.reset();}

	virtual void localDisplay() {}
	virtual void localSelect()  {}
	virtual void localSimulate(double dt){}
	virtual void culling(Array<DisplayObject*>&, const Frustum&);

	virtual void imposter();					// Display of collapsed object

	
private:	
	virtual GL_Matrix& matrix()	{ return _matrix;}	// Made specially for Cameras
	virtual void _LAPrepare(const GL_Matrix& m) {}	// Made specially for LA_DisplayObject's

	int		prepare(Array<Light*>& obj, Array<GL_Matrix>& mat, Scene* s, DisplayObject* mother=0);
	void	simulate(double dt);
	void	display();
	void	select(int what=-1);
	

public:											// default og standard constructor
												// Vil frst rotere s skalere og
												// tilslutt translatere/flytte objektet
	DisplayObject(const Vector<GLfloat,3>& trans  = Vector3D<GLfloat>(0,0,0),
				  const Point<GLfloat,3>&  scale   = Point3D<GLfloat>(1,1,1),
				  const Vector<GLfloat,3>& rotate = Vector3D<GLfloat>(1,0,0), Angle a=0);

	DisplayObject(const DisplayObject&  d);		// Copy constructor

	virtual ~DisplayObject();

	const GL_Matrix&	globalMatrix() const		{ return  _present; }
	Sphere<GLfloat,3>	surroundingSphere() const	{ return  _global_total_sphere;}
	Sphere<GLfloat,3>	cleanSurroundingSphere() const;
	Point<GLfloat,3>	centerPos() const			{ return  surroundingSphere().pos(); }

	DisplayObject*		find(GLuint name);

	void insert(DisplayObject* obj);
	void remove(DisplayObject* obj);

	virtual void rotate(Angle a, const Vector<GLfloat,3>& rot_axel);
	virtual void rotate(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& d);
	virtual void translate(const Vector<GLfloat,3>& trans_vector);

	virtual void rotateGlobal(Angle a, const Vector<GLfloat,3>& rot_axel);
	virtual void rotateGlobal(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& d);
	virtual void translateGlobal(const Vector<GLfloat,3>& trans_vector);

	virtual void scale(const Point<GLfloat,3>& scale_factor);	// Only geometry

	virtual const char* identity() const { return " NOTHING";}
	GLuint	getName() const				 { return _name; }	
	
	bool flipSelected()		 { _selected = !_selected; if(_selected) edit(); return _selected; }
	void setSelected(bool s) { _selected=s; }
	bool getSelected()		 { return _selected; }
	
	virtual void edit(int selector_id)	 {}
	virtual void edit(DisplayObject* lp) {}
	virtual void edit()					 {}
	virtual void selectEvent(int selector_id) {}
	virtual void toggleSelector()		{ _collapsed=!_collapsed; }
	virtual void setCollapsed(bool c)	{ _collapsed=c; }
	virtual void setVisible(bool v)		{ _visible=v; }
	virtual bool isVisible()			{ return _visible; }
	virtual void childrenOff() {for(int i=0;i<_object.size();i++) _object[i]->setVisible(false); }
	virtual void childrenOn()  {for(int i=0;i<_object.size();i++) _object[i]->setVisible(true); }

	virtual void editPos(Vector<float,3> delta){}
	int typeId() { return _type_id; }
	virtual void setMaterial(const Material& m){}

	Array<DisplayObject*>& children(){ return _object;}

	template <class T_Stream>
		friend T_Stream& operator<<(T_Stream& out, DisplayObject& s)
	{
		out << s._type_id << object_separator;
		out << s._matrix << object_separator
		    << s._scale  << object_separator;
/*		if(st)
		{
			out << _object.size() << object_separator;
			for(int i=0; i<_object.size(); i++)
				if(st>1 || (_object[i]->typeId() != 2))
					out << *(_object[i]);
		}*/
		return out;
	}
	template <class T_Stream>
		friend T_Stream& operator>>(T_Stream& in, DisplayObject& s)
	{
		static Separator os(object_separator);
		in >> s._type_id >> os;
		in >> s._matrix >> os >> s._scale >> os;
/*		if(st)
		{
			int nr;
			in >> nr >> os;
			for(int i=0; i<nr; i++)
					_object += newDisplayObject(in);
		}*/		return in;
	}
	
	template <class T_Stream>
	DisplayObject(T_Stream& in, int st) 
	{
		in >> *this;
		_name	= _free_name++;
		_aktiv	= false;
		_local_cs = true;
		_visible = true;
		_collapsed = false;
		_default	= Blue; 
		_marked		= _default.inverse();
		_ir = 0.07;
	}
};


inline
void DisplayObject::display()
{
	if(!_aktiv)	
	{
/*		if(_type_id != 11)
		{
			glPushMatrix();
			Sphere3D sp(_global_sphere,10,10);
			sp.display();
			glPopMatrix();
		}
*/
		glPushMatrix();
			if(_local_cs) _present.mult();
			_scale.glScaling();
			if(_collapsed)
			{
				GLboolean lg;
				glGetBooleanv(GL_LIGHTING,&lg);
				if(lg) glDisable(GL_LIGHTING);
				if(_selected)	_marked.glSet(); 
				else			_default.glSet();
				imposter(); 
				if(lg) glEnable(GL_LIGHTING);
			}
			else localDisplay();
		glPopMatrix();
	}
}

inline
void DisplayObject::select(int what)
{
		if(!_aktiv && (what==0 || what==_type_id || (what<0 && what+_type_id != 0)))
		{
			glPushMatrix();
			if(_local_cs) _present.mult();
			_scale.glScaling();
			GL_Color name(getName());
			name.glSet();
			if(_collapsed) imposter(); 
			else localSelect();
			glPopMatrix();
		}
}


} // end namespace GM_lib

#endif // _DISPLAYOBJECT_H__

