//////////////////////////////////////////////////////////////////////
// LA_DisplayObject.cpp: implementation of the LA_DisplayObject class.
//////////////////////////////////////////////////////////////////////

#include <GM_LA_DisplayObject.h>
using namespace GM_lib;



//////////////////////////////////////////////////////////////////////
// Construction/Destruction
//////////////////////////////////////////////////////////////////////


LA_DisplayObject::LA_DisplayObject(	const Point<GLfloat,3>&  pos,
									const Vector<GLfloat,3>& dir,
									const Vector<GLfloat,3>& up): DisplayObject(),_mat(),_mat_inv()
{
	set(pos,dir,up);
}
					 
LA_DisplayObject::LA_DisplayObject( const Point<GLfloat,3>&  look_at_pos,
									const Point<GLfloat,3>&  pos,
									const Vector<GLfloat,3>& up): DisplayObject(),_mat(),_mat_inv()
{
	Vector<GLfloat,3> dir = up.getLinIndVec();
	set(pos,dir,up);
	_keep_lock_at	= true;
	_lock_at_pos    = look_at_pos;
}

LA_DisplayObject::LA_DisplayObject( DisplayObject*			 lock_at_object,
									const Point<GLfloat,3>&  pos,
									const Vector<GLfloat,3>& up): DisplayObject(),_mat(),_mat_inv()
{
	Vector<GLfloat,3> dir = up.getLinIndVec();
	set(pos,dir,up);
	_keep_lock_at	= true;
	_lock_at_object	= lock_at_object;
}


LA_DisplayObject::LA_DisplayObject( const LA_DisplayObject& s): DisplayObject(s),_mat(s._mat),_mat_inv(s._mat_inv)
{
	set(s._pos,s._dir,s._up);
	_keep_lock_at	= s._keep_lock_at;
	_lock_at_object	= s._lock_at_object;
	_lock_at_pos	= s._lock_at_pos;
}


LA_DisplayObject::~LA_DisplayObject()
{
}



void LA_DisplayObject::set(const Point<GLfloat,3>&  p, const Vector<GLfloat,3>& d, const Vector<GLfloat,3>& u)
{
	_pos	= p;
	_dir	= d;
	_up		= u - (u*_dir)*_dir;
	_side	= _up^_dir;
	_keep_lock_at	= false;
	_lock_at_object	= 0;

	basisChange(_side, _up, _dir, _pos);
}


void LA_DisplayObject::lockAtTarget(DisplayObject* obj)
{ 
	if(obj)
	{
		_updateOrientation(obj->centerPos());
		basisChange(_side, _up, _dir, _pos);

		_keep_lock_at	= true;
		_lock_at_object = obj;		
	}
}

void LA_DisplayObject::lockAtTarget(const Point<GLfloat,3>& p)
{ 
	_lock_at_pos	= p;
	_keep_lock_at	= true; 
	_lock_at_object = 0;
}

void LA_DisplayObject::lockAtTarget(double d)
{ 
	_lock_at_pos	= _mat*(_pos + d*_dir);
	_keep_lock_at	= true; 
	_lock_at_object = 0;
}

void LA_DisplayObject::unLockTarget()
{ 
	_keep_lock_at	= false;
	_lock_at_object = 0;
}

void LA_DisplayObject::rotate(Angle a, const Vector<GLfloat,3>& rot_axel)
{
	if(_keep_lock_at)
	{
		Point3D<GLfloat> lock_at_p = _keepLockAt();
		double dir_length = (lock_at_p - _pos).length();
		Vector3D<GLfloat> t,ra=rot_axel;
		double dd= a.rad()/M_PI_4;
		while(dd > 1.0)
		{
			t = ra^_dir;
			t.setLength(dir_length);		
			move(Vector2D<GLfloat>(t*_side,t*_up));
			dd -= 1.0;
		}
		t = ra^_dir;
		t.setLength(tan(dd*M_PI_4)*dir_length);
		move(Vector2D<GLfloat>(t*_side,t*_up));
	}
	else
	{
		GL_Matrix m(a,rot_axel);
		_up   *= m;
		_dir  *= m;
		_side *= m;
	}
	basisChange(_side, _up, _dir, _pos);
}

void LA_DisplayObject::rotate(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& d)
{
	GL_Matrix m(a,p,d);
	_pos  *= m;
	_up   *= m;
	_dir  *= m;
	_side *= m;

	if(_keep_lock_at)
	{
		Point3D<GLfloat> lock_at_p = _keepLockAt();
		_updateOrientation(lock_at_p);
	}

	basisChange(_side, _up, _dir, _pos);
}

void LA_DisplayObject::translate(const Vector<GLfloat,3>& trans_vector)
{
	move(trans_vector);
}

void LA_DisplayObject::rotateGlobal(Angle a, const Vector<GLfloat,3>& rot_axel)
{
	rotate(a,_mat_inv*rot_axel);
}

void LA_DisplayObject::rotateGlobal(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& d)
{
	rotate(a,_mat_inv*p,_mat_inv*d);
}


void LA_DisplayObject::translateGlobal(const Vector<GLfloat,3>& trans_vector)
{
	move(_mat_inv*trans_vector);
}

void LA_DisplayObject::turn(Angle a)
{
	if(!_keep_lock_at)
	{
		GL_Matrix m(a,_up);
		_dir  *= m;
		_side *= m;
		basisChange(_side, _up, _dir, _pos);
	}
}

void LA_DisplayObject::roll(Angle a)
{
	GL_Matrix m(a,_dir);
	_up   *= m;
	_side *= m;
	basisChange(_side, _up, _dir, _pos);
}

void LA_DisplayObject::tilt(Angle a)
{
	if(!_keep_lock_at)
	{
		GL_Matrix m(a,_side);
		_up   *= m;
		_dir  *= m;
		basisChange(_side, _up, _dir, _pos);
	}
}

void LA_DisplayObject::move(GLfloat d)
{
	_pos += d*_dir;
	basisChange(_side, _up, _dir, _pos);
}

void LA_DisplayObject::move(const Vector<GLfloat,3>& t)
{
	Point3D<GLfloat> lock_at_p;

	if(_keep_lock_at)
		lock_at_p = _keepLockAt();

	_pos += t;

	if(_keep_lock_at)
		_updateOrientation(lock_at_p);
	basisChange(_side, _up, _dir, _pos);
}

void LA_DisplayObject::move(char c, double d)
{
	double dir_length;
	Point3D<GLfloat> lock_at_p;

	if(_keep_lock_at)
	{
		lock_at_p  = _keepLockAt();
		dir_length = (lock_at_p - _pos).length();
	}

	switch (c) {
	case 'u':
		_pos += d*_up;
		break;
	case 'd':
		_pos -= d*_up;
		break;
	case 'l':
		_pos += d*_side;
		break;
	case 'r':
		_pos -= d*_side;
		break;
	default:
		break;
	}

	if(_keep_lock_at)
	{
		_updateOrientation(lock_at_p);
		_pos	= lock_at_p - dir_length*_dir;
	}

	basisChange(_side, _up, _dir, _pos);
}

void LA_DisplayObject::move(const Vector<GLfloat,2>& t)
{
	double dir_length;
	Point3D<GLfloat> lock_at_p;

	if(_keep_lock_at)
	{
		lock_at_p  = _keepLockAt();
		dir_length = (lock_at_p - _pos).length();
	}

	_pos += t(0)*_side + t(1)*_up;

	if(_keep_lock_at)
	{
		_updateOrientation(lock_at_p);
		_pos	= lock_at_p - dir_length*_dir;
	}

	basisChange(_side, _up, _dir, _pos);
}


void LA_DisplayObject::localSimulate(double)
{
	if(_keep_lock_at)
	{
		_updateOrientation(_keepLockAt());
		basisChange(_side, _up, _dir, _pos);
	}
}


void LA_DisplayObject::_LAPrepare(const GL_Matrix& m)
{	
	_mat = m;
	_mat_inv = m.inverse2();
}
