//**************************************************************
// LA_DisplayObject.h: interface for the LA_DisplayObject class.
//**************************************************************
#ifndef _LA_DISPLAYOBJECT_H__
#define _LA_DISPLAYOBJECT_H__

#include <GM_DisplayObject.h>

namespace GM_lib {


class LA_DisplayObject : public DisplayObject  
{
protected:
	// Matrises from Scene to this
	GL_Matrix				_mat;
	GL_Matrix				_mat_inv;

	Point3D<GLfloat>		_pos;
	UnitVector3D<GLfloat>	_dir;
	UnitVector3D<GLfloat>	_side;
	UnitVector3D<GLfloat>	_up;

//private:

	Point3D<GLfloat>		_lock_at_pos;
	DisplayObject*			_lock_at_object;
	bool					_keep_lock_at;

	virtual
	void basisChange(const Vector<GLfloat,3>& x, const Vector<GLfloat,3>& y,
					 const Vector<GLfloat,3>& z, const Vector<GLfloat,3>& p)
													{ _matrix.basisChange(z,x,y,p);}

	Point<GLfloat,3>	_keepLockAt();
	void				_updateOrientation(const Point<GLfloat,3>& lock_at_p);

	// Virtual functions from DisplayObject
	void			localSimulate(double dt);
	void			_LAPrepare(const GL_Matrix& m);

public:
	LA_DisplayObject(const Point<GLfloat,3>&  pos = Point3D<GLfloat>(0,0,0),
					 const Vector<GLfloat,3>& dir = Vector3D<GLfloat>(1,0,0),
					 const Vector<GLfloat,3>& up = Vector3D<GLfloat>(0,0,1));

	LA_DisplayObject(const Point<GLfloat,3>&  look_at_pos,
					 const Point<GLfloat,3>&  pos = Point3D<GLfloat>(0,0,0),
					 const Vector<GLfloat,3>& up = Vector3D<GLfloat>(0,0,1));

	LA_DisplayObject(DisplayObject*			  lock_at_object,
					 const Point<GLfloat,3>&  pos = Point3D<GLfloat>(0,0,0),
					 const Vector<GLfloat,3>& up = Vector3D<GLfloat>(0,0,1));

	LA_DisplayObject(const LA_DisplayObject& s);

	virtual ~LA_DisplayObject();

	void set(const Point<GLfloat,3>&  pos,							// In scene coordinates
			 const Vector<GLfloat,3>& dir,
			 const Vector<GLfloat,3>& up);

	Point<GLfloat,3>	getPos()		{ return _mat*_pos; }		// In scene coordinates
	Vector<GLfloat,3>	getDir()		{ return _mat*_dir; }		// In scene coordinates
	Vector<GLfloat,3>	getSide()		{ return _mat*_side; }		// In scene coordinates
	Vector<GLfloat,3>	getUp()			{ return _mat*_up; }		// In scene coordinates


	virtual void		lockAtTarget(DisplayObject* obj);
	virtual void		lockAtTarget(const Point<GLfloat,3>& p);	// p must be in scene coordinates
	virtual void		lockAtTarget(double d);
	void				unLockTarget();
	bool				isLocked();
	Point<GLfloat,3>	getLockPos();								// In the coordinates of the scene.
	double				getLockDist();

	virtual void turn(Angle a);
	virtual void roll(Angle a);
	virtual void tilt(Angle a);
	virtual void move(GLfloat d);					// Moveing in shooting direction
	virtual void move(const Vector<GLfloat,3>& t);	// Move in direction t.
	virtual void move(char,double);					// Track u (up) d (down) l (left) r (right)
	virtual void move(const Vector<GLfloat,2>& t);	// Track in direction t.

	// LA_DisplayObject version of the virtual transformations of DisplayObject.

	// In local coordinates.
	void rotate(Angle a, const Vector<GLfloat,3>& rot_axel);
	void rotate(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& d);
	void translate(const Vector<GLfloat,3>& trans_vector);

	// In the global coordinates of the scene.
	void rotateGlobal(Angle a, const Vector<GLfloat,3>& rot_axel);
	void rotateGlobal(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& d);
	void translateGlobal(const Vector<GLfloat,3>& trans_vector);

	const char* identity() const { return "LA_DisplayObject";}

private:
	template <class T_Stream>
		void prOut(T_Stream& out) const//, const LA_DisplayObject& v)
	{
		out << _pos << object_separator << _dir << object_separator << _up << object_separator;
	}

	template <class T_Stream>
		void prIn(T_Stream& in)//, LA_DisplayObject& v)
	{ 
		static Separator os(object_separator);
		in >> _pos >> os >> _dir >> os >> _up >> os;
	}

public:
	template <class T_Stream>
		friend T_Stream& operator<<(T_Stream& out, const LA_DisplayObject& v)
	{
		out << *((DisplayObject*)& v);
		v.prOut(out);
		return out;
	}

	template <class T_Stream>
		friend T_Stream& operator>>(T_Stream& in, LA_DisplayObject& v)
	{ 
		in >> *((DisplayObject*)& v);
		v.prIn(in);
		return in;
	}

	template <class T_Stream>
		LA_DisplayObject(T_Stream& in, int st): DisplayObject(in,st ),_mat(),_mat_inv()	// Stream constructor
	{
		prIn(in);
		_side	= _up^_dir;
		_keep_lock_at	= false;
		_lock_at_object	= 0;
		basisChange(_side, _up, _dir, _pos);
	}

};





inline
bool LA_DisplayObject::isLocked()
{ 
	return _keep_lock_at;
}

inline
Point<GLfloat,3> LA_DisplayObject::getLockPos()
{
	if(_lock_at_object)
		return  _lock_at_object->centerPos();
	else
		return  _lock_at_pos;
}

inline
double LA_DisplayObject::getLockDist()
{
	if(_keep_lock_at)
		return  (getLockPos()-getPos()).length();
	else
		return  0.0;
}

inline
Point<GLfloat,3>  LA_DisplayObject::_keepLockAt()
{
	if(_lock_at_object)
		return _mat_inv*_lock_at_object->centerPos();
	else
		return _mat_inv*_lock_at_pos; 
}

inline
void LA_DisplayObject::_updateOrientation(const Point<GLfloat,3>& lock_at_p)
{
	_dir		= lock_at_p - _pos;
	_up			= _up - (_up*_dir)*_dir;
	_side		= _up^_dir;
}

} // end namespace GM_lib

#endif // _LA_DISPLAYOBJECT_H__

