
//********************************************************
// GM_LX_Array.h: interface for the LX_Array classes.
//********************************************************
#ifndef GM_LX_ARRAY_H_
#define GM_LX_ARRAY_H_

#include <vector>
using namespace std;

namespace GM_lib {


//*********************************************************
//****************  The LX_Array class  *******************
//************* Large and exspandable array. **************
//*********************************************************

template<class T>
struct LX_pt
{
	T* p;
	int size;
};



template<class T>
class LX_Array {

   LX_pt<T>*  	p;
   int  no_p;
   int  no1, no2;

   int	no_elem;
   int	max_elem;
   int  incr;

   void newRow(int n, T* p = NULL);
   void no1_2pluss();
   void no1_2minus();

protected:

   T&   elem(int) const;					// return given element.

public:

   LX_Array(int n = 0, int incr=100);						// Default constructor.
   LX_Array(const LX_Array<T>&, int incr=0);				// Copy  constructor.
   LX_Array(int n, const T* p, int inc=100, bool copy=true);// Standard constructor.

  ~LX_Array();									// Destructor
   
   void     clear(int inc=0);				// Remove all elements.

   T&       operator[](int);				// return given element.
   const T& operator()(int) const;			// return copy of given element.

   T&		last() {return (*this)[no_elem-1];} // return last element in array.

   LX_Array<T>& operator=(const LX_Array<T>&);	// Copy the other LX_Array.

   void     stripElements();				// Make each element exist only once.

   bool     exist(const T& t) const			// True if present.
               { return bool(index(t) >= 0); }

   int      index(const T&) const;			// Index in LX_Array (-1 if not present).

   void     alwaysInsert(const T&, bool first=false);
											// Inserting first or last in LX_Array.

   void     operator+=(const T& t)			// Insert last.
               { alwaysInsert(t); }

   bool     insert(const T&, bool first=false);
											// Insert only if not existing,
											// True if inserted (False existing)

   void     alwaysInsert(const LX_Array<T>&, bool first=false);// 
															// 
															// These tree functions
   void     operator+=(const LX_Array<T>& t)					// and operators works
               { alwaysInsert(t); }							// like the tree
															// functions above,
   bool     insert(const LX_Array<T>&, bool first=false);		// but are inserting
															// a whole LX_Array.
															//

   bool     removeIndex(int);				// Remove object with specified index

   bool     removeFirst() {return removeIndex(0);}		// Remove first object in array

   bool     removeLast() {return removeIndex(no_elem-1);}// Remove last object in array

   bool     remove(const T& obj)			// True if removed (False not existing)
               { return removeIndex(index(obj)); } 

   int      size() const					// Number of elements.
               { return no_elem;}

   int      maxSize() const					// Number of max possible elements.
               { return max_elem;}

   void     setSize(int);					// Set number of elements in the LX_Array.
											// New elements are not initialized
											// surplus element removed from end.

   void     setMax(int, int inc=0);					// Set number of possible elements
											// max_elem, no_elem is not changed.

   void     strip();						// Remove any surplus(max_elem=no_elem)

   void     resize(int n) {setSize(n);strip();}

   void     turn();							// Turn the LX_Array
											// (set first element as last etc.)

   void     swap(int i, int j)				// Swap two elements.
               { T tmp = (*this)[i]; (*this)[i] = (*this)[j]; (*this)[j] = tmp; }

   bool	isEmpty() const { return no_elem == 0; }

   bool     operator==(const LX_Array<T>&) const;		// EQUAL?
   bool     operator!=(const LX_Array<T>& a) const	// NOT EQUAL?
               { return !((*this) == a); }


 	// ********** IOSTREAM overloaded operators ***********
public:
	template<class T_Stream>
	friend T_Stream& operator<<(T_Stream& out, const LX_Array<T>& v)
	{
		out << v.size() << endl;
		for(int i = 0; i < v.size(); i++)
			out << v(i) << separator;
	}

	template<class T_Stream>
	friend T_Stream& operator>>(T_Stream& in,  LX_Array<T>& v)
	{
		int elements;
		in >> elements; v.setSize(elements);
		for(int i = 0; i < v.size(); i++)
			in >> v[i];
	}

};


//********************************************************
//**************** insert new element  *******************
//********************************************************

template<class T>
inline
void LX_Array<T>::no1_2pluss()
{
	if(no2 == (p[no1].size-1)) 
	{
		no1++;
		no2 = 0;
	}
	else	no2++;
	no_elem++;

}


//********************************************************
//**************** remove element      *******************
//********************************************************

template<class T>
inline
void LX_Array<T>::no1_2minus()
{
	if(no2 == 0)
	{
		no1--;
		no2 = p[no1].size-1;
	}
	else no2--;
	no_elem--;
}



//********************************************************
//**************** DEFAULT CONSTRUCTOR *******************
//********************************************************

template<class T>
inline
LX_Array<T>::LX_Array(int n, int inc)
{ 

	max_elem = no_elem = no_p = no1 = no2 = 0;
	incr = inc;
	p = NULL;
    if (n > 0)	newRow(n);
}


//********************************************************
//**************** COPY CONSTRUCTOR **********************
//********************************************************

template<class T>
inline
LX_Array<T>::LX_Array(const LX_Array<T>& da, int inc)
{ 
	max_elem = no_elem = no_p = no1 = no2 = 0;
	if(inc == 0)	incr = da.incr;
	else			incr = inc;
	p = NULL;

	if( da.size() > 0)
	{
		newRow(da.size());
		for(int i=0; i<da.size(); i++) p[0].p[i] = da(i);
		no_elem = da.size();
		no1 = 1;
		no2 = 0;
	}
}


//********************************************************
//*************** STANDARD CONSTRUCTOR *******************
//********************************************************

template<class T>
inline
LX_Array<T>::LX_Array(int n, const T* p, int inc, bool copy)
{
	max_elem = no_elem = no_p = no1 = no2 = 0;
	incr = inc;
	p = NULL;

	if (n>0)
	{
		if (copy)
		{
			newRow(n);
			for(int i=0; i<n; i++) p[0].p[i] = p[i];
		}
		else
			newRow(n,p);
		no_elem = n;
		no1 = 1;
		no2 = 0;
	}
}


//********************************************************
//**************** RESET THE LX_Array ********************
//********************************************************

template<class T>
inline
void  LX_Array<T>::clear(int inc)
{
	for(int i=0; i<no_p; i++)
		if(p[i].size > 0) {delete [] p[i].p; p[i].p=NULL; }
	if(no_p > 0 && p!=NULL) delete p;
 	max_elem = no_elem = no_p = no1 = no2 =0;
	p = NULL;
	if(inc>0)	incr = inc;
}


//********************************************************
//**************** elem(int i) -ORGINAL ELEMENT***********
//********************************************************

template<class T>
inline
T&  LX_Array<T>::elem(int j) const
{
#ifdef DEBUG
    if (j < 0 ||  j >= no_elem)
		printf("index %d is outside (0, %d)\n", j, no_elem);
#endif
	if(no_p == 1)	return p[0].p[j];
  int i;
 	for(i=0; i<no_p; i++)
		if(j < p[i].size)	break;
		else			j -= p[i].size;

	return p[i].p[j]; 
}


//********************************************************
//**************** OPERATOR[] -ORGINAL ELEMENT************
//********************************************************

template<class T>
inline
T&  LX_Array<T>::operator[](int j)
{
#ifdef DEBUG
    if (j < 0 ||  j >= no_elem)
		printf("index %d is outside (0, %d)\n", j, no_elem);
#endif

	if(no_p == 1)	return p[0].p[j];
  int i;
 	for(i=0; i<no_p; i++)
		if(j < p[i].size)	break;
		else			j -= p[i].size;

	return p[i].p[j]; 
}


//********************************************************
//**************** OPERATOR() - COPY OF ELEMENT **********
//********************************************************

template<class T>
inline
const T&   LX_Array<T>::operator()(int j) const
{
#ifdef DEBUG
    if (j < 0 ||  j >= no_elem)
		printf("index %d is outside (0, %d)\n", j, no_elem);
#endif
	if(no_p == 1)	return p[0].p[j];

  int i;
	for(i=0; i<no_p; i++)
		if(j < p[i].size)	break;
		else			j -= p[i].size;

	return p[i].p[j]; 
}


//********************************************************
//**************** INSERT(OBJECT) - INSERT IN LX_Array  *****
//********************************************************

template<class T>
inline
bool	 LX_Array<T>::insert(const T& obj, bool first)
{ 
	bool ret = true;
	if (exist(obj))
		ret = false;
	else
		alwaysInsert(obj,first);
	return ret;
}


//********************************************************
//************** INSERT(LX_Array) - INSERT HOLE LX_Array  ******
//********************************************************

template<class T>
inline
bool	 LX_Array<T>::insert(const LX_Array<T>& arr, bool first)
{ 
	bool ret = true;

	for(int i=0; i<arr.size(); i++)
	{
		if (exist(arr(i)))  ret = false;
		else			alwaysInsert(arr(i),first);
	}
	return ret;
}


} // end namespace GM_lib

#include "GM_LX_Array.c"

#endif /* GM_LX_ARRAY_H_ */

