//********************************************************
// GM_Material.h: Object describing a material.
//********************************************************
#ifndef _MATERIAL_H__
#define _MATERIAL_H__

#include <GM_OpenGL.h>
#include <GM_Texture.h>
#include <GM_String.h>

namespace GM_lib {


//**********************************************************************
//	Contains the following classes, predefinitions and external objects: 
//**********************************************************************

class Material;
class MaterialObject; 
class MaterialObjectList;
// A set of predefined Colors
// A set of predefined Materials


//**********************************************************************
// The Material class 
// containing:	ambient, diffuce and specular colors, 
//				shininess (0-100) and
//				texture ID and
//				source and destination blending factors
//**********************************************************************

class Material  
{
	GLenum				__source_blend_factor;
	GLenum				__destination_blend_factor;
	GLenum				__sided;

protected:
	GLuint				__texname;	
	GL_Color			__amb;
	GL_Color			__dif;
	GL_Color			__spc;
	float				__shininess;
	float				__transparancy;
public:																						
	Material(const GL_Color& amb = GL_Color(0.7,0.0,0.0,1.0),// Default/Standar constructor
			 const GL_Color& dif = GL_Color(0.7,0.0,0.0,1.0), 
			 const GL_Color& spc = GL_Color(0.7,0.0,0.0,1.0),
			 float shininess = 0.0,
			 GLuint texname = 0) 
		{ set(amb,dif,spc,shininess,texname);set(GL_ONE,GL_ONE);__sided = GL_FRONT_AND_BACK;__transparancy = 1;}

	Material(const Material&  v) {memcpy(this,&v,sizeof(Material));} // Copy constructor

	void set(const GL_Color& amb,			
			 const GL_Color& dif, 
			 const GL_Color& spc,
			 float shininess,
			 GLuint texname);

	void set(const GL_Color& amb,			
			 const GL_Color& dif, 
			 const GL_Color& spc);
	
	void set(const GLenum sfactor,
			 const GLenum dfactor);

	void set(float shininess); 

	void set(GLuint texname); 

	void setAmb(const GL_Color& amb)	{ __amb=amb; }
	void setDif(const GL_Color& dif)	{ __dif=dif; }
	void setSpc(const GL_Color& spc)	{ __spc=spc; }

	void setSided(GLenum s)		{__sided = s;}
	void setDoubleSided(bool s)	{ if(s) __sided = GL_FRONT_AND_BACK; else __sided = GL_FRONT;}

	void setTransparancy(double t)	{__transparancy = t;}

	int  texture() { return __texname; }

	virtual 
	void glSet();	// push this material to current GL-state.
	virtual 
	void glUnSet();

	void glSetInverse();
	bool isTransparent();

	Material& operator=(const Material& m) 
		{ memcpy(this,&m,sizeof(Material)); return *this; } // Make a copy 

	bool operator==(const Material& m) const
	{ return __amb == m.__amb && __dif == m.__dif && __spc == m.__spc && 
			 __shininess == m.__shininess && __texname == m.__texname;}

	bool operator<(const Material& m)const { return true; } // dummy for Array
	
	template <class T_Stream> 
	friend T_Stream& operator<<(T_Stream& out, const Material& m)
	{
		out << m.__amb << element_separator 
			<< m.__dif << element_separator
			<< m.__spc << element_separator
			<< m.__shininess << element_separator;
		if(m.__texname)
		{
			char *tx=Texture::findFileName(m.__texname);
			out << tx << group_separator;
		}
		else
		{
			char* no = "no-texture";
			out << no << group_separator;
		}

		return out;
	}

	template <class T_Stream>
	friend T_Stream& operator>>(T_Stream& in, Material& m)
	{
		static Separator es(element_separator);
		static Separator gs(group_separator);
		char* fn=0;
		in  >> m.__amb >> es 
			>> m.__dif >> es
			>> m.__spc >> es
			>> m.__shininess >> es
			>> fn >> gs;
		if(!(fn[0]=='n' && fn[1]=='o' && fn[2]=='-')) 
		{
			Texture tx(true);
			m.set(tx.LoadGL(fn)); 
		}
		m.set(GL_ONE,GL_ONE);
		delete fn;
		return in;
	}
};


//*******************************************
// Predefined Colors
//*******************************************

extern GL_Color White;
extern GL_Color Red;
extern GL_Color Green;
extern GL_Color Blue;
extern GL_Color Black;


//*******************************************
// Predefined Materials
//*******************************************

extern Material Brass;		
extern Material Bronze;			
extern Material PolishedBronze;
extern Material Chrome;	
extern Material Copper;			
extern Material PolishedCopper;
extern Material Gold;	
extern Material PolishedGold;
extern Material Pewter;	
extern Material Silver;			
extern Material PolishedSilver;
extern Material Emerald;
extern Material Jade;		
extern Material Obsidian;
extern Material Pearl;		
extern Material Ruby;			
extern Material Turquoise;
extern Material BlackPlastic;
extern Material BlackRubber;
extern Material Snow;
extern Material Plastic;
extern Material PolishedRed;
extern Material PolishedGreen;


//**********************************************************************
// The MaterialObject class 
// containing:	A name and
//				is a Material
//**********************************************************************

class MaterialObject: public Material
{
	GM_String		_name;			// Size of name is max 16 letters
	GM_String		_texture_name;	// Size of name is max 16 letters

public:
	MaterialObject(	const Material& mat = Snow, 
					const char* name = "Dummy",
					const char* tex_name = "\0") : Material(mat)
		{ _name = name; _texture_name = tex_name; }

	MaterialObject(	const Material& mat, 
					string name,
					string tex_name = "\0") : Material(mat)
		{ _name = name; _texture_name = tex_name; }

	MaterialObject(	const MaterialObject& m):Material(m)
		{ _name = m._name; _texture_name = m._texture_name; }

	void setMaterial(const Material& m)	{ Material::operator=(m); }
	void setName(const string& name)	{ _name = name; }
	void setName(const char* name)		{ _name = name; }
	void setTexture(const string& name)	{ _texture_name = name; }
	void setTexture(const char* name)	{ _texture_name = name; }

	void loadGL()
		{ if(_texture_name.size()>0) {Texture texture; __texname = texture.LoadGL(_texture_name.c_str());}}

	void deleteTexture()
		{ if(glIsTexture(__texname)) glDeleteTextures(1, &__texname); }

	const GM_String&	name()	   const { return _name; }
	const char*			nameC()	   const { return _name.c_str(); }
	const Material&		material() const { return *this; }

	bool is(const char* name) const 
		{ return _name == name; }
	bool is(const string& name) const 
		{ return _name == name; }
	bool is(const Material& m) const 
	{ return m == (Material)(*this); }

	MaterialObject& operator=(const MaterialObject& m)
	{ Material::operator=(m); _name = m._name; _texture_name = m._texture_name; return *this; }

//	bool operator==(const MaterialObject& m) const
//		{ m._name == _name; }
};


//********************************************
// List of materialObject
//*********************************************
 
class MaterialObjectList: Array<MaterialObject>
{
public:
	MaterialObjectList(bool init_=true) { if(init_) initPreDef(); }
	MaterialObjectList(char* file_name) { if(file_name) readFromFile(file_name);}

	void	initPreDef();
	bool	readFromFile(char* file_name);
	bool	storeToFile(char* file_name);
};

//*******************************************
// Inline member functions and operators
//*******************************************

inline
void Material::set(	const GL_Color& amb,			
					const GL_Color& dif, 
					const GL_Color& spc,
					float shininess,
					GLuint texname)
{
	__amb = amb;
	__dif = dif;
	__spc = spc;
	__shininess = shininess;
	__texname = texname;
}

inline
void Material::set(	const GL_Color& amb,			
					const GL_Color& dif, 
					const GL_Color& spc)
{
	__amb = amb;
	__dif = dif;
	__spc = spc;
}

inline
void Material::set(const GLenum sfactor,const GLenum dfactor)
{
	__source_blend_factor		= sfactor;
	__destination_blend_factor	= dfactor;
}

inline
void Material::set(float shininess)
{ 
	__shininess = shininess;
}

inline
void Material::set(GLuint texname)
{ 
	__texname = texname;
}

inline
bool Material::isTransparent()
{
	return ( !((__source_blend_factor==GL_ONE) && (__destination_blend_factor==GL_ONE)) );
};

} // end namespace GM_lib

#endif // _MATERIAL_H__

