
//*************************************************************************
//************ GM_PCurve.h: interface for the PCurve classes. *************
//***				This class is an abstract class, and				***
//***			  is the base class for parametric curves.				***
//*************************************************************************
#ifndef GM_PCURVE_H_
#define GM_PCURVE_H_

#include <GM_DVector.h>
#include <GM_Material.h>
#include <GM_LA_DisplayObject.h>

namespace GM_lib {

template <class T, int n>
class PCurve: public LA_DisplayObject
{
	bool				__dynamic;
	GLuint				__display_list;
	unsigned int		__rendermode;

	// Private help functions 
	void	_eval_(T t, int d) {	if(!(d <=__d &&  t ==__t)) { __t=t; __d=d; _eval(_shift(t),d); }}
	double	trapez(T a, T b, int nn);
	T		integral(T a, T b);	

 protected:
	T							__st_sam;		// Frst sampling value usually (startPar_u(),startPar_v())
	T							__end_sam;		// Last sampling value usually (endtPar_u(),endPar_v())
	int							__sam;			// Samplings-number u directions
	int							__defalt_d;		// used by operator() for number of derivative to evaluate. 
	Material					__material;		// Material data for the curve.
	GL_Color					__color;		// Color if lightning is disabled
	bool						__lighted;		// Use lighting and material otherwice color
	DVector<Point<GLfloat,3> >	__vertex;		// Vertex array for display
	GLfloat						__line_width;	// The line with used
	GLfloat						__point_size;	// The point size used
		
	DisplayObject*				__parent;		// The mother if present!

	
	//************************************************************
	// xxxxx     The result of the previous evaluation!!!   xxxxxx
	//************************************************************
	DVector<Vector<T,n> >	__p;		// Position and belonging derivatives // __p[1]
	T						__t;		// The parameter value used for last evaluation
	int						__d;		// Number of derivatives computed last time

	// Shift of parameter.
	T					__tr;	// translate
	T					__sk;	// skale
	T					_shift(T t) {return _st_p()+__sk*(t - startPar()); }

	//*********************************************************************
	//********   Virtual functions nessesary for inheritance !!!   ********
	//*  This functions must be implemented in the real parametric Curve  *
	//*********************************************************************
	virtual
	void	_eval(T t, int d) = 0;
	virtual
	T		_st_p()  = 0;
	virtual
	T		_end_p() = 0;

	//**************************************************************
	//********   Virtual functions already implemented !!!   *******
	//**  Must only be implemented if you want something special  **
	//**************************************************************
	virtual
	void _makeSample(Array<Point<T,n> >& a, T eps);	// Always smooth, requires derivatives 
	virtual 
	void _makeSample(Array<Point<T,n> >& a, int m1);// Given sampling rate 

	//**************************************************************************
	//*************     Functions inherited from DisplayObject.    *************
	//* You do not need to make these functions in the real parametric Curve   *
	//********      exept if you want to do something special.         *********
	//**************************************************************************
	void localDisplay();
	void localSelect();

 public:
	PCurve(int sampling = 100);
	PCurve(const PCurve<T,n>& c); 

	virtual
	~PCurve() {}

	void setDomain(T start, T end)	{__sk=(end-start)/(_end_p()-_st_p()); __tr=start-_st_p(); }
	void setScaleDomain(T sc)		{__sk = sc;}
	void setTransDomain(T tr)		{__tr = tr;}

	T deltaPar()	{ return __sk*(_end_p()-_st_p()); }
	T startPar()	{ return _st_p()+__tr; }
	T endPar()		{ return startPar()+deltaPar(); }


	void resetPlot(T epsilon, bool dynamic=false);
	void setMaterial(const Material& m) { __material=m; }
	void setColor(const GL_Color& m)	{ __color=m; }
	void setLighted(bool lighted) { __lighted = lighted; }


	void setEval(int d)			{ __defalt_d = d; }

	Point<T,n> operator()(T t)	{ _eval(t,__defalt_d); return _present*(Point<T,n>)__p[0]; }
	Vector<T,n> der1(T t)		{ _eval(t,1); return _present*__p[1]; }
	Vector<T,n> der2(T t)		{ _eval(t,2); return _present*__p[2]; }

	DVector<Vector<T,n> >& evaluate(T t, int d); 
	DVector<Vector<T,n> >& evaluateLocal(T t, int d) { _eval_(t,d); return __p; }
	DVector<Vector<T,n> >& evaluateMother(T t, int d);

	T radius(T t);
	T curvature(T t)	{ return 1.0/radius(t); }
	T speed(T t)		{ return der1(t).length(); }
	T curveLength();

	void setRenderMode(unsigned  int r) {__rendermode = r;}
	void toggleRenderMode() {if(__rendermode==GL_POINTS) __rendermode=GL_LINE_STRIP; else __rendermode=GL_POINTS;}

	virtual
	bool			closed() const		{ return false; }

	const char*		identity() const	{ return "PCurve";}

	//**************************************************************
	//	PCurve version og affin transformation
	//	Necessary because if there is a parent 
	//	then give this parent a message if the surface is edited
	//**************************************************************

	// In local coordinates.
	void rotate(Angle a, const Vector<GLfloat,3>& rot_axel);
	void rotate(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& d);
	void translate(const Vector<GLfloat,3>& trans_vector);

	// In the global coordinates of the scene.
	void rotateGlobal(Angle a, const Vector<GLfloat,3>& rot_axel);
	void rotateGlobal(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& d);
	void translateGlobal(const Vector<GLfloat,3>& trans_vector);
};

} // end namespace GM_lib



using namespace GM_lib;


template <class T, int n>
inline
PCurve<T,n>::PCurve(int sampl)
{
	__rendermode	= GL_LINE_STRIP;
	__line_width	= 2.0;
	__point_size	= 5.0;
	__lighted		= false;
	__sam			= sampl;
	__d				= -1; 
	__tr			= T(0); 
	__sk			= T(1); 
	__display_list	= 0; 
	_type_id		= 2000;
	__parent		= NULL;
	setEval(0);
}

template <class T, int n>
inline
PCurve<T,n>::PCurve(const PCurve<T,n>& c)
{
	__rendermode	= c.__rendermode;
	__line_width	= c.__line_width;
	__point_size	= c.__point_size;
	__lighted		= c.__lighted;
	__sam			= c.__sam;
	__p				= c.__p; 
	__t				= c.__t; 
	__d				= c.__d; 
	__tr			= c.__tr; 
	__sk			= c.__sk; 
	__display_list	= 0; 
	_type_id		= 2000;
	__parent		= c.__parent;
	setEval(0); 
}



template <class T, int n>
void PCurve<T,n>::localDisplay()
{
	GLboolean lg;
	if(__lighted)
	{
		__material.glSet();
	}
	else
	{
		glGetBooleanv(GL_LIGHTING,&lg);
		if(lg) glDisable(GL_LIGHTING);
		__color.glSet();
	}
	glLineWidth(__line_width);
	glPointSize(__point_size);
	if(__dynamic)
	{
		glVertexPointer(3,GL_FLOAT,0,__vertex[0].ptr());
		glDrawArrays(__rendermode,0,__vertex.dim());
	}
	else
	{
		if(__rendermode == GL_LINE_STRIP)	glCallList(__display_list);
		else if(__rendermode == GL_POINTS)	glCallList(__display_list+1);
	}
	glPointSize(1.0);
	glLineWidth(1.0);
	if(__lighted)
		__material.glUnSet();
	else
		if(lg) glEnable(GL_LIGHTING);
}


template <class T, int n>
void PCurve<T,n>::localSelect()
{
	glLineWidth(10);
	if(__dynamic)
	{
		glVertexPointer(3,GL_FLOAT,0,__vertex[0].ptr());
		glDrawArrays(GL_LINE_STRIP,0,__vertex.dim());
	}
	else
		glCallList(__display_list);
	glLineWidth(1.0);
}


template <class T, int n>
void  PCurve<T,n>::resetPlot(T eps, bool dynamic)
{
	if(__display_list)
	{
		glDeleteLists(__display_list,2);
		__display_list=0;
	}
	__dynamic = dynamic;

	if(eps==T(0) && __vertex.dim()>0) eps = __vertex.dim();

	Array<Point<T,n> > p;
	if(eps>1.0) _makeSample(p, (int)eps);
	else		_makeSample(p, eps);

	Sphere<GLfloat,3>  s;
	s.resetPos(Point<GLfloat,3>(p[0]));
	s += Point<GLfloat,3>(p[p.size()-1]);
	for(int i=p.size()-2; i>0; i--)
		s += Point<GLfloat,3>(p[i]);
	setSurroundingSphere(s);

	if(dynamic)
	{
		glEnableClientState(GL_VERTEX_ARRAY);
		__vertex.setDim(p.size());
		for(int i=0; i<p.size(); i++) __vertex[i] = Point<GLfloat,3>(p[i]);
	}
	else
	{
		__display_list = glGenLists(2);

		// Make displaylist for display of lined curve and select
		glNewList(__display_list, GL_COMPILE); 
		glBegin(GL_LINE_STRIP);
			for(int i=0;i<p.size();i++) glPoint(p[i]);
		glEnd();
		glEndList();

		// Make displaylist for display of pointed curve
		glNewList(__display_list+1, GL_COMPILE); 
		glBegin(GL_POINTS);
			for(int i=0;i<p.size();i++) glPoint(p[i]);
		glEnd();
		glEndList();
	}
}


template <class T, int n>
void  PCurve<T,n>::_makeSample(Array<Point<T,n> >& a, T eps)
{
	T t, ddt, dp5 = deltaPar()/5;
	bool go_on=true;
	for(t=startPar(); go_on; t+=ddt)
	{	
		_eval(t,2);
		a += __p[0];
		if (a.size() >7000) break;
		T _r = __p[1]*__p[1];
		Vector<T,n> r = __p[2]-((__p[2]*__p[1])/_r)*__p[1];
		_r = sqrt(double(_r));
		ddt = 2*_r/__p[1].length()*acos((_r-eps)/_r);
		if(ddt>dp5) ddt = dp5;
		if(ddt>endPar()-t) 
		{
			T oddt=ddt;
			ddt=endPar()-t; 
			go_on=false;
			if(ddt < oddt/5)
				a.removeIndex(a.size()-1);
		}
	}
	_eval(t,0);
	a += __p[0];
}

// BBang, 28.04.04
template <class T, int n>
void  PCurve<T,n>::_makeSample(Array<Point<T,n> >& a, int m1)
{
	T du = deltaPar()/(m1-1);
	a.setSize(m1);

	for(int i=0; i<m1; i++)
	{
		_eval(startPar()+i*du, 0);
		a[i] = __p[0];	
	}
}

template <class T, int n>
DVector<Vector<T,n> >& PCurve<T,n>::evaluate(T t, int d) 
{ 
	static DVector<Vector<T,n> >p;

	_eval(t,d); 
	p.setDim(d+1);
	p[0] = _present*static_cast<Point<T,n> &> (__p[0]);
	for(int i=1; i<=d; i++) p[i] = _present*__p[i];
	return p; 
} 

template <class T, int n>
DVector<Vector<T,n> >& PCurve<T,n>::evaluateMother(T t, int d) 
{
	static DVector<Vector<T,n> > p;

	_eval(t,d); 
	p.setDim(d+1);
	p[0] = _matrix*static_cast<Point<T,n> &> (__p[0]);
	for(int i=1; i<=d; i++) p[i] = _matrix*__p[i];
	return p; 
}

template <class T, int n>
double PCurve<T,n>::trapez(T a, T b,int nn)  // I(0,n)
{	
	int n=pow(2,nn);
	double h=(b-a)/n;
	double sum=0.0;
	for (int i=1; i<n; i++) sum+= speed(a+h*i);
	sum*=h;
	sum+=(speed(a)+speed(b))*0.5*h;
	return sum;
}
template <class T, int n>
T PCurve<T,n>::integral(T a, T b) //  I(v,n) almost like p. 23 (5.6 and 5.7)
{	
	int stp=8;
	GM_lib::DMatrix<double> In(10,10);

	In[0][0]=trapez(a,b,0);
	In[0][1]=trapez(a,b,1);

	In[1][0]=(pow(2,2*1)*In[0][1]-In[0][0])/(pow(2,2*1)-1.0);

	for(int v=1;v<stp+1;v++) 
	{
		In[0][v]=trapez(a,b,v);
		for(int k=1;k<=v;k++) In[k][v-k]=(pow(2,2*k)*In[k-1][v-k+1]-In[k-1][v-k])/(pow(2,2*k)-1.0);
	}
	return In[stp][0];
}

template <class T, int n>
T PCurve<T,n>::curveLength() 
{ 
	int step=8192;
	T sum=T(0);
	T dt=deltaPar()/(T)(step-1);
// for(int i=0; i<step; i+=2)  sum+=integral(startPar()+i*dt,startPar()+(i+1)*dt);
	
	T d1,d2,d3;
	for(int i=0; i<step; i+=2) 
	{ 
		d1=speed(startPar()+dt*i);
		d2=4*speed(startPar()+dt*(i+1));
		d3=speed(startPar()+dt*(i+2));
		sum+=(d1+d2+d3)*dt/3.0;
	}
	return sum;
}


//****************************************************
//	PCurve version og affin transformation
//	Necessary because if there is a parent then
//	give the parent a message if the surface is edited
//****************************************************

// In local coordinates.
template <class T, int n>
void PCurve<T,n>::rotate(Angle a, const Vector<GLfloat,3>& rot_axel) 
{
	LA_DisplayObject::rotate(a,rot_axel); 
	if(__parent) __parent->edit(this); 
}
template <class T, int n>
void PCurve<T,n>::rotate(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& d)
{
	LA_DisplayObject::rotate(a,p,d);
	if(__parent) __parent->edit(this);
}
template <class T, int n>
void PCurve<T,n>::translate(const Vector<GLfloat,3>& trans_vector)
{
	LA_DisplayObject::translate(trans_vector);
	if(__parent)__parent->edit(this);
}

// In the global coordinates of the scene.
template <class T, int n>
void PCurve<T,n>::rotateGlobal(Angle a, const Vector<GLfloat,3>& rot_axel)
{
	LA_DisplayObject::rotateGlobal(a,rot_axel);
	if(__parent) __parent->edit(this);
}
template <class T, int n>
void PCurve<T,n>::rotateGlobal(Angle a, const Point<GLfloat,3>& p,const UnitVector<GLfloat,3>& d)
{
	LA_DisplayObject::rotateGlobal(a,p,d);
	if(__parent) __parent->edit(this);
}
template <class T, int n>
void PCurve<T,n>::translateGlobal(const Vector<GLfloat,3>& trans_vector)
{
	LA_DisplayObject::translateGlobal(trans_vector);
	if(__parent) __parent->edit(this);
}


#endif   /* GM_PCURVE_H_ */
