
//********************************************************
// GM__T_Array.h: interface for the T_Array classes.
//********************************************************
#ifndef GM_T_ARRAY_H_
#define GM_T_ARRAY_H_

#include <stddef.h>

namespace GM_lib {


//*********************************************************
//****************  The T_Array class  ********************
//************* Tiny and fast exessable array. ************
//*********************************************************



template<class T>
class T_Array {

   T*  	p;

   short	no_elem;
   short	max_elem;


   static int incr;
 
   T&   elem(int) const;					// return given element.

public:

   T_Array(int n = 0);							// Default constructor.
   T_Array(const T_Array<T>&);					// Copy  constructor.
   T_Array(const Array<T>&);					// Copy  constructor.
   T_Array(int n, const T* p, bool copy=true);	// Standard constructor.

  ~T_Array();									// Destructor
   
   void     clear();							// Remove all elements.

   T&       operator[](int);					// return given element.
   const T& operator()(int) const;				// return copy of given element.

   T&		last() {return (*this)[no_elem-1];} // return last element in array.

   T_Array<T>& operator=(const T_Array<T>&);	// Copy the other T_Array.

   void     stripElements();					// Make each element exist only once.

   bool     exist(const T& t) const				// True if present.
               { return bool(index(t) >= 0); }

   int      index(const T&) const;				// Index in T_Array (-1 if not present).

   void     alwaysInsert(const T&, bool first=false);
												// Inserting first or last in T_Array.

   void     operator+=(const T& t)				// Insert last.
               { alwaysInsert(t); }

   bool     insert(const T&, bool first=false);
												// Insert only if not existing,
												// True if inserted (False existing)

   void     alwaysInsert(const T_Array<T>&, bool first=false);
															// These tree functions
   void     operator+=(const T_Array<T>& t)					// and operators works
               { alwaysInsert(t); }							// like the tree
															// functions above,
   bool     insert(const T_Array<T>&, bool first=false);	// but are inserting
															// a whole T_Array.

   bool     removeIndex(int);					// Remove object with specified index

   bool     removeFirst() {return removeIndex(0);}		// Remove first object in array

   bool     removeLast() {return removeIndex(no_elem-1);}// Remove last object in array

   bool     remove(const T& obj)				// True if removed (False not existing)
               { return removeIndex(index(obj)); } 

   int      size() const						// Number of elements.
               { return no_elem;}

   int      maxSize() const						// Number of max possible elements.
               { return max_elem;}

   void     setSize(int);						// Set number of elements in the T_Array.
												// New elements are not initialized
												// surplus element removed from end.

   void     resetSize();						// Set number of elements in the array to 0
												// without doing anything else! FAST VERSON.

   void     setMax(int);						// Set number of possible elements
												// max_elem, no_elem is not changed.

   void     strip();							// Remove any surplus(max_elem=no_elem)

   void     resize(int n) {setSize(n);strip();}

   void     turn();								// Turn the T_Array
												// (set first element as last etc.)

   void     swap(int i, int j)					// Swap two elements.
               { T tmp = (*this)[i]; (*this)[i] = (*this)[j]; (*this)[j] = tmp; }

   bool		isEmpty() const { return no_elem == 0; }

   bool     operator==(const T_Array<T>&) const;	// EQUAL?
   bool     operator!=(const T_Array<T>& a) const	// NOT EQUAL?
               { return !((*this) == a); }


   static
   void	    setIncr(int i) { incr = i; }

 	// ********** IOSTREAM overloaded operators ***********

#if defined(IOSTREAMH) || defined(__IOSTREAM_H) || defined(_IOSTREAM_)
private:
   ostream& _prOut(ostream& out) const;
   istream& _prIn(istream& in) ;
public:

  friend
   ostream& operator<<(ostream& out, const T_Array<T>& v)
	   { return v._prOut(out); }

  friend
   istream& operator>>(istream& in, T_Array<T>& v)
	   { return v._prIn(in); }

   void print(char prompt[]="T_Array<T>", ostream & out = cout) const
	   { out << prompt << ": " << (*this) << "\n"; }

#endif
};


//********************************************************
//**************** DEFAULT CONSTRUCTOR *******************
//********************************************************

template<class T>
inline
T_Array<T>::T_Array(int n)
{ 
	no_elem = 0;
    if (n > 0)	{ p = new T[n]; max_elem = n; }
	else		{ p = NULL;		max_elem = 0; }
}


//********************************************************
//**************** COPY CONSTRUCTOR **********************
//********************************************************

template<class T>
inline
T_Array<T>::T_Array(const T_Array<T>& da)
{
	no_elem = da.no_elem;

	if( no_elem > 0)
	{
		p = new T[da.size()]; 
		max_elem = da.no_elem;
		for(int i=0; i<da.size(); i++) p[i] = da.p[i];
	}
	else
	{
		no_elem = max_elem = 0;
		p = NULL;
	}
}


//********************************************************
//*************** STANDARD CONSTRUCTOR *******************
//********************************************************

template<class T>
inline
T_Array<T>::T_Array(int n, const T* pt, bool copy)
{
	if (n>0)
	{
		max_elem = no_elem = n;
		if (copy)
		{
			p = new T[n]; 
			for(int i=0; i<n; i++) p[i] = pt[i];
		}
		else
			p = pt;
	}
	else
	{
		no_elem = max_elem = 0;
		p = NULL;
	}
}


//********************************************************
//**************** RESET THE T_Array ********************
//********************************************************

template<class T>
inline
void  T_Array<T>::clear()
{
	if(max_elem > 0) 
	{ 
		delete [] p;
		p = NULL;
		max_elem = no_elem = 0;
	}
}


//********************************************************
//**************** elem(int i) -ORGINAL ELEMENT***********
//********************************************************

template<class T>
inline
T&  T_Array<T>::elem(int j) const
{
#ifdef DEBUG
    if (j < 0 ||  j >= no_elem)
		printf("index %d is outside (0, %d)\n", j, no_elem);
#endif

	return p[j]; 
}


//********************************************************
//**************** OPERATOR[] -ORGINAL ELEMENT************
//********************************************************

template<class T>
inline
T&  T_Array<T>::operator[](int j)
{
#ifdef DEBUG
    if (j < 0 ||  j >= no_elem)
		printf("index %d is outside (0, %d)\n", j, no_elem);
#endif

	return p[j]; 
}


//********************************************************
//**************** OPERATOR() - COPY OF ELEMENT **********
//********************************************************

template<class T>
inline
const T& T_Array<T>::operator()(int j) const
{
#ifdef DEBUG
    if (j < 0 ||  j >= no_elem)
		printf("index %d is outside (0, %d)\n", j, no_elem);
#endif

	return p[j]; 
}


//********************************************************
//**************** INSERT(OBJECT) - INSERT IN T_Array  *****
//********************************************************

template<class T>
inline
bool	 T_Array<T>::insert(const T& obj, bool first)
{ 
	bool ret = true;
	if (exist(obj))  ret = false;
	else             alwaysInsert(obj,first);
	return ret;
}


//********************************************************
//************** INSERT(T_Array) - INSERT HOLE T_Array  ******
//********************************************************

template<class T>
inline
bool	 T_Array<T>::insert(const T_Array<T>& arr, bool first)
{ 
	bool ret = true;

	for(int i=0; i<arr.size(); i++)
	{
		if (exist(arr(i)))  ret = false;
		else			alwaysInsert(arr(i),first);
	}
	return ret;
}


//********************************************************
//**************** PRIVATE OSTREAM FUNCTIONS *************
//********************************************************


#if defined(IOSTREAMH) || defined(__IOSTREAM_H) || defined(_IOSTREAM_)
template <class T>
inline
ostream& T_Array<T>::_prOut(ostream& out) const
{  
	out << size() << "\n"; 
	for (int i=0; i<no_elem;i++) out << (*this)(i) << ' '; 
	out << endl;
	return out;
}

template <class T>
inline
istream& T_Array<T>::_prIn(istream& in)
{  
	int no; in >> no; setSize(no); 
	for (int i=0; i<no; i++)  in >> (*this)[i];
	return in;
}
#endif

//********************************************************
//**************** COPY CONSTRUCTOR from standar array ***
//********************************************************


template <class T>
inline
T_Array<T>::T_Array(const Array<T>& a)
{
	clear();
	for(int i=0; i<a.size(); i++) alwaysInsert(a(i));
	return *this;
}


//********************************************************
//************ Reset array (0 elements fast version) *****
//********************************************************


template<class T>
inline
void T_Array<T>::resetSize()
{ 
	no_elem=0;
}

} // end namespace GM_lib

#include "GM_T_Array.c"

#endif /* GM_T_ARRAY_H_ */

