#include	"Mesh3DS.h"

//********************************************************************
// Constructor preparing the reading by making a new H3dsScene object
//********************************************************************
Mesh3DS::Mesh3DS (const char * filename, int threshold, bool force_2sided)
{
	InFile						= 0;
	_default_material_warning	= 0;
	force2sided					= force_2sided;
	_display_list				= 0;
	_num_display_list			= 0;
	cos_threshold				= (GLfloat)cos(threshold*M_PI_180);
	MScene						= new H3dsScene();
}


//*****************************************************************************
// Destructor deleting displaylists, this might be done by the materialObjects
//*****************************************************************************
Mesh3DS::~Mesh3DS ()
{
	if(glIsList(_display_list)) glDeleteLists(_display_list, _num_display_list);
	delete	MScene;
}


//*****************************************************************************************
// Calling the display for each polygon, this might be separeted in several DisplayObjects
//*****************************************************************************************
void Mesh3DS::localDisplay()
{
	for( int i=0; i<_num_display_list; i++)
	{
		poly_list[i].mat().glSet();
		glCallList(_display_list+i); 
		poly_list[i].mat().glUnSet();
	}
}


//********************************************************
// The same comments for localSelect as for localDisplay
//********************************************************
void Mesh3DS::localSelect()
{
	for( int i=0; i<_num_display_list; i++)
	{
		glCallList(_display_list+i); 
	}
}


//*******************************************************************
// Make a new polygonObject in the polygon list,
// and set the given material (mat) into this new object.
//*******************************************************************
void Mesh3DS::AddMaterial(const MaterialObject& mat) // (const H3dsMat& mat)
{	
	poly_list += POLYGONLIST();
	POLYGONLIST& poly_entry = poly_list.last();
	poly_entry.material = mat;
	if(force2sided) poly_entry.material.setDoubleSided(true); // Propagate flag which forces 2-Sided Materials
	poly_entry.material.loadGL();
}


//*******************************************************************
// Make a new Triangle in the triangle list of a Polygon object
// with the material: mat_name.
//*******************************************************************
void Mesh3DS::AddPolygon (const Point3D<GLuint>& tri, const GM_String& mat_name)
{
	int			mat_num = FindMaterialIndex(mat_name);	// Find the polygon list index we belong to
	TRIANGLE	surface(tri);
	if (!CalculateSurfaceNormal(surface)) return;		// Ignore degenerate polygons
	poly_list[mat_num].triangle_index_list += surface;	// Attach a new triangle to the polygon list
}


//****************************************************************************
//	Returns the angle between the triangle legs at the specified vertex. 
//****************************************************************************
GLfloat Mesh3DS::Angle (const TRIANGLE& triangle, GLint vertex)
{
	GLint a, b;
	if (triangle.vertex(0) == vertex)		{ a = triangle.vertex(1); b = triangle.vertex(2); }
	else if (triangle.vertex(1) == vertex)	{ a = triangle.vertex(0); b = triangle.vertex(2); }
	else if (triangle.vertex(2) == vertex)	{ a = triangle.vertex(1); b = triangle.vertex(0); }
	else return 0.0;

	Vector3D<float> v1(vertex_list[a].pos() - vertex_list[vertex].pos());
	Vector3D<float> v2(vertex_list[b].pos() - vertex_list[vertex].pos());
	return (v1.angle(v2)).rad();
}


//****************************************************************************
//	Updates the surface normal for the triangle, conter clockwise orientation 
//****************************************************************************
bool Mesh3DS::CalculateSurfaceNormal(TRIANGLE& triangle)
{
	Vector3D<float> a((vertex_list[triangle.vertex(1)]).pos()-(vertex_list[triangle.vertex(0)]).pos());
	Vector3D<float> b((vertex_list[triangle.vertex(2)]).pos()-(vertex_list[triangle.vertex(0)]).pos());
	Vector3D<float>	normal = a^b;

	if (normal.length() < 1e-16)       return false;  // Ignore degenerate polygons
	else { triangle.setNormal(normal); return true; }
}



//****************************************************************************
//	Generate the per-vertex normals. Make new vertices if two surfaces are
//	more than a threshold angle apart. Thus, a cube would have six surface
//	normals but if the threshold was 90 degrees, it would have 24 
//	vertex/normal pairs.
//****************************************************************************
void Mesh3DS::CalculateVertexNormals()
{
	for (int i=0; i < vertex_list.size(); i++)			// Calculate a normal for every vertex
	{
		Vector3D<float>  n(0.0f,0.0f,0.0f);
		bool new_vertex_already_made = false;
		if	(!vertex_list[i].isEmpty())
		{
			TRIANGLE& first_surface = *(vertex_list[i][0]);	// Get first TRIANGLE Object
			n += Angle(first_surface, i)*first_surface.normal();
			for (int j=1; j < vertex_list[i].size(); j++)
			{
				TRIANGLE& current_surface =  *(vertex_list[i][j]);
				if	(((first_surface.normal())*(current_surface.normal())) <= cos_threshold)
				{
					if (!new_vertex_already_made)	// Put this triangle at another vertex
					{								// Make a copy of this vertex at the end of the vertex list
						new_vertex_already_made	= true;
						vertex_list += vertex_list[i];
						vertex_list.last().resetSize();
						vertex_list.last() += current_surface;
					}
						// Change the vertex indices of this current surface so that we point
						// to the new vertex instead of the one we were originally pointing to
					if (current_surface.vertex(0) == i) current_surface.setIndex(0, vertex_list.size()-1);
					if (current_surface.vertex(1) == i) current_surface.setIndex(1, vertex_list.size()-1);
					if (current_surface.vertex(2) == i) current_surface.setIndex(2, vertex_list.size()-1);
				}
				else
					n += Angle(current_surface, i) * current_surface.normal();
			}
			n.normalize();
		} // end if
		
		vertex_list[i].setNormal(n);	// Update the normal for this vertex
	} // end for each vertex
}


//****************************************************************
// creates the displaylists, one for each polygon in the poly_list
//****************************************************************
void Mesh3DS::CompileDisplayList()
{
	int	i, j, k;

	_num_display_list	= poly_list.size();
	_display_list		= glGenLists(_num_display_list);

	for(i=0; i < poly_list.size(); i++)
	{
		glNewList(_display_list+i, GL_COMPILE);
			glBegin(GL_TRIANGLES);
			for (j=0; j < poly_list[i].triangle_index_list.size(); j++)
				for(k=0; k<3; k++)
				{
					glTexCoord(texture_coord(i, j, k)); 
					glVertex(vertex(i, j, k));
				}
			glEnd();
		glEndList();
	}
}


//***************************************************************
// Convert the data read from the file to the Mesh3DS format
//***************************************************************
void Mesh3DS::Convert_to_OpenGL ()
{
	int				i, j, k, current;
	VERTEX			v;

	poly_list.setMax(_default_material_warning+MScene->material.size());
	if	(_default_material_warning) // If the model has default materials in it, substitute an snow material
	{		
		MaterialObject  m;
		m.setName("WG_DEFAULT");
		AddMaterial(m);
	}

	for (i=0; i < MScene->material.size(); i++)	// Feed each material and its properties to OpenGL
		AddMaterial(MScene->material[i]);

	for (k=i=0; i < MScene->meshobjlist.size(); i++)
		for (j=0; j < MScene->meshobjlist[i]->vertlist.size(); j++) k++;
	vertex_list.setMax(k+50);

	for (current=i=0; i < MScene->meshobjlist.size(); i++)
	{
		for (j=0; j < MScene->meshobjlist[i]->vertlist.size(); j++)
		{
			v.pos() = MScene->meshobjlist[i]->vertlist(j);		
			swap(v.pos()[1],v.pos()[2]);			// Swap y and z as 3DS MAX treats Y-axis as front-to-back
			if	((MScene->meshobjlist[i]->maplist.size() > 0) && (j < MScene->meshobjlist[i]->maplist.size()))
				v.setTextureCoord(MScene->meshobjlist[i]->maplist(j)); 
			else
				v.setTextureCoord(Point2D<float>(0.0,0.0));
			vertex_list += v; 	
		}
		for (j=0; j < MScene->meshobjlist[i]->bindings.size(); j++)
			for	(k=0; k < MScene->meshobjlist[i]->bindings[j].faces.size(); k++)
				AddPolygon(Point3D<GLuint>(				// Flip polygon orientation by using points (0,2,1)
				 MScene->meshobjlist[i]->facelist[MScene->meshobjlist[i]->bindings[j].faces[k]].p[0]+current,
				 MScene->meshobjlist[i]->facelist[MScene->meshobjlist[i]->bindings[j].faces[k]].p[2]+current,
				 MScene->meshobjlist[i]->facelist[MScene->meshobjlist[i]->bindings[j].faces[k]].p[1]+current),
				 MScene->meshobjlist[i]->bindings[j].name);
		current += MScene->meshobjlist[i]->vertlist.size();
	}
	FinishModel();
}


//****************************************************************************
//	Performs some model clean-up that can only be done once we have the entire model.
//	Note that this code finds the vertex normals by first calculating 
//  the angle that each polygon shares with that vertex.
//	It then averages all the normals together, each one weighted on 
//	how much of an angle they contribute to the vertex.
//	Also, if we use a threshold for normals (other than 90 degress) so that surface edges are preserved,
//	we have to add a new vertex to the list with the same X,Y,Z coordinates but different normal coordinates.
//	Thus, edge preservation may look good, but it's expensive.
//****************************************************************************
void Mesh3DS::FinishModel()
{
		LinkVerticesToPolygons();
		CalculateVertexNormals();
		CompileDisplayList();
}



//********************************************************
//	The function for reading all polygons completely
//********************************************************
bool Mesh3DS::HRead3dsScene(void *ptr, int size) 
{
  	InFile	= (FILE *) ptr;
	short	id;
    int		len, pc = Dgetpos();

    Dread	(&id,	sizeof(id));
    Dread	(&len,	sizeof(len));
	if((int) id != CHUNK_MAIN)
	{
	   	printf("Not in 3DS format.\n");
    	return false;
    }
    ReadMainBlocks (pc+len);
  	return true;
}


//*********************************************************
//	Build links from the vertices back to the polygon lists
//*********************************************************
void Mesh3DS::LinkVerticesToPolygons()
{
	for (int i=0; i < poly_list.size(); i++)	// For each material in the poly_list:
	{
		for (int j=0; j < poly_list[i].triangle_index_list.size(); j++)	// For each triangle:
		{
			// Insert a reference to this triangle into a list for each of the three vertices it touches.
			vertex_list[poly_list[i].triangle_index_list[j].vertex(0)] += poly_list[i].triangle_index_list[j];
			vertex_list[poly_list[i].triangle_index_list[j].vertex(1)] += poly_list[i].triangle_index_list[j];
			vertex_list[poly_list[i].triangle_index_list[j].vertex(2)] += poly_list[i].triangle_index_list[j];
		}
	}
}


//*******************************************************************
//	The external function for reading and initializing the 3Ds object
//*******************************************************************
bool Mesh3DS::parse3DS (char* filename) 
{
  	long		size;
  	FILE		*inf;
  
  	if	(!(inf = fopen((char *)filename, "rb")))	// Open & prepare input file
	{
	    printf	("Failed to open %s\n", filename);
    	return	false;
	}
  	if	(fseek (inf, 0, SEEK_END))
	{
	    printf	("Error seeking %s\n", filename);
    	fclose	(inf);
    	return	false;
	}
  	if	((size = ftell (inf)) == -1L)
	{
	    printf	("Error sizing %s\n", filename);
    	fclose	(inf);
    	return	false;
	}
  	rewind	(inf);
	if	(!HRead3dsScene(inf, size))
	{
	    printf	("Failed to parse %s\n", filename);
    	fclose	(inf);
    	return	false;
	}
  	fclose(inf);
	Convert_to_OpenGL (); // Convert the internal data structures to OpenGL's structures
	delete MScene;
  	return	true;
}


void Mesh3DS::ReadDefaultMaterial(H3dsMeshObj& meshobj, short nf)
{
	meshobj.bindings += H3dsMatList();
	H3dsMatList&	binding_entry = meshobj.bindings.last();

	binding_entry.faces.setSize(nf);
	for	(int n=0; n < nf; n++)
		binding_entry.faces[n] = n;	// Bind the faces to the material
}



//*********************************************************
//	Reads in the face list for the current mesh object. 
//	Each face is a set of 3 indexes into the mesh object's 
//	vertex list and a set of options.
//*********************************************************
short Mesh3DS::ReadFaceList(H3dsMeshObj& meshobj)
{
	short		nv;
	Dread(&nv, sizeof (nv));
	int current=meshobj.facelist.size();
	meshobj.facelist.setSize(current+nv);

	for (int n=0; n < nv; n++)
	{
		Dread(&(meshobj.facelist[current+n]),sizeof(meshobj.facelist[current+n]));
		//	0x07 means that the AB, BC, and CA edges are all wound 
		//	in the reverse order, so swap the last two vertices 
		if (meshobj.facelist[current+n].flags == 0x07) 
			swap(meshobj.facelist[current+n].p[1],meshobj.facelist[current+n].p[2]);
	}
	return (nv);
}



void Mesh3DS::ReadDoublesidedChunk (bool& material_doublesided)
{
	material_doublesided = true;
} 


//************************************************************
//	Assigns the material in this chunk to the faces specified. 
//	Each material has a list of indexes into the face array 
//	to mark which faces use which material.
//************************************************************
void Mesh3DS::ReadFaceMaterial (H3dsMeshObj& meshobj)
{
	unsigned short	nf;
	H3dsMatList		binding_entry;

	ReadName(binding_entry.name, MAXNAMESIZE);			// Read first the name,
	Dread(&nf, sizeof(nf));								// then the number of faces,
	binding_entry.faces.setSize(nf);
	Dread(binding_entry.faces.ptr(),nf*sizeof(short));	// and at last the faces (index list)

	meshobj.bindings += binding_entry;
} 


//**************************************************
// Reads in texture file name for a polygon object
//**************************************************
void Mesh3DS::ReadTextureChunk(int p, GM_String& texture)
{
  	unsigned short	id;
  	int				len, pc;

  	while ((pc = Dgetpos()) < p)
	{
	    Dread	(&id,	sizeof (id));
    	Dread	(&len,	sizeof (len));
    	switch	((unsigned int) id) {
      		case CHUNK_MAPFILENAME: 
				ReadName(texture, MAXNAMESIZE);
				break;
      		default: 
				Dsetpos(pc+len);
    	} // endof switch
  	} // endof while
}  


//*****************************
// Reads in a color assignment
//*****************************
void Mesh3DS::ReadShadingChunk(int p, GL_Color& rgb)
{
  	short	id;
  	int		len, pc;

  	while ((pc = Dgetpos()) < p)
	{
	    Dread	(&id,	sizeof (id));
    	Dread	(&len,	sizeof (len));
    	switch	((unsigned int)id)
		{
      		case CHUNK_RGB1:
				ReadRGBFloat(rgb);
				break;
      		case CHUNK_RGB2:
				ReadRGBByte (rgb);
				break;
      		default:
				Dsetpos(pc+len);
    	}
  	}
}


//*************************************
// Reads in mesh objects and materials
//*************************************
void Mesh3DS::ReadObjMeshBlocks(int p)
{
  	unsigned short id;
  	int len, pc;

  	while ((pc = Dgetpos()) < p)
	{
	    Dread	(&id,	sizeof (id));
    	Dread	(&len,	sizeof (len));
    	switch	((unsigned int)id)
		{
      		case CHUNK_MATERIAL: 
				ReadMaterialBlocks (pc+len); 
				break;
		    case CHUNK_OBJBLOCK: 
				ReadObjBlocks (pc+len); 
				break;
			default: 
				Dsetpos (pc+len);
		}
	}
}


//****************************************************************
//  Reads in object blocks, which are cameras, meshes, or lights;
//  this code only supports meshes and ignores the rest
//****************************************************************
void Mesh3DS::ReadObjBlocks(int p)
{
  	short		id;
  	int			len, pc;
  	GM_String	name;
	
	ReadName(name, MAXNAMESIZE); // The object name is always the first item 

  	while	((pc = Dgetpos()) < p)
	{
	    Dread	(&id, sizeof(id));
    	Dread	(&len, sizeof(len)); 
		switch	((int)id)
		{
      		case CHUNK_TRIMESH:
				ReadTriMeshBlocks(pc+len, name);
				break;
		//  case CHUNK_LIGHT:
	  	//	case CHUNK_CAMERA:
      		default:
				Dsetpos(pc+len);	
		}	
	}  	
	Dsetpos(p);
}



//*********************************************
//  Reads in a material objects; the object are  
//  direct put into a GM_lib::MaterialObject 
//*********************************************
void Mesh3DS::ReadMaterialBlocks (int p)
{
  	unsigned short	id;
  	int				len, pc;
	MaterialObject	mat;
	GM_String		name;
	GL_Color		col;
	bool			ds;
	float			tr;
	mat.set(20.0f);	// until we find out how to read shininess.

  	while		((pc = Dgetpos()) < p)
	{
	    Dread	(&id,	sizeof(id));
    	Dread	(&len,	sizeof(len));
    	switch	((unsigned int)id)
		{
    	case CHUNK_MATNAME: 
			ReadName				(name, MAXNAMESIZE); 
			mat.setName(name);
      		break;
    	case CHUNK_AMBIENT: 
      		ReadShadingChunk		(pc+len, col);
			mat.setAmb(col);
      		break;
    	case CHUNK_DIFFUSE: 
      		ReadShadingChunk		(pc+len, col);
			mat.setDif(col);
      		break;
    	case CHUNK_SPECULAR:
      		ReadShadingChunk		(pc+len, col);
			mat.setSpc(col);
      		break;
    	case CHUNK_TEXTURE:
      		ReadTextureChunk		(pc+len, name);
			mat.setTexture(name);
      		break;
    	case CHUNK_TRANSPARENCY:
      		ReadTransparencyChunk	(pc+len, tr);
			mat.setTransparancy(tr);
      		break;
    	case CHUNK_DOUBLESIDED:
      		ReadDoublesidedChunk	(ds);
			mat.setDoubleSided(ds);
      		break;
    	default: 
      		Dsetpos					(pc+len);
      		break;
    	}
  	}
  	MScene->material += mat;
}


//*********************************************
//  Reads in objects; this would be the start 
//  of where keyframer support would be added
//*********************************************
void Mesh3DS::ReadMainBlocks(int p)
{
  	short	id;
  	int		len, pc;
  
  	while	((pc = Dgetpos()) < p)
	{
	    Dread	(&id, sizeof(id));
    	Dread	(&len, sizeof(len));
    	switch	((unsigned int)id)
		{
      	case CHUNK_OBJMESH: 
			ReadObjMeshBlocks(pc+len);
			break;
      	case CHUNK_KEYFRAMER:
      	default:
			Dsetpos (pc+len);
    	}
	}
}


//*********************************************
//  Reads the transparancy in percent
//	and transelt it to a part: [0.0, 1.0]
//*********************************************
void Mesh3DS::ReadTransparencyChunk (int p, float& transparency)
{
  	short	id, amount;
  	int		len, pc;

  	while((pc=Dgetpos()) < p)
	{
    	Dread(&id, sizeof(id));
    	Dread(&len, sizeof(len));
    	switch((unsigned int)id)
		{
      		case CHUNK_AMOUNT:
				Dread(&amount, sizeof(short));
				break;
      		default: 
				Dsetpos (pc+len);
    	}
  	}
  	transparency = (float)(amount/100.0);
	if (transparency < .01)  transparency = 1.0;
}  


//******************************************************
//	Eats a triangle mesh, which is a combination of:
//	vertices, faces, and bindings of material to faces. 
//******************************************************
void Mesh3DS::ReadTriMeshBlocks(int p, const GM_String& name)
{ 
	short	id;
	short	fl;
	int		len, pc;
	bool	found_facematial= false;
	H3dsMeshObj* meshobj = new H3dsMeshObj();

	meshobj->name = name;

	while ((pc = Dgetpos()) < p)
	{
		Dread	(&id,	sizeof (id));
		Dread	(&len,	sizeof (len));
		switch	((int)id)
		{
		case CHUNK_VERTLIST:
			ReadVertList		(*meshobj);
			break;
		case CHUNK_FACELIST:
			fl = ReadFaceList	(*meshobj);
			break;
		case CHUNK_MAPLIST:
			ReadMapList			(*meshobj);
			break;
		case CHUNK_TRMATRIX:
			ReadTraMatrix		(*meshobj);
			break;
		case CHUNK_FACEMAT:
			ReadFaceMaterial	(*meshobj);
			found_facematial	= true;
			break;
		// case CHUNK_SMOOLIST:
		default:
			Dsetpos (pc+len);
		}
	}
	if (!found_facematial)
	{
		_default_material_warning = 1;
		ReadDefaultMaterial(*meshobj, fl);
	}
	MScene->meshobjlist += meshobj;
}


//*************************************************************
//	Reads in the list of vertices for the current mesh object
//*************************************************************
void Mesh3DS::ReadVertList(H3dsMeshObj& meshobj)
{
	short		nv;
	Dread(&nv, sizeof(nv));
	int current = meshobj.vertlist.size();
	meshobj.vertlist.setSize(current+nv);
	Dread(&(meshobj.vertlist[current]),3*nv*sizeof(float));
}



void Mesh3DS::zoom(GLfloat _zoom)
{
	for (int i=0; i < vertex_list.size(); i++)
		vertex_list[i].pos() *= _zoom;
	CompileDisplayList();
}


Sphere<float,3> Mesh3DS::sphere()
{
	Sphere<float,3> sp(vertex_list[0].p);
	for (int i=1; i < vertex_list.size(); i++)
		sp += vertex_list[i].p;
	return sp;
}
