#ifndef __MESH3DS_H_
#define __MESH3DS_H_

#include <GM_DisplayObject.h>
#include <GM_Texture.h>
#include <GM_Material.h>
#include <GM_LX_Array.h>
using namespace GM_lib;

#include "Load3DSfile.h"



class TRIANGLE	// A TRIANGLE with a normal attached
{
	friend class Mesh3DS;

	Point3D<GLuint>		p;		// Index in vertex-list to the 3 points in the triangle
	UnitVector3D<float> n;		// Normal Vector to the triangle

	void setIndex(const Point3D<GLuint>& pp) { p = pp;}
	void setIndex(int i, int index) { p[i] = index; }
	void setNormal(const Vector3D<float>& nn) { n = nn;}

	const UnitVector3D<float>& normal()  const { return n; }
	const Point3D<GLuint>& vertices()  const { return p; }
	int	  vertex(int i)  const { return p(i); }

public:
	TRIANGLE(const Point3D<GLuint>& pp = Point3D<GLuint>(GLuint(0)),
			const Vector3D<float>& nn = Vector3D<float>())
					{ p = pp; n = nn;}

	TRIANGLE(const TRIANGLE& s) 
					{ p = s.p; n = s.n; }
};

class VERTEX  
{
	friend class Mesh3DS;

	Point2D<float>		t;			// Texture Coordinates for this vertex
	Vector3D<float>		n;			// Normal Vector for this vertex						
	Point3D<float>		p;			// Position for this vertex
	Array<TRIANGLE*>	list;		// List of all triangles around this vertex

	Arrow<float,3>	vertex() const { return Arrow<float,3>(p,n); }
	Point3D<float>& pos() { return p; }

	void		setNormal(const Vector3D<float>& nn)		{ n=nn; }
	void		setTextureCoord(const Point2D<float>& tt)	{ t = tt; }
	const		Point2D<float>& texCoord() const			{ return t; }

	void		operator+=(TRIANGLE& s)		{ list += &s; }
	TRIANGLE*	operator[](int i)			{ return list[i]; }
	int			size()						{ return list.size(); }
	bool		isEmpty()					{ return list.isEmpty(); }
	void		resetSize()					{ list.resetSize(); }

public:

	VERTEX( const Point2D<float>& texcoords_ = Point2D<float>(0,0), 
			const Vector3D<float>& normal_ = Vector3D<float>(0.0), 
			const Point3D<float>& pos_ = Point3D<float>(0.0))
					{ t   = texcoords_; n	= normal_; p	= pos_; }

	VERTEX( const VERTEX& v) 
					{ t = v.t; n = v.n; p = v.p; list = v.list; }
};


class POLYGONLIST 
{
	friend class Mesh3DS;

	MaterialObject		material;
	LX_Array<TRIANGLE>	triangle_index_list; // Array of TRIANGLE objects which have indexes into vertex_list[]

	const GM_String& name() const							{ return material.name(); }
	void			 setMaterialName(const GM_String& mn)	{ material.setName(mn); }
	void			 setTextureName(const GM_String& tn)	{ material.setTexture(tn); }

	MaterialObject& mat()									{ return material; }

public:
	POLYGONLIST (){}
	POLYGONLIST(const POLYGONLIST& p)
	{ 
	  material = p.material;
	  triangle_index_list = p.triangle_index_list;
	}

	~POLYGONLIST()
	{ material.deleteTexture();}

	bool operator<(const POLYGONLIST& a) const { return true; } // dummy
};


class	Mesh3DS 							// 3D Studio MAX mesh loader class
{
public:
	Mesh3DS(const char*	filename,			// Class constructor
		    int			threshold, 
			bool		force_2sided);	

	~Mesh3DS();								// Class destructor

	bool		parse3DS(char *);			// Load data structures, display lists, etc.
	void		zoom(GLfloat);				// Scale mesh & recompute vertex/surface normals
	void		localDisplay();				// Display mesh onto current Rendering Context
	void		localSelect();				// Display mesh onto current Rendering Context

	Sphere<float,3> sphere();

private:	
	
	Array<POLYGONLIST>	poly_list;			// Object array of POLGONLIST objects
	LX_Array<VERTEX>	vertex_list;		// Object array of VERTEX objects

	int			_display_list;
	int			_num_display_list;
	int			_default_material_warning;

	GLfloat		cos_threshold;
	bool		force2sided;				// Flag to force 2-Sided materials

	FILE		*InFile;
	H3dsScene	*MScene;


	Point2D<float> texture_coord(int poly, int tri, int vertex)
		{ return vertex_list[poly_list[poly].triangle_index_list[tri].vertex(vertex)].texCoord(); }

	Arrow<float,3> vertex(int poly, int tri, int vertex)
		{ return vertex_list[poly_list[poly].triangle_index_list[tri].vertex(vertex)].vertex(); }


	int			Dgetpos					(void);
	void		Dread					(void *dest, int len);
	void		Dsetpos					(int pos);

	int			FindMaterialIndex		(const GM_String& name);
	short		ReadFaceList			(H3dsMeshObj& meshobj);
	void		Convert_to_OpenGL		();
	void		ReadVertList			(H3dsMeshObj& meshobj);
	void		ReadMapList				(H3dsMeshObj& meshobj);
	void		ReadTraMatrix			(H3dsMeshObj& meshobj);
	void		ReadFaceMaterial		(H3dsMeshObj& meshobj);
	void		ReadDefaultMaterial		(H3dsMeshObj& meshobj, short nf);
	void		ReadTriMeshBlocks		(int p, const GM_String& name);
	void		ReadObjBlocks			(int p);
	void		ReadRGBFloat			(GL_Color& rgb);
	void		ReadRGBByte				(GL_Color& rgb);
	void		ReadShadingChunk		(int p, GL_Color& rgb);
	void		ReadTextureChunk		(int p, GM_String& texture);
	void		ReadDoublesidedChunk	(bool& material_doublesided);
	void		ReadAmountChunk			(short& amount);
	void		ReadTransparencyChunk	(int p, float& transparency);
	void		ReadMaterialBlocks		(int p);
	void		ReadObjMeshBlocks		(int p);
	void		ReadMainBlocks			(int p);
	void		AddPolygon				(const Point3D<GLuint>& tri, const GM_String& MaterialName);
//	void		AddMaterial				(const H3dsMat& );
	void		AddMaterial				(const MaterialObject&);
	void		FinishModel				();
	void		LinkVerticesToPolygons	();
	void		CompileDisplayList		();
	void		CalculateVertexNormals	();
	bool		CalculateSurfaceNormal	(TRIANGLE& ThisTriangle);
	bool		HRead3dsScene			(void *ptr, int size);
	GLfloat		Angle					(const TRIANGLE& triangle, GLint vertex);
	void		ReadName				(GM_String& name, int size);
};



//****************************************************
// Overloaded basic "File reading tools"( 3 functions)
//****************************************************

inline
void Mesh3DS::Dread (void *dest, int len)	
{
  	if (fread(dest, len, 1, InFile) != 1)	printf("Error reading file\n");
}

inline
int Mesh3DS::Dgetpos(void)
{
  	long pos;	
	if ((pos=ftell(InFile)) == -1L)			printf("Error getting fileposition\n");
  	return (int)pos;
}

inline
void Mesh3DS::Dsetpos(int pos)
{
  	if (fseek(InFile, (long)pos, SEEK_SET) != 0) printf("Error moving filepointer\n");
}


//***************************************************
// Reads in an RGB Color triplet (without alpa value)
//***************************************************

inline
void Mesh3DS::ReadRGBByte(GL_Color& rgb)	// byte format
{
	unsigned char c[3];
	Dread(c, 3*sizeof(unsigned char));
    rgb.set(c[0],c[1],c[2]);
}

inline
void Mesh3DS::ReadRGBFloat( GL_Color & rgb)	// float format
{
	float c[3];
	Dread(c, 3*sizeof(float));
	rgb.set(c[0],c[1],c[2]);
}


//***********************************************************************
//	Reads in the mapping list/texture coords. for the current mesh object
//***********************************************************************
inline
void Mesh3DS::ReadMapList( H3dsMeshObj& meshobj)
{
	short			nv;
	Dread(&nv, sizeof(short));
	meshobj.maplist.setSize(nv);
	Dread(meshobj.maplist[0].ptr(),nv*sizeof(Point2D<float>));
}


//*************************************************************
// Reads in the translation matrix for the current mesh object.
// I'm not sure when this is used in a 3DS model.
//**************************************************************
inline 
void Mesh3DS::ReadTraMatrix (H3dsMeshObj& meshobj)
{
	Vector3D<float> m[4];
	Dread (m[0].ptr(), 4*sizeof(Vector3D<float>));
	meshobj.mat.basisChange(m[0],m[1],m[2],m[3]);
}


//*********************************************************
// Search for a material with a given name in the poly_list.
//*********************************************************
inline 
int Mesh3DS::FindMaterialIndex(const GM_String& name)
{
	for (int n=0; n < poly_list.size(); n++)
		if (poly_list[n].material.is(name)) return n;
	return	-1;					// Error: Can't find material
}


//***************************************************
//  Reads in the name of an object ******************
//***************************************************
inline
void Mesh3DS::ReadName(GM_String& name, int size)
{
	char n[256];
	Dread(n, 1);
	for(int i=1; i < size && n[i-1] != '\0'; i++)
		Dread(n+i, 1);
	n[i-1]='\0';
	name = n;
}

#endif