#ifndef _RANDOM_H_
#define _RANDOM_H_

#include <ctime>

class AbstractRandomGenerator
{
	public:
		
		virtual long double operator()() = 0;
		
		virtual long long Rand() = 0;
		
		virtual long long Rand( long long max ) = 0;
		
		virtual void Seed( long long seed ) = 0;
		
		virtual long double Expectation() const = 0;
		
		virtual long double Variance() const = 0;
		
		virtual long double Minimum() const = 0;
		
		virtual long double Maximum() const = 0;


};

class Random : public AbstractRandomGenerator
{
	public:

		Random( const long double& variance = 0.0, const long double& expectation = 0.0 ) : variance( variance ), expectation( expectation )
		{
			Seed( (unsigned int) time( 0 ) );

			minimum = - sqrt( 3 * variance ) + expectation;
			maximum = sqrt( 3 * variance ) + expectation;
		}

		Random( const long double& minimum, const long double& maximum, int dummy ) : minimum( minimum ), maximum( maximum )
		{
			Seed( (unsigned int) time( 0 ) );

			expectation = ( maximum + minimum ) / 2.0;
			variance = ( maximum - expectation ) * ( maximum - expectation ) / 3.0;
		}

		virtual ~Random()
		{
		}

		Random( const Random& right ) : 
			minimum( right.minimum ), 
			maximum( right.maximum ), 
			expectation( right.expectation ), 
			variance( right.variance )
		{
		}

		Random& operator=( const Random& right )
		{
			if( this != &right )
			{
				minimum = right.minimum;
				maximum = right.maximum;
				
				expectation = right.expectation;
				variance = right.variance;
			}

			return *this;
		}


		virtual long double operator()()
		{
			return minimum + (long double) ( ( ( maximum - minimum ) * Rand() ) / ( (long double) M - 1.0 ) );
		}

		virtual long long Rand()
		{
			current = ( A * current + B ) % M;
			return current;
		}

		virtual long long Rand( long long max )
		{
			return ( (long double) Rand() * max ) / (long double) M;
		}

		virtual void Seed( long long seed )
		{
			current = seed;
		}

		virtual long double Expectation() const
		{
			return expectation;
		}
		
		virtual long double Variance() const
		{
			return variance;
		}

		virtual long double Minimum() const
		{
			return minimum;
		}

		virtual long double Maximum() const
		{
			return maximum;
		}



	private:

		long double minimum;
		long double maximum;

		long double expectation;
		long double variance;

		long long current;

		static long long A;
		static long long B;
		static long long M;

};

long long Random::A = 1664525;
long long Random::B = 1013904223;
long long Random::M = 4294967296;

#endif // _RANDOM_H_
