#ifndef _COSTFUNCTION_H_
#define _COSTFUNCTION_H_

#include "GM_DVector.h"
#include <GM_DArrayN.h>
#include <Definitions.h>

#include "Lifting.h"
#include "WaveFLWT.h"

using namespace GM_lib;

static double epsilon, p;
double (*infocost)(DArrayN<double,1>, const int, const int);

/************************/
/* Hedge data structure */
/************************/
template<class T>
class Hedge 
{
	public:
		Hedge( int blocks, T** contents, DVector<int> levels, void *tag );
		~Hedge() {}

		T** contents;   // an array of data structures containing coefficients in the adapted basis 
		void *tag;         // a pointer to data structure containing extra information about the decomposition 

		DVector<int> levels;   // the sequence of levels in encounter order 
		int blocks; // the number of elements in the levels array, also in the contents array 
};

template<class T> 
Hedge<T>:: Hedge (int blocks, T** contents, DVector<int> levels, void *tag )
{
	this->blocks = blocks;
	this->tag = tag;
	this->levels=levels;
	this->contents = contents;
}


/****************/
/* Binary tree */
/****************/
template<class T>
class BTN
{
	public:
		BTN(void *content, BTN *left, BTN *right, void *tag );
		~BTN() {}

		void *tag, //a pointer to a data structure containing information about this node 
         *content;           
         BTN *left,*right;    // pointer to the left and right descendent of this node   

		 BTN* makebtnt (int level);
		 BTN *btnt2btn ( BTN *root, int level, int block );
		 BTN *btn2branch ( BTN *self, int level, int block );
};

template<class T> 
BTN<T>:: BTN (void *content, BTN *left, BTN *right, void *tag )
{
	BTN *node = (BTN *)malloc((unsigned)sizeof(BTN));
    node->tag     = (void *)tag;
    node->content = (void *)content;
    node->left    = (BTN *)left;
    node->right   = (BTN *)right;
}
 
//Build a completely empty tree down to a specified level with a recursive function
template<class T> 
BTN<T>* BTN<T>:: makebtnt (int level )
{
    BTN *root = makebtn(NULL, NULL, NULL, NULL);

    if ( level > 0 ) 
	{
        root->left  = makebtnt( level-1 );
        root->right = makebtnt( level-1 );
    }
    return root;
}


// Get a pointer to a particular node in a BTN tree.
template<class T> 
BTN<T>* BTN<T>::btnt2btn ( BTN *root, int level, int block )
{
    BTN *node;
    if ( root == NULL || level == 0 )
        node = root;
    else {
        if ( EVEN(block) )
            node = btnt2btn( root->left, level-1, block>>1 );
        else
            node = btnt2btn( root->right, level-1, (block-1)>>1 );
    }

    return node;
}

//Link together one branch between the root and a target specified by its level and block indices
template<class T> 
BTN<T>* BTN<T>::btn2branch ( BTN *self, int level, int block )
{
    if ( level > 0 ) {
        if ( EVEN(block) ) {
            if ( self->left == NULL )
                self->left = makebtn( NULL, NULL, NULL, NULL );
            self = btn2branch( self->left, level-1, block>>1 );
        } else {
            if ( self->right == NULL )
                self->right = makebtn( NULL, NULL, NULL, NULL );
            self = btn2branch( self->right, level-1, (block-1)>>1 );
        }
    }
    return self;
}


/******************/
/* Cost Functions */
/******************/
template<class T> T
bbwp2 ( Hedge<T> *graph, DArrayN<T,2> &in, const int x, const int y, const N, const nTilde,
                                 const int s, const int l )
{
	DVector<int> size;
	in.getSize(size);
	
	T cost = 0;
	T *out;

	for(int i=0 ; i<size.dim() ; i++)
		for (int j=0 ; j<size[i] ; j++ ) 
			cost += (T)(*infocost)( in[j], 0, size[i]-1);

    if ( s<l ) {
        int block = graph->blocks;
        Array<DArrayN<T,2> > k = DSFLWT2D<T>( in, N, nTilde );   // separate them 

         T kcost = NULL;
		 for (int j=0 ; j<4 ; j++ )
            kcost += bbwp2<T>( graph, k[j], size[0]>>1, size[1]>>1, N, nTilde, s+1, l );

       if ( kcost < cost )  cost = kcost;
         else { 
            in = IFLWTP2D<T>(in, N, nTilde );
            graph->levels[block] = s;
            out = graph->contents[block];
           
			for(int i=0, t=0; i<size[0]; i++)
				for(int j=0; j<size[1]; j++, t++)
					out[t] = in[i][j];

            graph->blocks = block + 1;
            graph->contents[graph->blocks] = out + size[0]*size[1];
        }
    } else {
        graph->levels[graph->blocks] = l;
        out = graph->contents[graph->blocks];
        
		for(int i=0, t=0; i<size[0]; i++)
			for(int j=0; j<size[1]; j++, t++)
				out[t] = in[i][j];
        
        graph->blocks++;
        graph->contents[graph->blocks] = out + size[0]*size[1];
    }

    return cost;
} 

template<class T> T
bbwp1 ( Hedge<T> *graph, DArrayN<T,1> &in, const int x, const N, const nTilde,
                                 const int s, const int l )
{
	T cost = 0;
	T *out;
	
	cost = (T)(*infocost)( in, 0, x-1);

    if ( s<l ) 
	 {
       int block = graph->blocks;
        Array<DArrayN<T,1> > k = DSFLWT1D<T>( in, N, nTilde );
        T kcost = NULL;
			   
        for (int j=0 ; j<2 ; j++ )
           kcost += bbwp1<T>(graph, k[j], x>>1, N, nTilde, s+1, l);
		

        if ( kcost < cost ) cost = kcost;
         else { 
            in = IFLWTP1D<T>( in, N, nTilde );
            graph->levels[block] = s;  // set the lvl for this block 
			out = graph->contents[block];

            for (int i=0 ; i<x ; i++ )
                out[i] = in[i];
            graph->blocks = block + 1; // there is now one more block 
            //set contents[] so it points to next block 
			graph->contents[graph->blocks] =  out+x;
       }
    } else {
        graph->levels[graph->blocks] = l;

		out = graph->contents[graph->blocks];
		
        for (int i=0 ; i<x ; i++ )
            out[i] = in[i];

		graph->blocks++;
		graph->contents[graph->blocks]=  out+x;
    }
    return cost;
} 


//*****************************
// Aditive cost functions 
//*****************************

// Thresh function: measures the cost of a sequence of coefficients by the "number exceeding a threshold". 
template<class T>
T thresh (DArrayN<T,1> data, const int least, const int final)
{
    T cost = 0;

    // Threshold information cost functional 
    for (int k=least; k<final ; k++ )
       if ( abs(data[k]) > epsilon )
            cost += (T)1;

    return cost;
}

// lpnormp functions calculates the concentration in l^p by summing the pth powers of their coefficients, for 0<p<2.
template<class T> T
l1norm (DArrayN<T,1> data, const int least, const int final )
{
    T cost = 0;

    // l^1 information cost functional 
    for (int k=least ; k<final ; k++ )
        cost += abs(data[k]);

    return cost;
}

template<class T> T
lpnormp (DArrayN<T,1> data, const int least, const int final)
{
    T cost = 0;
    
    // l^p information cost functional 
    for (int k=least ; k<final ; k++ ) {
        if ( abs(data[k]) > (T)0 )
            cost += (T)exp( p*(T)log((T)abs(data[k])) );
    }
    return cost;
}

// Entropy 
template<class T> T
ml2logl2 (DArrayN<T,1> data, const int least, const int final )
{
     T cost = 0;
  
    // -(l^2)log(l^2) or "entropy" information cost functional 
    for (int k=least ; k<final ; k++ ) {
        T dataSQR = data[k]*data[k];
        if ( dataSQR > (T)0 )
            cost -= dataSQR * (T)log(dataSQR);
    }
    return cost;
}


// Distortion function finds the amount of distorion of the given area using the Mean Squared Error of the signal.
template<class T> T
mse (DArrayN<T,1> data, const int least, const int final )
{
    T cost = 0;

    for (int i=least ; i<final ; i++ )
        cost += data[i] * data[i];
    cost /= (T)(final-least+1);

    return cost;
}


//*****************************
// Not Additive Cost functions 
//*****************************
//Gauss-Markov 
template<class T> T
logl2 (DArrayN<T,1> data, const int least, const int final )
{
    T cost = 0;
   
    // log(l^2) or Gauss-Markov information cost functional 
    for (int k=least ; k<final ; k++ )
        cost += (T)log( (T)(data[k] * data[k]) );

    return cost;
}

#endif