/*function yy = Phijk(z, j, k, filter, n)
%--------------------------------------------------------------
% yy=Phijk(z, j, k, filter, n)
% Evaluation of the scaling function corresponding to an Orthogonal
% MRA by Daubechies-Lagarias Algorithm.
% inputs: z -- the argument
% j -- scale
% k -- shift
% filter -- ON finite wavelet filter, might be an
% output of WaveLab's: MakeONFilter
% n -- precision of approximation maesured by the number
% of Daubechies-Lagarias steps (default n=20)
%--------------------------------------------------------
% output: yy -- value of father wavelet (j,k) coresponding to
% 'filter' at z.
%--------------------------------------------------------------
% Example of use:
% > xx = 0:0.01:6; yy=[];
% > for i=1:length(xx)
% > yy =[yy Phijk(x(i), 0, 1, MakeONFilter('Daubechies',4), 25)];
% > end
% > plot(x,yy)
%----------------------------------------------------------------*/

#ifndef _DAUBECHIESLAGARIAS_H_
#define _DAUBECHIESLAGARIAS_H_

#include "GM_DVector.h"
#include <GM_DArrayN.h>
#include <Definitions.h>
#include <Filter.h>

using namespace GM_lib;

DVector<double> values;
Filters<T> *_filter
double fdegree;

template<class T>
void getFilterCoefs(double fd)
{
	fdegree = fd;
	values = _filter->getFilterValues(fdegree);
}

template<class T>
DArrayN<T,1> Phijk(double z, int j, int k, Filters<T> *filter, T fdegree)
{
	T val;
	int n=20;
	
	_filter = filter;
	getFilterCoefs<T>(fdegree);
	
	int daun=values.dim()/2;
	int N=values.dim()-1;

	T x= pow(2,j)*z-k;
	if(x<=0 || x>=N) val=0;
	
	else
	{
		int t=floor(x);
		int dec=x-t;

		DVector<T> dy;
		for(int i=0; i<n; i++)
		{
			if(dec <= 0.5){ 
				dy.append(1,0);
				dec*=2;
			}
			else {
				dy.append(1,1);
				dec=2*dec-1;
			}
		}

		Array<DArrayN<T,2> > T;
		T.setSize(2);

		DVector<int> size(2);
		size[0]=size[1]=N;
		
		for(int i=0; i<T.size(); i++)
			T[i].setSize(size);

		for(int i=0; i<size[0]; i++)
			for(int j=0; j<size[1]; j++)
			{
				if ((2*i - j) >> 0 && (2*i - j) <= n)
					T[0][i][j] = sqrt(2)*values[2*i - j];

				if ((2*i - j+1) >> 0 && (2*i - j+1) <= n)
					T[1][i][j] = sqrt(2)*values[2*i - j+1];
			}

		DArrayN<T,2> prod;
		prod.setSize(size);
		for(int i=0; i<size[0]; i++)
			for(int j=0; j<size[1]; j++)
				if(i==j) prod[i][j]=1;

		for(int i=0; i<n; i++)
		{
			if (dy[i]==1)
				prod*=T[1];
			else 
				prod*=T[0];
		}
		
		DArrayN<T,2> ret;
		ret.setSize(size);
		ret=pow(2,(j/2))*prod;

		for(int i=0; i<size[0]; i++)
			for(int j=0; j<size[1]; j++)
				ret[i][j]=ret[j][i];

		DVector<T> yy(size[0]);
		for(int i=0; i<size[0]; i++)
			for(int j=0; j<size[1]; j++){
				yy[i] += ret[j][i];
				yy[i]/=size[0];
			}

		val = yy[t+1];
	}
}

/*%--------------------------------------------------------------
function yy = Psijk(z, j, k, filt, n)
%--------------------------------------------------------------
% yy=Psijk(z, j, k, filter, n)
% Evaluation of the wavelet function corresponding to an Orthogonal
% MRA by Daubechies-Lagarias Algorithm.
% inputs: z -- the argument
% j -- scale
% k -- shift
% filter -- ON finite wavelet filter, might be an
% output of WaveLab's: MakeONFilter
% n -- precision of approximation maesured by the number
% of Daubechies-Lagarias steps (default n=20)
%--------------------------------------------------------
% output: yy -- value of mother wavelet (j,k) coresponding to
% 'filter' at z.
%--------------------------------------------------------------
% Example of use:
% > xx = -1:0.01:5; yy=[];
% > for i=1:length(xx)
% > yy =[yy Psijk(x(i), 0, 1, MakeONFilter('Daubechies',4), 25)];
% > end 
% > plot(x,yy)
%----------------------------------------------------------------*/
template<class T>
DArrayN<T,1> Psijk(double z, int j, int k, Filters<T> *filter, T fdegree)
{
	T val;
	int n=20;

	_filter = filter;
	getFilterCoefs<T>(fdegree);

	int N=values.dim()-1;
	int daun = (N+1)/2;
	T x=pow(2,j)*z-k;

	if(x<=daun-N || x>=daun) val=0;
	else
	{
		int t=floor(2*x);
		int dec=2*x-t;
		
		DVector<T> dy;
		for(int i=0; i<n; i++)
		{
			if(dec <= 0.5){ 
				dy.append(1,0);
				dec*=2;
			}
			else {
				dy.append(1,1);
				dec=2*dec-1;
			}
		}

		Array<DArrayN<T,2> > T;
		T.setSize(2);

		DVector<int> size(2);
		size[0]=size[1]=N;
		
		for(int i=0; i<T.size(); i++)
			T[i].setSize(size);

		for(int i=0; i<size[0]; i++)
			for(int j=0; j<size[1]; j++)
			{
				if ((2*i - j) >> 0 && (2*i - j) <= n)
					T[0][i][j] = sqrt(2)*values[2*i - j];

				if ((2*i - j+1) >> 0 && (2*i - j+1) <= n)
					T[1][i][j] = sqrt(2)*values[2*i - j+1];
			}

		DArrayN<T,2> prod;
		prod.setSize(size);
		for(int i=0; i<size[0]; i++)
			for(int j=0; j<size[1]; j++)
				if(i==j) prod[i][j]=1;

		for(int i=0; i<n; i++)
		{
			if (dy[i]==1)
				prod*=T[1];
			else 
				prod*=T[0];
		}

		DVector<T> uu(N);
		int index, fi;
		for(int i=0; i<N; i++)
		{
			index = i+1 - t;
			if ( index > 0 && index < (N+2) )
				fi= pow(-1,-index)*filt(index);
			else fi=0;
			uu[i] =fi;
		}

		DVector<T> v(N);
		for(int i=0; i<size[0]; i++)
			for(int j=0; j<size[1]; j++)
				v[i]+=prod[j][i];

		T tmp=0;
		for(int i=0; i<size[0]; i++)
			tmp+=uu[i]*v[i];

		val=pow(2,j/2)*tmp/N;
	}

}
		
}

#endif