
#pragma once

#include <stdio.h>
#include <GM_DVector.h>
#include <GM_Array.h>
using namespace GM_lib;

#define MAX 20

namespace Wave_lib {

	template<class T>
	class FilterCoefs {
		public:
			char				name[20];
			double					dim;
			
			// BiOrthoFilters has Two Filter Coefficient Vectors, but
			// OrthoFilters has Only One Filter Coefficient Vectors
			Array<DVector<T> >	values;						

			// not yet in use, will move the filters inside her, after calculation
			Array<DVector<T> > Lo_D;		// LowPass decompositin filters,    2 for the case of BiOrtogonalFilters
			Array<DVector<T> > Hi_D;		// HighPass decompositin filters,   2 for the case of BiOrtogonalFilters
			Array<DVector<T> > Lo_R;		// LowPass reconstruction filters,  2 for the case of BiOrtogonalFilters
			Array<DVector<T> > Hi_R;		// HighPass reconstruction filters, 2 for the case of BiOrtogonalFilters

			bool operator<(const FilterCoefs<T> &f)const;		// for use with GM_Array sorting function
	};

	
	template<class T>
	bool FilterCoefs<T>::operator<(const FilterCoefs<T> &f)const {
		return (dim < f.dim);
	}



	template<class T>
	class Filters {

		protected:
			Array<FilterCoefs<T> > _filter;

			void _qmt(DVector<T> &f, DVector<T> &v, int p=0);					// Quadrature mirror filter calculation

			//virtual void  calcFilterCoefs(double dim, FilterCoefs<T> &dbn) = 0;	// calculates Filter Coefs, eks. Daubechies Coefs
			virtual void calcFilters(FilterCoefs<T> &f, int p=0) = 0;			// calculates High and Low pass wavelet filters
		public:
			Filters();
			virtual ~Filters() {}
			
			virtual void  calcFilterCoefs(double dim, FilterCoefs<T> &dbn) = 0;	// calculates Filter Coefs, eks. Daubechies Coefs

			int _getFilter(double dim);

			// will only give Lo_D_1, Hi_D_1, Lo_R_1, Hi_R_1 for BiOrtogonal filters, Ortogonal filters main func.
			void getFilterD(double dim, DVector<T> &LowPass, DVector<T> &HighPass);
			void getFilterR(double dim, DVector<T> &LowPass, DVector<T> &HighPass);
			
			// Spesial funct in the case of BiOrtogonal filters, but can be used by Ortogonal filters
			void getFilterD(double dim, Array<DVector<T> > &LowPass, Array<DVector<T> > &HighPass);	
			void getFilterR(double dim, Array<DVector<T> > &LowPass, Array<DVector<T> > &HighPass);
			DVector<T> getFilterValues(double dim);
	};



	template<class T>
	Filters<T>::Filters() {
		
	}


	// ************************************************************************
	// **************************    protected func   *************************
	// ************************************************************************

	template<class T>
	int Filters<T>::_getFilter(double dim) {

		int index = -1;
		int size  = _filter.size();
		for (int i=0; i<size; ++i) {
			if (_filter[i].dim == dim) { index=i; break; }
		}

		if (index<0) {
			// must calculate a new filters with dim
			_filter.alwaysInsert(FilterCoefs<T>());
			_filter[size].dim = dim;

			// calculating Filter Coefs
			calcFilterCoefs(dim, _filter[size]);
			index = size;
		}
	
		// calculating LowPass and HighPass filters
		calcFilters(_filter[index]);

		return index;
	}

	template<class T>
	void Filters<T>::_qmt(DVector<T> &f, DVector<T> &v,int p) {

		// Compute quadrature mirror filter.
		v = f.getReversed();
//		v = f.reverse();
		int i=1-p%2;
		for (; i<v.dim(); i+=2) {
			v[i] = -v[i];
		}
	}
	
	// ************************************************************************
	// **************************    public func   ****************************
	// ************************************************************************

	template<class T>
	void Filters<T>::getFilterD(double dim, DVector<T> &LowPass, DVector<T> &HighPass) {

		int index = _getFilter(dim);

		LowPass  = _filter[index].Lo_D[0];
		HighPass = _filter[index].Hi_D[0];

	}


	template<class T>
	void Filters<T>::getFilterR(double dim, DVector<T> &LowPass, DVector<T> &HighPass) {

		int index = _getFilter(dim);

		LowPass  = _filter[index].Lo_R[0];
		HighPass = _filter[index].Hi_R[0];
	}
	
	template<class T>
	void Filters<T>::getFilterD(double dim, Array<DVector<T> > &LowPass, Array<DVector<T> > &HighPass) {
		
		int i;
		int index = _getFilter(dim);
		int size  = _filter[index].Lo_D.size();

		// Give all LowPass decomp filters
		LowPass.setSize(size);
		for (i=0; i<size; ++i)
			LowPass[i]  = _filter[index].Lo_D[i];

		// Give all HighPass decomp filters
		HighPass.setSize(size);
		for (i=0; i<size; ++i)
			HighPass[i] = _filter[index].Hi_D[i];
	}
	
	template<class T>
	void Filters<T>::getFilterR(double dim, Array<DVector<T> > &LowPass, Array<DVector<T> > &HighPass) {
		
		int i;
		int index = _getFilter(dim);
		int size  = _filter[index].Lo_D.size();

		// Give all LowPass recon filters
		LowPass.setSize(size);
		for (i=0; i<size; ++i)
			LowPass[i]  = _filter[index].Lo_R[i];

		// Give all HighPass recon filters
		HighPass.setSize(size);
		for (i=0; i<size; ++i)
			HighPass[i] = _filter[index].Hi_R[i];
	}

	template<class T>
	DVector<T> Filters<T>::getFilterValues(double dim) {

		int index = -1;
		int size  = _filter.size();
		for (int i=0; i<size; ++i) {
			if (_filter[i].dim == dim) { index=i; break; }
		}

		if (index<0) {
			// must calculate a new filters with dim
			_filter.alwaysInsert(FilterCoefs<T>());
			_filter[size].dim = dim;
			calcFilterCoefs(dim, _filter[size]);
			index = size;
		}
		
		return _filter[index].values[0];
	}

} // namespace Wave_lib

