#ifndef _LIFTING_H_
#define _LIFTING_H_

#include "GM_DVector.h"
#include <GM_DMatrix.h>
#include <GM_DArrayN.h>

using namespace GM_lib;

template<class T>
void GetLifting (DArrayN<T,1> &lift, DVector<T> filt, int step, int N, int R, int L, bool type )
{
	DArrayN<double,2> moment;   // Integral and Moments 

	int len,         // number of coefficients in this level 
		step2,       // step size between same coefficients 
		noGammas;    // number of Gamma coeffs 

	
	DVector<int> msize(2);
	msize[0] = L;	msize[1] = R;
	moment.setSize(msize);

	// Initialize Integral and Moments 
	// Integral is equal to one at the beginning since all the filter 
	// coefficients must add to one for each gamma coefficient. 
	// 1st moment = k, 2nd moment = k^3, 3rd moment = k^3, ... 

	for (int j=0 ; j<msize[0] ; j++ )    
		for (int t=0 ; t<msize[1] ; t++ ) 
		{
			if ( j==0 && t==0 )
				moment[j][t] = (T)1.0;   
			else
				moment[j][t] = (T)pow( j,t);
	}

	// Calculate number of coefficients and step size 
	len = (int)(L+step-1)/step;
	step2 = step << 1;

	// Check for number of Gamma coefficients 
	noGammas  = len >> 1;   //Number of Gammas  = floor(len/2) 

	//****************************
	// Decide order of operations 
	//****************************
	if ( type == false ) {   // Direct transform 
		moment =UpdateMoment<T> ( moment, filt, step2, noGammas, len, N, R );
		UpdateLifting<T> ( lift, moment, len, step2, noGammas, R );
	}
	else {   // Inverse transform
		UpdateLifting<T> ( lift, moment, len, step2, noGammas, R );
		moment = UpdateMoment<T> ( moment, values, step2, noGammas, len, N, R );
	}

}

// UpdateMoment function: calculates the integral-moment tuple for the current  level of calculations.
template<class T>
DArrayN<T,2> UpdateMoment (DArrayN<T,2> moment, DVector<T> filter, int step2, int noGammas, int len, int N, int R )
{
	int i, j,              // counters 
		row, col,          // indices of the matrices 
		idxL, idxG,        // pointers to Lambda & Gamma coeffs, resp. 
		top1, top2, top3;  // number of iterations for L<R, L=R, L>R 

		//*************************************
		// Update Integral-Moments information 
		//*************************************

		// Calculate number of iterations for each case 
		top1 = (N>>1) - 1;                 // L < R 
		top3 = (N>>1) - ((len) & 0x01);    // L > R 
		top2 = noGammas - (top1 + top3);   // L = R 

		// Cases where nbr. left Lambdas < nbr. right Lambdas 
		int t=0;   // initialize pointer to first row 
		idxG = step2>>1;      // second coefficient is Gamma 
		for ( row = 1 ; row <= top1 ; row++ ) {
			idxL = 0;   // the first Lambda is always the first coefficient 
			t += N;   // go to next filter row 
			for ( col = 0 ; col < N ; col++ ) {
				for ( j = 0 ; j < R ; j++ ) {
					moment[idxL][j] += filter[t+col]*moment[idxG][j];
				}
				// Jump to next Lambda coefficient 
				idxL += step2;
			}
			idxG += step2;   // go to next Gamma coefficient 
		}

		// Cases where nbr. left Lambdas = nbr. right Lambdas 
		t += N;   // go to last filter row 
		for ( i = 0 ; i < top2 ; i++ ) {
			idxL = i*step2;
			for ( col = 0 ; col < N ; col++ ) {
				for ( j = 0 ; j < R ; j++ ) {
					moment[idxL][j] += filter[t+col]*moment[idxG][j];
				}
				// Jump to next Lambda coefficient 
				idxL += step2;
			}
			idxG += step2;   // go to next Gamma coefficient and stay 
							// in the same row of filter coefficients 
		}

		// Cases where nbr. left Lambdas > nbr. right Lambdas 
		for ( row = top3 ; row >= 1 ; row-- ) {
			idxL = (top2-1)*step2;   // first Lambda is always in this place 
			t -= N;   // go to previous filter row 
			for ( col = N-1 ; col >= 0 ; col-- ) {
				for ( j = 0 ; j < R ; j++ ) {
					moment[idxL][j] += filter[t+col]*moment[idxG][j];
				}
				// Jump to next Lambda coefficient and next filter row 
				idxL += step2;
			}
			idxG += step2;   // go to next Gamma coefficient 
		}

	return moment;
}


// UpdateLifting function: calculates the lifting coefficients using the given integral-moment tuple
template<class T>
void UpdateLifting (DArrayN<T,1> &lifting, DArrayN<T,2> moment, int len, int step2, int noGammas, int R )
{
	int lcIdx,             // index of lifting vector 
		i, j,              // counters 
		row, col,          // indices of the matrices 
		idxL, idxG,        // pointers to Lambda & Gamma coeffs, resp. 
		top1, top2, top3;  // number of iterations for L<R, L=R, L>R 
		
	T d;                 // used by the LU routines 

	//********************************
	// Calculate Lifting Coefficients 
	//********************************

	lcIdx = 0;         // first element of lifting vector 

	// Allocate space for the indx[] vector 
	DVector<int> indx(R+1);
	DVector<T> b(R+1);

	//********************************
	// Calculate Lifting Coefficients 
	//********************************

	lcIdx = 0;         // first element of lifting vector

	DArrayN<T,2> lift;
	DVector<int> lsize(2);
	lsize[0] = R+1;	lsize[1] = R+1;
	lift.setSize(lsize);

	// Calculate number of iterations for each case 
	top1 = (R>>1) - 1;                 // L < R 
	top3 = (R>>1) - ((len) & 0x01);    // L > R 
	top2 = noGammas - (top1 + top3);   // L = R 

    // Cases where nbr. left Lambdas < nbr. right Lambdas 
		idxG = step2>>1;   // second coefficient is Gamma 
		for ( i=0 ; i<top1 ; i++ ) {
			idxL = 0;   // the first Lambda is always the first coefficient 
			for ( col=1 ; col<=R ; col++ ) {
				// Load temporary matrix to be solved 
				for ( row=1 ; row<=R ; row++ )
					lift[row][col]= moment[idxL][row-1];    
				// Jump to next Lambda coefficient 
				idxL += step2;
			}
			// Apply LU decomposition to lift[][] 
			LUdcmp<T> (lift, R, indx, &d);

			// Load independent vector 
			for ( j=1 ; j<=R ; j++)
				b[j] = moment[idxG][j-1];   
			// Apply back substitution to find lifting coefficients 
			LUbksb<T> (lift, R, indx, b);
			for ( col=1 ; col<=R ; col++ )    // save them in lifting vector 
				lifting[lcIdx++] = b[col];

			idxG += step2;   // go to next Gamma coefficient 
		}

		// Cases where nbr. left Lambdas = nbr. right Lambdas 
		for ( i=0 ; i<top2 ; i++ ) {
			idxL = i*step2;
			for ( col=1 ; col<=R ; col++ ) {
				// Load temporary matrix to be solved 
				for ( row=1 ; row<=R ; row++ )
					lift[row][col] = moment[idxL][row-1];    
				// Jump to next Lambda coefficient 
				idxL += step2;
			}
			// Apply LU decomposition to lift[][] 
			LUdcmp<T> (lift, R, indx, &d);

			// Load independent vector 
			for ( j=1 ; j<=R ; j++)
				b[j] = moment[idxG][j-1];   

			// Apply back substitution to find lifting coefficients 
			LUbksb<T> (lift, R, indx, b);
			for ( col=1 ; col<=R ; col++ )    // save them in lifting vector 
				lifting[lcIdx++] = b[col];

			idxG += step2;   // go to next Gamma coefficient 
		}

		// Cases where nbr. left Lambdas > nbr. right Lambdas 
		for ( i=0 ; i<top3 ; i++ )
		{
			idxL = (top2-1)*step2;   // first Lambda is always in this place 
			for ( col=1 ; col<=R ; col++ )
			{
				// Load temporary matrix to be solved 
				for ( row=1 ; row<=R ; row++ )
					lift[row][col] = moment[idxL][row-1];    
				// Jump to next Lambda coefficient 
				idxL += step2;
			}
			// Apply LU decomposition to lift[][] 
			LUdcmp<T> (lift, R, indx, &d);
			// Load independent vector 
			for ( j=1 ; j<=R ; j++)
				b[j] = moment[idxG][j-1];  
			// Apply back substitution to find lifting coefficients 
			LUbksb<T> (lift, R, indx, b);
			for ( col=1 ; col<=R ; col++ )    // save them in lifting vector 
				lifting[lcIdx++] = b[col];

			idxG += step2;   // go to next Gamma coefficient 
		}
}

template<class T>
void LUdcmp (DArrayN<double,2> &a, int n, DVector<int> &indx, T *d )
{
	int i, imax, j, k;
	T big, dum, sum, temp;

	DArrayN<T,1> tmp;   //Store the implicit scaling of each row 
	DVector<int> size(1);
	size[0]=n+1;
	tmp.setSize(size);
	*d=(T)1;                    

	for (i=1;i<=n;i++) {   // Loop over rows to get the implicit scaling 
		big=(T)0;          // information. 
		for (j=1;j<=n;j++)
			if ((temp=(T)fabs((T)a[i][j])) > big)
				big=temp;
		if (big == (T)0.0) exit(0);

		// Nonzero largest element. 
		tmp[i]=(T)1/big;    // Save the scaling. 
	}
	for (j=1;j<=n;j++) {     // This is the loop over columns of Crout's method. 
		for (i=1;i<j;i++) {  // Sum form of a triangular matrix except for i=j. 
			sum=a[i][j];     
			for (k=1;k<i;k++) sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
		}
		big=(T)0;      // Initialize for the search for largest pivot element. 
		imax = 0;        // Set default value for imax 
		for (i=j;i<=n;i++) {  // This is i=j of previous sum and i=j+1...N 
			sum=a[i][j];      // of the rest of the sum. 
			for (k=1;k<j;k++)
				sum -= a[i][k]*a[k][j];
			a[i][j]=sum;
			if ( (dum=tmp[i]*(T)fabs((T)sum)) >= big) {
				big=dum;
				imax=i;
			}
		}
		if (j != imax) {         
			for (k=1;k<=n;k++) {  
				dum=a[imax][k];
				a[imax][k]=a[j][k];
				a[j][k]=dum;
			}
			*d = -(*d);         
			tmp[imax]=tmp[j];       
		}
		indx[j]=imax;
		if (a[j][j] == (T)0.0)
			a[j][j]=(T)1.0e-20;
		// If the pivot element is zero the matrix is singular (at least 
		// to the precision of the algorithm). For some applications on 
		// singular matrices, it is desiderable to substitute TINY for zero. 
		if (j != n) {           // Now, finally divide by pivot element. 
			dum=(T)1/(a[j][j]);
			for ( i=j+1 ; i<=n ; i++ )
				a[i][j] *= dum;
		}
	}                      
}
	    

template<class T>
void LUbksb (DArrayN<double,2> a, int n, DVector<int> indx, DVector<T> &b )
{
	int i,ii=0,ip,j;
		T sum;

		for (i=1;i<=n;i++)    // When ii is set to a positive value, it will 
		{	ip= indx[i];       // become the index of the first nonvanishing 
			sum=b[ip];         // element of b. We now do the forward substitution. 
			b[ip]=b[i];        // The only new wrinkle is to unscramble the 
			if (ii)            // permutation as we go. 
				for (j=ii;j<=i-1;j++) sum -= a[i][j]*b[j];
			else if (sum)      // A nonzero element was encountered, so from now on 
				ii=i;          // we will have to do the sums in the loop above 
			b[i]=sum;
		}

		for (i=n;i>=1;i--) //Do the backsubstitution.
		{   
			sum=b[i];
			for (j=i+1;j<=n;j++) sum -= a[i][j]*b[j];
			b[i]=sum/a[i][i];  // Store a component of the solution vector

			if (( (T)-100 * (T)6E-8 < b[i] ) && ( b[i] < (T)100 * (T)6E-8 ))
				 b[i] = (T)0;   // Verify small numbers. 
		}               
}

#endif