
/*  Using the "Lifting Scheme" and the "Square 2D" method, the coefficients
 *  of a "Biorthogonal Wavelet" or "Second Generation Wavelet" are found for
 *  a two-dimensional data set. The same scheme is used to apply the inverse
 *  operation. */


#include "WaveFLWT.h"
#include <GM_Tid.h>
#include <GM_DArrayN.h>
 
using namespace GM_lib;


//static DArrayN<T,n> filter = NULL;      /* Vector with filter coefficients */
//static DArrayN<T,2> liftingX = NULL;    /* Matrix with lifting coefficients in X */
//static DArrayN<T,2> liftingY = NULL;    /* Matrix with lifting coefficients in Y */

namespace Wave_lib
{

	template<class T, int n>
	T Neville ( const T *x, const T *f, const int n, const T xx )
	{
		register int i,j;
		DArrayN<T,n> vy;
		T y;

		vy = DArrayN<T,n>( 0, (long)(n-1) );   /* allocate a temporary vector */
		for ( i=0; i<n; i++ ) {
			vy[i] = f[i];
			for ( j=i-1; j>=0; j--) {
				T den = x[i] - x[j];
				vy[j] = vy[j+1] + (vy[j+1] - vy[j]) * (xx - x[i]) / den;
			}
		}
		y = vy[0];
		return y;
	}


	template<class T, int n>
	DArrayN<T,n> GetFilter ( const int N )
	{
		static Flt xa[40], ya[40], x;
		static int row, col, cc, Nrows;
		DArrayN<T,n> filter, ptr;

		assert ( N < sizeof xa / sizeof *xa );   /* N is greater than dim(xa) */

		/* Check for validity of N */
		if ( ODD (N) || (N == 0) )
			return (DArrayN<T,n>)NULL;   /* N is not even and greater than zero */

		/* Number of cases for filter calculations */
		Nrows = (N>>1) + 1;    /* N/2 + 1 */

		/* Allocate memory for filter matrix */
		filter = vector( 0, (long)(Nrows*N - 1) );
		ptr = filter;

		/* Generate values of xa */
		xa[0] = (Flt)0.5*(Flt)(1 - N);    /* -N/2 + 0.5 */
		for (col = 1 ; col < N ; col++)
			xa[col] = xa[col-1] + (Flt)1;

		/* Find filter coefficient values */
		filter += ( (Nrows*N) - 1 );   /* go to last position in filter matrix */
		for (row = 0 ; row < Nrows ; row++) {
			x = (Flt)row;
			for (col = 0 ; col < N ; col++) {
				for (cc = 0 ; cc < N ; cc++)
					ya[cc] = (Flt)0;
				ya[col] = (Flt)1;
				*(filter--) = Neville (xa, ya, N, x);
			}
		}
		return ptr;
	}


	template<class T, int n>
	void GetDual ( const int width, const int height, const int N, const int nTilde,
	                               const int levels )
	{
		int row, col, noGammas, nX, nY;
		T maxN;
		DArrayN<T,n> filterPtr;

		/* Calculate Filter Coefficients */
		filter = GetFilter(N);

		/* Allocate lifting coefficient matrices */
		maxN = (T)MAX(N, nTilde) - (T)1;   /* max vanishing moments */
		noGammas = width >> 1;    /* number of Gammas in X direction */
		if (noGammas > 0) {
			nX = (int)logBaseN ( (T)(width-1)/maxN, (T)2 );   /* iterations in X */
			nX = (nX < 0) ? 0 : MIN(levels, nX);   /* find lowest level */
			liftingX = DArrayN<T,n>( 0, (long)(nX-1), 0, (long)(noGammas*nTilde - 1) );
			for ( row=0 ; row<nX ; row++ )
				for ( col=0 ; col<noGammas*nTilde ; col++ )
					liftingX[row][col] = (T)0;
		}
		noGammas = height >> 1;    /* number of Gammas in Y direction */
		if (noGammas > 0) {
			nY = (int)logBaseN ( (T)(height-1)/maxN, (T)2 );   /* iterations in Y */
			nY = (nY < 0) ? 0 : MIN(levels, nY);   /* find lowest level */
			liftingY = matrix( 0, (long)(nY-1), 0, (long)(noGammas*nTilde - 1) );
			for ( row=0 ; row<nY ; row++ )
				for ( col=0 ; col<noGammas*nTilde ; col++ )
					liftingY[row][col] = (T)0;
		}
	}
	

	template<class T, int n>
	void GetReal ( const int width, const int height, const int N, const int nTilde,
	                                const int levels )
	{
		int x, y, k, n, nX, nY, step, max_step, level;
		T  maxN;
		DArrayN<T,n> liftPtr;
		char buf[BUFSIZ];

		/* Calculate number of iterations n.  */
		maxN = (T)MAX(N, nTilde) - (T)1;   /* max vanishing moments */
		/* Iterations in X */
		nX = (width == 1) ? 0 : (int)logBaseN ( (T)(width-1)/maxN, (T)2 );
		nX = (nX < 0) ? 0 : MIN (levels, nX);   /* find lowest level */
		/* Iterations in Y */
		nY = (height == 1) ? 0 : (int)logBaseN ( (T)(height-1)/maxN, (T)2 );
		nY = (nY < 0) ? 0 : MIN (levels, nY);   /* find lowest level */
		/* Total number of iterations */
		n = MAX(nX, nY);   /* find minimum number of iterations */
		max_step = (n==0) ? 0 : 1<<n;   /* maximum step */

		/* Find lifting coefficients for all levels */
		level = 0;
		for ( step = 1 ; step < max_step ; step <<= 1 )
		{
			if (nX-- > 0) 
			{
				/* Get lifting coefficients for rows */
				GetLifting (liftingX[nX], filter, step, N, nTilde, width, FALSE, FALSE);
			}

			if (nY-- > 0) 
			{
				/* Get lifting coefficients for columns */
				GetLifting (liftingY[nY], filter, step, N, nTilde, height, TRUE, FALSE);
			}
			level++;   /* go to next level */
		}
	}


	template<class T, int n>
	DArrayN<T,n> FLWT1D (DArrayN<T,n> Data, const int width,
						const int N, const int nTilde,
						const int levels, const bool inverse )
	{
		int i, k, x, s,       /* counters */
			step, max_step,   /* step size and max./init. step */
			nX;               /* number of iterations in X and Y directions */
		T maxN;             /* maximum number of vanishing moments */

		//Use GetDual() and GetReal() before calling this function
  
			
		return Data;
                  
		
	   // GetDual(width,0, N,  nTilde, levels);GetReal( width,0, N,  nTilde, levels);

		/* Calculate number of iterations n.  */
		maxN = (T)((N > nTilde)? (N):(nTilde)) - (T)1;   /* max vanishing moments */
		/* Iterations in X */
		nX = (width == 1) ? 0 : (int)((T)log((T)(width-1)/maxN)/(T)log((T)2) );
		nX = (nX < 0) ? 0 : ((levels < nX) ? (levels):(nX));   /* find lowest level */


		if ( !inverse ) {    /* Forward transform */
			/* Calculate wavelet coefficients */
			max_step = (nX==0) ? 0 : 1<<nX;   /* maximum step */
			for ( step=1, s=nX-1 ; step<max_step ; step<<=1, s-- ) {
				/* Apply forward transform to the vector */
				//FLWT1D_Predict ( Data, (long)width, 1,(long)step, (long)N );
				//FLWT1D_Update ( Data, (long)width, 1,(long)step, (long)nTilde, &liftingX[s] );
			}
		}
	}

	
	template<class T, int n>
	void FLWT1D_Predict (DArrayN<T,n> vect, const long size, const long incr,
								const long step, const long N )
	{
		register DArrayN<T,n> lambdaPtr,       /* pointer to Lambda coefficients */
						gammaPtr,        /* pointer to Gamma coefficients */
						fP, filterPtr;   /* pointers to filter coefficients */
		register long len,               /* number of coefficients at current level */
					j,                 /* counter for the filter cases */
					stepIncr;          /* step size between coefficients of the same type */

		long stop1,                      /* number of cases when L < R */
			stop2,                      /* number of cases when L = R */
			stop3,                      /* number of cases when L > R */
			soi;                        /* increment for the middle cases */

		/************************************************/
		/* Calculate values of some important variables */
		/************************************************/
		len       = CEIL(size, step);   /* number of coefficients at current level */
		stepIncr  = (step*incr) << 1;   /* step size betweeen coefficients */

		/************************************************/
		/* Calculate number of iterations for each case */
		/************************************************/
		j     = ODD(len);
		stop1 = N >> 1;
		stop3 = stop1 - j;                /* L > R */
		stop2 = (len >> 1) - N + 1 + j;   /* L = R */
		stop1--;                          /* L < R */

		/***************************************************/
		/* Predict Gamma (wavelet) coefficients (odd guys) */
		/***************************************************/

		filterPtr = filter + N;   /* position filter pointer */

		/* Cases where nbr. left Lambdas < nbr. right Lambdas */
		gammaPtr = vect + (stepIncr >> 1);   /* second coefficient is Gamma */
		while(stop1--) {
			lambdaPtr = vect;   /* first coefficient is always first Lambda */
			j = N;
			do {   /* Gamma update (Gamma - predicted value) */
				*(gammaPtr) -= (*(lambdaPtr)*(*(filterPtr++)));
				lambdaPtr   += stepIncr;   /* jump to next Lambda coefficient */
			} while(--j);   /* use all N filter coefficients */
			/* Go to next Gamma coefficient */
			gammaPtr += stepIncr;
		}

		/* Cases where nbr. left Lambdas = nbr. right Lambdas */
		soi = 0;
		while(stop2--) {
			lambdaPtr = vect + soi;   /* first Lambda to be read */
			fP = filterPtr;   /* filter stays in same values for this cases */
			j = N;
			do {   /* Gamma update (Gamma - predicted value) */
				*(gammaPtr) -= (*(lambdaPtr)*(*(fP++)));
				lambdaPtr   += stepIncr;   /* jump to next Lambda coefficient */
			} while(--j);   /* use all N filter coefficients */
			/* Move start point for the Lambdas */
			soi += stepIncr;
			/* Go to next Gamma coefficient */
			gammaPtr += stepIncr;
		}

		/* Cases where nbr. left Lambdas > nbr. right Lambdas */
		fP = filterPtr;   /* start going backwards with the filter coefficients */
		vect += (soi-stepIncr);   /* first Lambda is always in this place */
		while (stop3--) {
			lambdaPtr = vect;   /* position Lambda pointer */
			j = N;
			do {   /* Gamma update (Gamma - predicted value) */
				*(gammaPtr) -= (*(lambdaPtr)*(*(--fP)));
				lambdaPtr   += stepIncr;   /* jump to next Lambda coefficient */
			} while(--j);   /* use all N filter coefficients */
			/* Go to next Gamma coefficient */
			gammaPtr += stepIncr;
		}
	}


	template<class T, int n>
	void FLWT1D_Update (DArrayN<T,n> vect, const long size, const long incr,
								const long step, const long nTilde,
								const DArrayN<T,n>* lc )
	{
		register DArrayN<T,n> lcPtr,   /* pointer to lifting coefficient values */
						vL,      /* pointer to Lambda values */
						vG;      /* pointer to Gamma values */
		register long j;         /* counter for the lifting cases */

		long len,                /* number of coefficietns at current level */
			stop1,              /* number of cases when L < R */
			stop2,              /* number of cases when L = R */
			stop3,              /* number of cases when L > R */
			noGammas,           /* number of Gamma coefficients at this level */
			stepIncr,           /* step size between coefficients of the same type */
			soi;                /* increment for the middle cases */


		/************************************************/
		/* Calculate values of some important variables */
		/************************************************/
		len      = CEIL(size, step);   /* number of coefficients at current level */
		stepIncr = (step*incr) << 1;   /* step size between coefficients */
		noGammas = len >> 1 ;          /* number of Gamma coefficients */

		/************************************************/
		/* Calculate number of iterations for each case */
		/************************************************/
		j	  = ODD(len);
		stop1 = nTilde >> 1;
		stop3 = stop1 - j;                   /* L > R */
		stop2 = noGammas - nTilde + 1 + j;   /* L = R */
		stop1--;                             /* L < R */

		/**********************************/
		/* Lift Lambda values (even guys) */
		/**********************************/

		lcPtr = lc;   /* position lifting pointer */

		/* Cases where nbr. left Lambdas < nbr. right Lambdas */
		vG = vect + (stepIncr >> 1);   /* second coefficient is Gamma */
		while(stop1--) {
			vL = vect;   /* lambda starts always in first coefficient */
			j = nTilde;
			do {
				*(vL) += (*(vG)*(*(lcPtr++)));   /* lift Lambda (Lambda + lifting value) */
				vL    += stepIncr;               /* jump to next Lambda coefficient */
			} while(--j);   /* use all nTilde lifting coefficients */
			/* Go to next Gamma coefficient */
			vG += stepIncr;
		}

		/* Cases where nbr. left Lambdas = nbr. right Lambdas */
		soi = 0;
		while(stop2--) {
			vL = vect + soi;   /* first Lambda to be read */
			j = nTilde;
			do {
				*(vL) += (*(vG)*(*(lcPtr++)));   /* lift Lambda (Lambda + lifting value) */
				vL    += stepIncr;               /* jump to next Lambda coefficient */
			} while(--j);   /* use all nTilde lifting coefficients */
			/* Go to next Gamma coefficient */
			vG  += stepIncr;
			/* Move start point for the Lambdas */
			soi += stepIncr;
		}

		/* Cases where nbr. left Lambdas = nbr. right Lambdas */
		vect += (soi - stepIncr);   /* first Lambda is always in this place */
		while(stop3--) {
			vL = vect;   /* position Lambda pointer */
			j = nTilde;
			do {
				*(vL) += (*(vG)*(*(lcPtr++)));   /* lift Lambda (Lambda + lifting value) */
				vL    += stepIncr;               /* jump to next Lambda coefficient */
			} while(--j);   /* use all nTilde lifting coefficients */
			/* Go to next Gamma coefficient */
			vG += stepIncr;
		}
	}
}