#ifndef _WAVEPACKETS_H_
#define _WAVEPACKETS_H_

#include <GM_Array.h>
#include "GM_DVector.h"
#include <GM_DArrayN.h>
#include <Definitions.h>

#include "Lifting.h"
#include "CostFunction.h"
#include "WaveFLWT.h"

using namespace GM_lib;

template<class T>  DVector<int>
DFLWTPackets (DArrayN<T,2> &Data, int N, int nTilde, int maxLevel, int *blocks)
{
   Hedge<T> *graph;        // hedge with the best basis of Data 
   int  nX, nY,          // number of levels in X and Y 
        L;               // maximum level in decomposition 
                      // size of contents vector in hedge 
    T maxN,            // maximum number of vanishing moments 
        cost;            // cost of the best basis tree 

	DVector<int> size;
	Data.getSize(size);

    // Calculate maximum level 
    maxN = (T)((N > nTilde)? (N):(nTilde)) - (T)1;
    nX = (size[0] == 1) ? 0 : (int)((T)log((T)(size[0]-1)/maxN)/(T)log((T)2) );
	nX = (nX < 0) ? 0 : (maxLevel<nX)? maxLevel:nX;
    nY = (size[1] == 1) ? 0 : (int)((T)log((T)(size[1]-1)/maxN)/(T)log((T)2) );
	nY = (nY < 0) ? 0 : (maxLevel<nY)? maxLevel:nY;
	L = nX > nY ? nX : nY;

    // Allocate initial hedge 
    *blocks = (1<<2*L) + 1;
	T** bestCoeffs = (T**)InitMem2D<T>(ceil((T)(size[0]*size[1])/(T)*blocks),*blocks );
    DVector<int> levels(*blocks); // levels in which the regions are located

    graph = new Hedge<T> (*blocks, bestCoeffs, levels, NULL );
    graph->blocks = 0;
    
    // Find best basis 
    cost = bbwp2<T>( graph, Data, size[0], size[1], N, nTilde, 0, L );
    
    // Put results in final buffer 
	DVector<int> outlevels(graph->blocks); 
    *blocks = graph->blocks;
  
	DArrayN<T,1> container;
	DVector<int> csize(1);
	csize[0]=size[0]*size[1];
	container.setSize(csize);

	for (int i=0, t=0; i<graph->blocks ; i++ )
	{
		int length = csize[0]/(1<<(2*graph->levels[i]));
		outlevels[i] = graph->levels[i];
		for (int j=0 ; j<length ; j++, t++)
			container[t] = graph->contents[i][j];
	}

	for (int i=0, t=0; i<size[0]; i++) 
		for (int j=0; j<size[1]; j++) 
			Data[i][j]= container[t++];

	return outlevels;
}


template<class T> void
IFLWTPackets (DArrayN<T,2> &Data, int blocks, DVector<int> levels, int N, int nTilde )
{
	int length;

    Hedge<T> *graph;
	DVector<int> size;
	Data.getSize(size);

    // Allocate a hedge for the manipulation of the basis 
	T** content = (T**)malloc( (size_t)(blocks * sizeof( T* )) );   // row of new pointers 
    graph  = new Hedge<T>( blocks, content, levels, NULL );

	T** tmp =InitMem2D<T>( size[0], size[1] );
	for (int i=0 ; i<size[0] ; i++ )
		for (int j=0 ; j<size[1] ; j++ )
			tmp[i][j]=Data[i][j];

    // Extract hedge of coefficients 
	DArrayN<T,1> container;
	DVector<int> csize(1);
	csize[0]=size[0]*size[1];
	container.setSize(csize);
 
	int column = 0;
    int maxl = -999;
    for (int i=0 ; i<graph->blocks ; i++ ) {
		maxl = ( maxl > levels[i] ) ? maxl : levels[i];      // find maximum level 
        length = csize[0]/(1<<(2*levels[i]));   // length of vector block 
        graph->contents[i] = tmp[0] + column;   // pointer to beginning of vector block 
        column += length;                        // place for next column pointer 
    }

	DArrayN<T,1> out;
	Array<DArrayN<T,2> > K;
	K.setSize(4);

    // Reconstruct original data
    if ( maxl > 0 )
	{
		while ( maxl ) {
			// Find the nodes in this level and merge them 
			i = 0;
			while ( i < graph->blocks ) {
				int block;

				// Find a node in this level 
				while ( levels[i] < maxl && i < graph->blocks ) i++; 
				if ( i >= graph->blocks )
					break;
				block = i;
				length = csize[0]/(1<<(2*levels[block]));

				// Take all the four siblings and merge them 

				DVector<int> ksize(2);
				ksize[0]=ksize[1]=(int)sqrt( (T)length );// assuming all regions are square 
				for(int i=0; i<K.size(); i++)
					K[i].setSize(ksize);

				int k1=0,k2=0,k3=0,k4=0;
				for (int x=0; x<ksize[0] ; x++ )
					for (int y=0 ; y<ksize[1] ; y++ )
					{
						K[0][x][y] = graph->contents[block][k1++];
						K[1][x][y] = graph->contents[block+1][k2++];
						K[2][x][y] = graph->contents[block+2][k3++];
						K[3][x][y] = graph->contents[block+3][k4++];
					}

				out = ISFLWT2D<T>( K, ksize[0]<<1, ksize[1]<<1, N, nTilde );
	          
				// Update pointers and levels 
				graph->blocks -= 3;
				levels[block] -= 1;
				i = block + 1;
				for (int j=i ; j<graph->blocks ; j++ ) {
					graph->contents[j] = graph->contents[j+3];
					levels[j] = levels[j+3];
				}
				// Copy new node in graph->contents 
				length *= 4;
				for (int j=0 ; j<length ; j++ )
						graph->contents[block][j] =  out[j];
			}
			maxl--; 
		}
    
        for (int i=0, t=0; i<graph->blocks ; i++ )
		{
            length = csize[0]/(1<<(2*levels[i]));
            for (int j=0 ; j<length ; j++, t++ ) 
                container[t] = graph->contents[i][j];
        }
	}

	for (int i=0, t=0; i<size[0]; i++) 
		for (int j=0; j<size[1]; j++,t++) 
			Data[i][j]= container[t];
}


template<class T> DVector<int>
DFLWTPackets1D ( DArrayN<T,1> &Data, int N, int nTilde, int maxLevel, int *blocks)
{
    Hedge<T> *graph;        // hedge with the best basis of Data 
 
	int nX;              // number of levels in X 
    T maxN,            // maximum number of vanishing moments 
        cost;            // cost of the best basis tree 

	DVector<int> size;
	Data.getSize(size);

       // Calculate maximum level 
    maxN = (T)((N > nTilde)? (N):(nTilde)) - (T)1;
    nX = (size[0] == 1) ? 0 : (int)((T)log((T)(size[0]-1)/maxN)/(T)log((T)2) );
	nX = (nX < 0) ? 0 : (maxLevel<nX)? maxLevel:nX;

    // Allocate initial hedge 
    *blocks = (1<<nX) + 1;
	DVector<int> levels(*blocks); // levels in which the regions are located 

	T** bestCoeffs = (T**)InitMem2D<T>(ceil(size[0]/(T)*blocks),*blocks );
	graph = new Hedge<T>( *blocks, bestCoeffs, levels, NULL );
    graph->blocks = 0;
    
    // Find best basis 
    cost = bbwp1<T>( graph, Data, size[0], N, nTilde, 0, nX);

    // Put results in final buffer 
    DVector<int> outlevels(graph->blocks);
    *blocks = graph->blocks;

	int t=0;
    for (int i=0; i<graph->blocks ; i++ ) {
        int length = size[0]/(1<<(graph->levels[i]));

        outlevels[i] = graph->levels[i];
        for (int j=0 ; j<length ; j++, t++ )
           Data[t] = graph->contents[i][j];
    }

	return outlevels;
}
 
template<class T> void
IFLWTPackets1D ( DArrayN<T,1> &Data, int blocks, DVector<int> levels, int N, int nTilde )
{

    Hedge<T> *graph;
    int length, maxl;

	DVector<int> insize;
	Data.getSize(insize);

    // Allocate a hedge for the manipulation of the basis 
    T** content = (T**)malloc( (size_t)(blocks * sizeof( T* )) );   // row of new pointers 

    // row of new pointers 
    graph  = new Hedge<T>( blocks, content, levels, NULL );
	T *tmp = (T*) malloc((unsigned)(insize[0])*sizeof(T));
	for (int i=0 ; i<insize[0] ; i++ )
		tmp[i]= Data[i];

    // Extract hedge of coefficients 
    int  offset=0;
    maxl = -999;
    for (int i=0 ; i<graph->blocks ; i++ ) 
	{
        maxl = (maxl > levels[i]) ? maxl:levels[i];     // find maximum level 
        length = insize[0]/(1<<(levels[i]));  // length of vector block 
      
		graph->contents[i] = tmp+offset;      // pointer to beginning of vector block 
        offset+= length; 
    }
	
	DArrayN<T,1> out;
	Array<DArrayN<T,1> > K;
	K.setSize(2);

    // Reconstruct original image 
  if ( maxl > 0 ) 
	{
		 while ( maxl) {

			// Find the nodes in this level and merge them 
			int i = 0;
			while ( i < graph->blocks ) {
				int block;

				// Find a node in this level 
				while ( levels[i] < maxl && i < graph->blocks ) i++;
				if ( i >= graph->blocks )
					break;
				block = i;
				length = insize[0]/(1<<(levels[block]));
				// Take the two siblings and merge them 
				
				DVector<int> size(1);
				size[0]=length;
				for(int k=0; k<K.size(); k++)
					K[k].setSize(size);

				for (int x=0 ; x<size[0] ; x++ )
				{
					K[0][x] = graph->contents[block][x];
					K[1][x] = graph->contents[block+1][x];
				}

				out = ISFLWT1D<T> ( K, size[0]<<1, N, nTilde );
	          
				// Update pointers and levels 
				graph->blocks--;
				levels[block]--;
				i = block + 1;
				for (int j=i ; j<graph->blocks ; j++ ) {
					graph->contents[j] = graph->contents[j+1];
					levels[j] = levels[j+1];
				}
				// Copy new node in graph->contents 
				length *= 2;
				for ( j=0 ; j<length ; j++ )
					graph->contents[block][j] = out[j];
			
			}
			maxl--;
		}

		int t=0;
        for ( i=0; i<graph->blocks ; i++ ) 
		{
            length = insize[0]/(1<<(levels[i]));
            for (int j=0 ; j<length ; j++,t++ ) 
                Data[t] = graph->contents[i][j];
        }
    }
}

template<class T> T**
InitMem2D( int width, int height )
{
    T** buff = (T**)malloc( (size_t)(height*sizeof(T*)) );  // First allocate a row of pointers 
    buff[0] = (T*)malloc( (size_t)(width*height*sizeof(T))); // Allocate entire 2D vector 

    // Set the row-pointers to point correctly into the 2D vector 
    for (int i=1 ; i<height ; ++i ) 
        buff[i] = buff[i-1] + width;

    return buff;
}

#endif